#!/bin/bash

# This script checks compatibility of user supplied CXXFLAGS with
# the GMP library.
# Exit is 0 iff CXXFLAGS and GMP are compatible.

# This script expects two args:
#   first is the full path of the GMP library,
#   second is the full path of the GMP header file.

if [ "$#" != "4" ]
then
  echo "$0: ERROR: requires 4 args (CXX, CXXFLAGS, full path of gmp.h & full path of libgmp.a)"
  exit 1
fi

CXX="$1"
CXXFLAGS="$2"
GMP_H="$3"
GMP_LIB="$4"

GMP_HDR_DIR=`dirname "$GMP_H"`

umask 22

# Create tmp directory, put C prog in it, compile and run.
# TMP_DIR depends on hostname, userid, and process number to avoid unfortunate
# name clashes if several people try to install CoCoALib simultaneously.
TMP_DIR=gmp-check-cxxflags-$HOSTNAME-$UID-$$
/bin/rm -rf $TMP_DIR && mkdir $TMP_DIR
if [ $? -ne 0 ]; then
  echobox "$0: failed to create temporary directory \"$TMP_DIR\""
  exit 1
fi

cd $TMP_DIR
cat > TestProg.c <<EOF
#include "gmp.h"

int main()
{
  mpz_t A; mpz_init_set_ui(A,1);
  mpz_t B; mpz_init_set_ui(B,1);
  mpz_add(A, A, B);
  if (mpz_cmp_ui(A, 2) != 0)
    return 1;
  return 0;
}
EOF


# Compile test prog using given CXX and CXXFLAGS, and GMP header and library
$CXX $CXXFLAGS TestProg.c -o TestProg -I "$GMP_HDR_DIR" "$GMP_LIB" 2> /dev/null

# Check whether compilation failed; if so, complain.
if [ $? -ne 0 ]
then
  cd ..
  /bin/rm -rf $TMP_DIR
  exit 2
fi

# Compilation succeeded, so try running $PROG.
./TestProg  2>/dev/null

# Check whether execution failed; if so, complain (probably linker problems).
if [ $? -ne 0 ]
then
  cd ..
  /bin/rm -rf $TMP_DIR
  exit 3
fi

cd ..
/bin/rm -rf $TMP_DIR
exit 0
