#ifdef CoCoA_WITH_NORMALIZ
//   Copyright (c)  2010-2014 Anna M. Bigatti, Christof Soeger

//   This file is part of the source of CoCoALib, the CoCoA Library.

//   CoCoALib is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.

//   CoCoALib is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.

//   You should have received a copy of the GNU General Public License
//   along with CoCoALib.  If not, see <http://www.gnu.org/licenses/>.


// Source code for Normaliz integration

#include "CoCoA/ExternalLibs-Normaliz.H"
#include "CoCoA/BigInt.H"
#include "CoCoA/BigRat.H"
#include "CoCoA/IntOperations.H"
#include "CoCoA/TmpPPVector.H"
#include "CoCoA/convert.H"
#include "CoCoA/error.H"
#include "CoCoA/matrix.H"
#include "CoCoA/SparsePolyRing.H"

#include "CoCoA/io.H"  // just for debugging
#include <iostream>  // just for debugging
#include <sstream>  // for ErrorMessage

#include "gmpxx.h"
#include "libnormaliz/cone.h"
#include "libnormaliz/map_operations.h"
using libnormaliz::Cone;
using libnormaliz::ConeProperties;



// #include <list>
// using std::list;
#include <map>
using std::map;
#include <vector>
using std::vector;



namespace CoCoA
{


  class TemporarilySet
  {
  public:
    TemporarilySet(bool& var, bool value): myVar(var), myOrigValue(var) { myVar = value; }
    ~TemporarilySet() { myVar = myOrigValue; }
  private: // data members
    bool& myVar;
    bool myOrigValue;
  };

  namespace Normaliz
  {

    libnormaliz::InputType ToInputType(const std::string& TypeString)
    {
      return libnormaliz::to_type(TypeString);
    }

    class ConeImpl: protected IntrusiveReferenceCount
    {
      friend class SmartPtrIRC<const ConeImpl>; // Morally "friend Cone", so it can alter reference count.

      public:
        ConeImpl(const std::vector<std::vector<BigInt> >& v, libnormaliz::Type::InputType InputType);
        ConeImpl(const std::map< libnormaliz::InputType, std::vector<std::vector<BigInt> > >& m);
        ConeImpl(const PPVector& v, libnormaliz::Type::InputType InputType);
        friend std::vector<std::vector<BigInt> > HilbertBasis(const cone& c);
        friend std::vector<std::vector<BigInt> > ModuleGenerators(const cone& c);
        friend std::vector<std::vector<BigInt> > Generators(const cone& c);
        friend std::vector<std::vector<BigInt> > ExtremeRays(const cone& c);
        friend std::vector<std::vector<BigInt> > VerticesOfPolyhedron(const cone& c);
        friend std::vector<std::vector<BigInt> > Deg1Elements(const cone& c);
        friend std::vector<std::vector<BigInt> > GeneratorsOfToricRing(const cone& c);
        friend std::vector<std::vector<BigInt> > SupportHyperplanes(const cone& c);
        friend std::vector<std::vector<BigInt> > Equations(const cone& c);
        friend std::vector<std::vector<BigInt> > Congruences(const cone& c);
        friend std::vector<std::vector<BigInt> > ExcludedFaces(const cone& c);
        //friend std::vector<BigInt> HVector(const cone& c);
        friend HPSeries HilbertSeries(const cone& c);
        friend RingElem HilbertPoly(const cone& c);
        friend BigRat multiplicity(const cone& c);
        friend std::vector<BigRat> grading(const cone& c);
        friend bool IsPointed(const cone& c);
        friend bool IsInhomogeneous(const cone& c);
        friend bool IsIntegrallyClosed(const cone& c);
        friend bool IsDeg1HilbertBasis(const cone& c);
        friend long EmbeddingDim(const cone& c);
        friend long rank(const cone& c);
        friend long RecessionRank(const cone& c);
        friend long AffineDim(const cone& c);
        friend long ModuleRank(const cone& c);
        friend std::vector<BigInt> dehomogenization(const cone& c);
        friend BigInt shift(const cone& c);

      public:
        void myComputation(const libnormaliz::ConeProperties& CPs) const;
        void myComputation(libnormaliz::ConeProperty::Enum CP) const;    
        void myComputation(libnormaliz::ConeProperty::Enum CP1, libnormaliz::ConeProperty::Enum CP2) const;    
        void myComputation() const;   //default: compute everything possible 
        bool isComputed(libnormaliz::ConeProperty::Enum CP) const;    
        bool isComputed(libnormaliz::ConeProperty::Enum CP1, libnormaliz::ConeProperty::Enum CP2) const;    

      private:
        mutable bool mySmallConeIsGood;
        mutable libnormaliz::Cone<long> myConeLong; // be careful!
        mutable libnormaliz::Cone<mpz_class> myConeMPZ; // be careful!
    };


    // implementation of cone

    cone::cone(const std::vector<std::vector<BigInt> >& v, libnormaliz::Type::InputType InputType) : mySmartPtr(new ConeImpl(v, InputType)) {}

    cone::cone(const std::map< libnormaliz::InputType, std::vector<std::vector<BigInt> > >& m) : mySmartPtr(new ConeImpl(m)) {}

    cone::cone(ConstMatrixView M, libnormaliz::Type::InputType InputType) : mySmartPtr(new ConeImpl(MatrixToVecVecBigInt(M), InputType)) {}

    cone::cone(const ConeImpl* ptr): mySmartPtr(ptr) {}
    cone::cone(const cone& c) : mySmartPtr(c.mySmartPtr) {}
    cone::~cone() {}

    void cone::myComputation(const libnormaliz::ConeProperties& CPs) const { mySmartPtr->myComputation(CPs); }
    void cone::myComputation(libnormaliz::ConeProperty::Enum CP) const { mySmartPtr->myComputation(CP); }
    void cone::myComputation(libnormaliz::ConeProperty::Enum CP1, libnormaliz::ConeProperty::Enum CP2) const { mySmartPtr->myComputation(CP1, CP2); }
    void cone::myComputation() const { mySmartPtr->myComputation(); }
    bool cone::isComputed(libnormaliz::ConeProperty::Enum CP) const { return mySmartPtr->isComputed(CP); }
    bool cone::isComputed(libnormaliz::ConeProperty::Enum CP1, libnormaliz::ConeProperty::Enum CP2) const { return mySmartPtr->isComputed(CP1, CP2); }
    const ConeImpl* cone::operator->() const { return mySmartPtr.operator->(); }

    namespace  // conversion functions, implementation at the end of the file
    {
      std::vector<BigInt> LongVToBigIntV(const std::vector<long>& VIn);
      std::vector<BigInt> MPZ_classVToBigIntV(const std::vector<mpz_class>& VIn);
      std::vector<BigRat> LongVToBigRatV(const std::vector<long>& VIn, const BigInt& denom);
      std::vector<BigRat> MPZ_classVToBigRatV(const std::vector<mpz_class>& VIn, const BigInt& denom);
      std::vector<long> BigIntVToLongV(const std::vector<BigInt>& VIn);
      std::vector<mpz_class> BigIntVToMPZ_classV(const std::vector<BigInt>& VIn);
      void ConvertFromNormaliz(std::vector<std::vector<BigInt> >& v, const std::vector<std::vector<long> >& l);
      void ConvertFromNormaliz(std::vector<std::vector<BigInt> >& v, const std::vector<std::vector<mpz_class> >& l);
      std::vector<std::vector<long> >  ReturnVecVecLong(const std::vector<std::vector<BigInt> >& v);
      map< libnormaliz::InputType, vector<vector<long> > >  ReturnMapVecVecLong(const map< libnormaliz::InputType, vector<vector<BigInt> > > m);
      std::vector<std::vector<mpz_class> >  ReturnVecVecMPZ_class(const std::vector<std::vector<BigInt> >& v);
      map< libnormaliz::InputType, vector<vector<mpz_class> > >  ReturnMapVecVecMPZ_class(const map< libnormaliz::InputType, vector<vector<BigInt> > > m);
      std::vector<std::vector<BigInt> > PPVectorToVecVecBigInt(const PPVector& ppv);
      void VecVecBigIntToPPVector(PPVector& ppv, const std::vector<std::vector<BigInt> >& M);
    } // end of anonymous namespace

    //this is used in CoCoA5 methods, so not in the anonymous namespace
    std::vector<std::vector<BigInt> > MatrixToVecVecBigInt(ConstMatrixView M);


    // implementation of ConeImpl
    ConeImpl::ConeImpl(const std::vector<std::vector<BigInt> >& v, libnormaliz::Type::InputType InputType):
          myConeLong(ReturnVecVecLong(v), InputType),
          myConeMPZ(ReturnVecVecMPZ_class(v), InputType)
      {
        mySmallConeIsGood = (myConeLong.getBasisChange().get_rank() > 0);
      }

    ConeImpl::ConeImpl(const std::map< libnormaliz::InputType, std::vector<std::vector<BigInt> > >& m):
      myConeLong(ReturnMapVecVecLong(m)),
      myConeMPZ(ReturnMapVecVecMPZ_class(m))
    {}

    void ConeImpl::myComputation(const libnormaliz::ConeProperties& CPs) const
    {
      try
      {
        if (mySmallConeIsGood)
        {
          TemporarilySet save(libnormaliz::test_arithmetic_overflow, true);
          libnormaliz::ConeProperties missing = myConeLong.compute(CPs);
          if (missing.any())
          {
            std::ostringstream os;
            os << missing;
            CoCoA_ERROR("Normaliz cannot compute "+ os.str(), "myComputation long");
            return;
          }
        }
      }
      catch (libnormaliz::ArithmeticException&) { mySmallConeIsGood = false; }
      if ( !mySmallConeIsGood )
      {
        libnormaliz::ConeProperties missing = myConeMPZ.compute(CPs);
        if (missing.any())
        {
          std::ostringstream os;
          os << missing;
          CoCoA_ERROR("Normaliz cannot compute "+ os.str(), "myComputation MPZ");
        }
      }
    }

    void ConeImpl::myComputation(libnormaliz::ConeProperty::Enum CP) const
    {
      myComputation(ConeProperties(CP));
    }

    void ConeImpl::myComputation(libnormaliz::ConeProperty::Enum CP1,
                                 libnormaliz::ConeProperty::Enum CP2) const
    {
      myComputation(ConeProperties(CP1, CP2));
    }

    void ConeImpl::myComputation() const
    {
      myComputation(ConeProperties(libnormaliz::ConeProperty::DefaultMode));
    }

    bool ConeImpl::isComputed(libnormaliz::ConeProperty::Enum CP) const
    {
      if (mySmallConeIsGood)
      {
        return myConeLong.isComputed(CP);
      }
      else
      {
        return myConeMPZ.isComputed(CP);
      }
    }

    bool ConeImpl::isComputed(libnormaliz::ConeProperty::Enum CP1,
                              libnormaliz::ConeProperty::Enum CP2) const
    {
        return isComputed(CP1) && isComputed(CP2);
    }

    // friend functions which are the interface to the user
    std::vector<std::vector<BigInt> > HilbertBasis(const cone& c)
    {
      std::vector<std::vector<BigInt> > v;
      c->myComputation(libnormaliz::ConeProperty::HilbertBasis);
      if (c->mySmallConeIsGood)
        ConvertFromNormaliz(v, c->myConeLong.getHilbertBasis());
      else
        ConvertFromNormaliz(v, c->myConeMPZ.getHilbertBasis());
      return v;
    }

    std::vector<std::vector<BigInt> > ModuleGenerators(const cone& c)
    {
      std::vector<std::vector<BigInt> > v;
      c->myComputation(libnormaliz::ConeProperty::ModuleGenerators);
      if (c->mySmallConeIsGood)
        ConvertFromNormaliz(v, c->myConeLong.getModuleGenerators());
      else
        ConvertFromNormaliz(v, c->myConeMPZ.getModuleGenerators());
      return v;
    }

    std::vector<std::vector<BigInt> > Deg1Elements(const cone& c)
    { 
      std::vector<std::vector<BigInt> > v;
      c->myComputation(libnormaliz::ConeProperty::Deg1Elements);
      if (c->mySmallConeIsGood)
        ConvertFromNormaliz(v, c->myConeLong.getDeg1Elements());
      else
        ConvertFromNormaliz(v, c->myConeMPZ.getDeg1Elements());
      return v;
    }

    std::vector<std::vector<BigInt> > Generators(const cone& c)
    {
      std::vector<std::vector<BigInt> > v;
      c->myComputation(libnormaliz::ConeProperty::Generators);
      if (c->mySmallConeIsGood)
        ConvertFromNormaliz(v, c->myConeLong.getGenerators());
      else
        ConvertFromNormaliz(v, c->myConeMPZ.getGenerators());
      return v;
    }

    std::vector<std::vector<BigInt> > ExtremeRays(const cone& c)
    {
      std::vector<std::vector<BigInt> > v;
      c->myComputation(libnormaliz::ConeProperty::ExtremeRays);
      if (c->mySmallConeIsGood)
        ConvertFromNormaliz(v, c->myConeLong.getExtremeRays());
      else
        ConvertFromNormaliz(v, c->myConeMPZ.getExtremeRays());
      return v;
    }

    std::vector<std::vector<BigInt> > VerticesOfPolyhedron(const cone& c)
    {
      std::vector<std::vector<BigInt> > v;
      c->myComputation(libnormaliz::ConeProperty::VerticesOfPolyhedron);
      if (c->mySmallConeIsGood)
        ConvertFromNormaliz(v, c->myConeLong.getVerticesOfPolyhedron());
      else
        ConvertFromNormaliz(v, c->myConeMPZ.getVerticesOfPolyhedron());
      return v;
    }

    std::vector<std::vector<BigInt> > GeneratorsOfToricRing(const cone& c)
    {
      std::vector<std::vector<BigInt> > v;
      c->myComputation(libnormaliz::ConeProperty::GeneratorsOfToricRing);
      if (c->mySmallConeIsGood)
        ConvertFromNormaliz(v, c->myConeLong.getGeneratorsOfToricRing());
      else
        ConvertFromNormaliz(v, c->myConeMPZ.getGeneratorsOfToricRing());
      return v;
    }


    // The following constraints depend all on ConeProperty::SupportHyperplanes
    std::vector<std::vector<BigInt> > SupportHyperplanes(const cone& c)
    { 
      std::vector<std::vector<BigInt> > v;
      c->myComputation(libnormaliz::ConeProperty::SupportHyperplanes);
      if (c->mySmallConeIsGood)
        ConvertFromNormaliz(v, c->myConeLong.getSupportHyperplanes());
      else
        ConvertFromNormaliz(v, c->myConeMPZ.getSupportHyperplanes());
      return v;
    }

    std::vector<std::vector<BigInt> > Equations(const cone& c)
    { 
      std::vector<std::vector<BigInt> > v;
      c->myComputation(libnormaliz::ConeProperty::SupportHyperplanes);
      if (c->mySmallConeIsGood)
        ConvertFromNormaliz(v, c->myConeLong.getEquations());
      else
        ConvertFromNormaliz(v, c->myConeMPZ.getEquations());
      return v;
    }

    std::vector<std::vector<BigInt> > Congruences(const cone& c)
    { 
      std::vector<std::vector<BigInt> > v;
      c->myComputation(libnormaliz::ConeProperty::SupportHyperplanes);
      if (c->mySmallConeIsGood)
        ConvertFromNormaliz(v, c->myConeLong.getCongruences());
      else
        ConvertFromNormaliz(v, c->myConeMPZ.getCongruences());
      return v;
    }

    // excluded faces are only available if they were explicitly given
    std::vector<std::vector<BigInt> > ExcludedFaces(const cone& c)
    {
      std::vector<std::vector<BigInt> > v;
      c->myComputation(libnormaliz::ConeProperty::ExcludedFaces);
      if (c->mySmallConeIsGood)
        ConvertFromNormaliz(v, c->myConeLong.getExcludedFaces());
      else
        ConvertFromNormaliz(v, c->myConeMPZ.getExcludedFaces());
      return v;
    }



/* REPLACED by the better function HilbertSeries
    std::vector<BigInt> HVector(const cone& c)
    { 
      c->myComputation(libnormaliz::ConeProperty::HilbertSeries);
      if (c->mySmallConeIsGood)
        return MPZ_classVToBigIntV(c->myConeLong.getHilbertSeries().getNum());
      else
        return MPZ_classVToBigIntV(c->myConeMPZ.getHilbertSeries().getNum());
    }
*/
    HPSeries HilbertSeries(const cone& c)
    { 
      c->myComputation(libnormaliz::ConeProperty::HilbertSeries);
      const libnormaliz::HilbertSeries& HS = c->mySmallConeIsGood ? c->myConeLong.getHilbertSeries() : c->myConeMPZ.getHilbertSeries();
      return HPSeries(MPZ_classVToBigIntV(HS.getNum()), 
                      libnormaliz::to_vector(HS.getDenom()));
    }

    RingElem HilbertPoly(const cone& c)
    { 
      c->myComputation(libnormaliz::ConeProperty::HilbertSeries);
      const libnormaliz::HilbertSeries& HS = c->mySmallConeIsGood ? c->myConeLong.getHilbertSeries() : c->myConeMPZ.getHilbertSeries();
      if (HS.getPeriod() != 1)
      {
        CoCoA_ERROR("Hilbert function is a quasi-polynomial of periode > 1. This function works for regular polynomials only.", "HilbertPoly");
      }
      vector<BigRat> coeffs = MPZ_classVToBigRatV( HS.getHilbertQuasiPolynomial()[0],
                                                   BigInt(HS.getHilbertQuasiPolynomialDenom().get_mpz_t()));
      PolyRing QQt = AsPolyRing(RingQQt(1));
      const RingElem t = indet(QQt,0);
      RingElem tpower = one(QQt);
      RingElem hpoly(QQt);
      for (long i=0; i<len(coeffs); ++i)
      {
          hpoly += coeffs[i] * tpower;   // NB tpower = t^i
          tpower *= t;
      }

      return hpoly;
    }

    BigRat multiplicity(const cone& c)
    {
      c->myComputation(libnormaliz::ConeProperty::Multiplicity);
      if (c->mySmallConeIsGood)
        return BigRat(c->myConeLong.getMultiplicity().get_mpq_t());
      else
        return BigRat(c->myConeMPZ.getMultiplicity().get_mpq_t());
    }



    bool IsPointed(const cone& c)
    {
      c->myComputation(libnormaliz::ConeProperty::ExtremeRays);
      if (c->mySmallConeIsGood)
        return c->myConeLong.isPointed();
      else
        return c->myConeLong.isPointed();
    }

    bool IsInhomogeneous(const cone& c)
    {
      // is always known
      if (c->mySmallConeIsGood)
        return c->myConeLong.isInhomogeneous();
      else
        return c->myConeLong.isInhomogeneous();
    }

    bool IsIntegrallyClosed(const cone& c)
    {
      c->myComputation(libnormaliz::ConeProperty::HilbertBasis);
      if (c->mySmallConeIsGood)
        return c->myConeLong.isIntegrallyClosed();
      else
        return c->myConeLong.isIntegrallyClosed();
    }

    bool IsDeg1HilbertBasis(const cone& c)
    {
      c->myComputation(libnormaliz::ConeProperty::HilbertBasis, libnormaliz::ConeProperty::Grading);
      if (c->mySmallConeIsGood)
        return c->myConeLong.isDeg1HilbertBasis();
      else
        return c->myConeLong.isDeg1HilbertBasis();
    }

   

    // dimension and rank invariants
    long EmbeddingDim(const cone& c)
    {
      // is always known
      if (c->mySmallConeIsGood)
        return c->myConeLong.getEmbeddingDim();
      else
        return c->myConeLong.getEmbeddingDim();
    }

    long rank(const cone& c)
    {
      c->myComputation(libnormaliz::ConeProperty::ExtremeRays);
      if (c->mySmallConeIsGood)
        return c->myConeLong.getRank();
      else
        return c->myConeMPZ.getRank();
    }

    std::vector<BigRat> grading(const cone& c)
    {
      c->myComputation(libnormaliz::ConeProperty::Grading);
      if (c->mySmallConeIsGood)
      {
        return LongVToBigRatV(c->myConeLong.getGrading(), BigInt(c->myConeLong.getGradingDenom()));
      }
      else
      {
        return MPZ_classVToBigRatV(c->myConeMPZ.getGrading(), BigInt(c->myConeMPZ.getGradingDenom().get_mpz_t()));
      }
    }

    // only for inhomogeneous case:
    long RecessionRank(const cone& c)
    {
      c->myComputation(libnormaliz::ConeProperty::RecessionRank);
      if (c->mySmallConeIsGood)
        return c->myConeLong.getRecessionRank();
      else
        return c->myConeMPZ.getRecessionRank();
    }
    long AffineDim(const cone& c)
    {
      c->myComputation(libnormaliz::ConeProperty::ExtremeRays);
      if (c->mySmallConeIsGood)
        return c->myConeLong.getRank();
      else
        return c->myConeMPZ.getRank();
    }
    long ModuleRank(const cone& c)
    {
      c->myComputation(libnormaliz::ConeProperty::ModuleRank);
      if (c->mySmallConeIsGood)
        return c->myConeLong.getModuleRank();
      else
        return c->myConeMPZ.getModuleRank();
    }

    BigInt shift(const cone& c)
    {
      c->myComputation(libnormaliz::ConeProperty::Shift);
      if (c->mySmallConeIsGood)
        return BigInt(c->myConeLong.getShift());
      else
        return BigInt(c->myConeMPZ.getShift().get_mpz_t());
    }
 
    std::vector<BigInt> dehomogenization(const cone& c)
    {
      c->myComputation(libnormaliz::ConeProperty::Dehomogenization);
      if (c->mySmallConeIsGood)
        return LongVToBigIntV(c->myConeLong.getDehomogenization());
      else
        return MPZ_classVToBigIntV(c->myConeMPZ.getDehomogenization());
    }

    // not so important in this library
    // size_t getTriangulationSize() const;
    // Integer getTriangulationDetSum() const;


/**************  applications to monomials  **************/

    namespace // anonymous for file local fns
    {
      PPVector HilbertBasis_PPVector(const PPVector& ppv, libnormaliz::InputType t)
      {
        cone c (PPVectorToVecVecBigInt(ppv), t);
        std::vector<std::vector<BigInt> > res = HilbertBasis(c);
        //create an emptyPPVector with the same PPMonoid
        PPVector ppv_res(PPM(ppv), DMR(ppv));
        VecVecBigIntToPPVector(ppv_res,res);
        return ppv_res;
      }
/*      PPVector HilbertBasis_PPVector(const std::vector<std::vector<BigInt> >& Mat, libnormaliz::InputType t, ring R)
      {
        SparsePolyRing S(AsSparsePolyRing(R));
        cone c (Mat, t);
        std::vector<std::vector<BigInt> > res = HilbertBasis(c);
        //create an emptyPPVector with the same PPMonoid
        PPVector ppv_res(S.myPPM());
        VecVecBigIntToPPVector(ppv_res,res);
        return ppv_res;
      } */
    }

    PPVector NormalToricRing(const PPVector& ppv)
    { 
      return HilbertBasis_PPVector(ppv, libnormaliz::Type::normalization);
    }

    PPVector IntClosureToricRing(const PPVector& ppv)
    {
      return HilbertBasis_PPVector(ppv, libnormaliz::Type::integral_closure);
    }

    PPVector IntClosureMonIdeal(const PPVector& ppv)
    {
      cone c (PPVectorToVecVecBigInt(ppv), libnormaliz::Type::rees_algebra);
      vector<vector<BigInt> > res = HilbertBasis(c);
      vector<vector<BigInt> > res1;
      for (vector<vector<BigInt> >::const_iterator it=res.begin(); it!=res.end(); ++it) {
        //BigInt tmp = it->back();
        if (it->back() == 1 ) {
          res1.push_back(*it);
          res1.back().pop_back();
        }
      }

      //create an emptyPPVector with the same PPMonoid
      PPVector ppv_res1(PPM(ppv), DMR(ppv));
      VecVecBigIntToPPVector(ppv_res1,res1);
      return ppv_res1;

    }

    //if you want also the whole normalization of the rees algebra you have to specify an extra symbol
    //it may be done by NewSymbols(1), which creates an unique new symbol (with funny name)
/*    pair<PPVector,PPVector> IntClosureMonIdeal(const PPVector& ppv, symbol var)
    { 
      //check if the symbol is in the ring
      cone c (PPVectorToVecVecBigInt(ppv), libnormaliz::Type::rees_algebra);
      vector<vector<BigInt> > res = HilbertBasis(c);
      vector<vector<BigInt> > res1;
      for (vector<vector<BigInt> >::const_iterator it=res.begin(); it!=res.end(); ++it) {
        BigInt tmp = it->back();
        if (it->back() == 1 ) {
          res1.push_back(*it);
          res1.back().pop_back();
        }
      }
      //create an emptyPPVector with the same PPMonoid
      PPVector ppv_res1(PPM(ppv), DMR(ppv));
      VecVecBigIntToPPVector(ppv_res1,res1);
      // create new PPM with extra symbol
      PPVector ppv_res(PPM(ppv), DMR(ppv));
      VecVecBigIntToPPVector(ppv_res,res);

      return make_pair(ppv_res1, ppv_res);
    }
*/     

/*    PPVector EhrhartRing(const PPVector& ppv)  // additional symbol?
    {
      // will give runtime error because ring has not enough variables
      return HilbertBasis_PPVector(ppv, libnormaliz::Type::polytope);
    }
*/

/**************  torus invariants and valuation rings  **************/
/*
    PPVector torusInvariants (Matrix T, Ring R)
    {
      if (NumIndets(R)!=NumCols(T)) 
      {
        CoCoA_Error("torusInvariants: wrong number of columns in matrix", "torusInvariants");
      }

      return HilbertBasis_PPVector(T, equations, R);
    }

// M2 code, to be converted
finiteDiagInvariants = method(Options=>true);
finiteDiagInvariants (Matrix,Ring) := opts>>o -> (M,R) ->
(
  if(NumIndets R != NumCols M-1)
  then error("finiteDiagInvariants: wrong number of columns in matrix");

  cone:= normaliz(allComputations=>o.allComputations, grading=>o.grading,M,6);
  rt := cone#"gen";
  if(instance(rt,Nothing,rt)) then return createEmptyMonomialSubalgebra R;

  S := createMonomialSubalgebra intmat2mons(rt,R);
  S.cache#"cone"=cone;
  return S;
);

diagInvariants = method(Options=>true);
diagInvariants (Matrix,Matrix,Ring) := opts>>o -> (T,F,R) ->
(
  if(NumIndets R != NumCols T or NumIndets R != NumCols F -1)
  then error("diagInvariants: wrong number of columns in matrix");

  cone:=(normaliz(allComputations=>o.allComputations, grading=>o.grading,{(T,5),(F,6)}));
  rt := cone#"gen";

  if(instance(rt,Nothing)) then return createEmptyMonomialSubalgebra R;

  S := createMonomialSubalgebra intmat2mons(rt,R);
  S.cache#"cone"=cone;
  return S;
);

intersectionValRings = method(Options=>true)
intersectionValRings (Matrix,Ring) := opts>>o -> (V,R) ->
(
    if(NumIndets(R)!=NumCols(V))
    then(
          error("intersectionValRings: wrong number of columns in matrix");
    );

    I:=id_(ZZ^(NumCols(V))); -- identity matrix
    V1:=I||V;

    M:=runNormaliz(allComputations=>o.allComputations, grading=>o.grading, V1,4);


    if(not nmzGen) then return; -- M=null

    vr:=createMonomialSubalgebra intmat2mons(M#"gen",R);
    vr.cache#"cone"=M;
    return vr;
)


intersectionValRingIdeals = method(Options=>true)
intersectionValRingIdeals (Matrix,Ring) := opts>>o -> (V,R) ->
(
    nc:=NumCols(V);
    if(NumIndets(R)!=nc-1)
    then(
         error("intersectionValRingIdeals: wrong number of columns in matrix");
    );

    I:=id_(ZZ^nc); -- identity matrix
    V1:=I||V;
    V1=mutableMatrix(V1);

    for i from 0 to numRows(V)-1
    do(
       V1_(i+nc,nc-1)=-V1_(i+nc,nc-1);
    );
    V1=matrix(V1);

    nmzCone:=runNormaliz(allComputations=>o.allComputations, grading=>o.grading, V1,4);
    if(not nmzGen) then return; -- nmzCone=null
    M:=nmzCone#"gen";

    R1:=createMonomialSubalgebra intmat2mons(M,R,0,NumCols M-1);
    R1.cache#"cone"=nmzCone; -- ????? TODO
    R2:=intmat2mons(M,R,1,NumCols M-1);

    return new HashTable from {"subalgebra"=>R1, "module generators"=>R2};
)
*/

/********************************************************
 ***               conversion functions               ***
 ********************************************************/

    namespace  // conversion functions
    {
      std::vector<BigInt> LongVToBigIntV(const std::vector<long>& VIn)
      {
        std::vector<BigInt> v;
        v.reserve(VIn.size());
        for (vector<long>::const_iterator it=VIn.begin(); it!=VIn.end(); ++it)
          v.push_back(BigInt(*it));
        return v;
      }

      std::vector<BigInt> MPZ_classVToBigIntV(const std::vector<mpz_class>& VIn)
      {
        std::vector<BigInt> v;
        v.reserve(VIn.size());
        for (vector<mpz_class>::const_iterator it=VIn.begin(); it!=VIn.end(); ++it)
          v.push_back(BigInt(it->get_mpz_t()));
        return v;
      }

      std::vector<BigRat> LongVToBigRatV(const std::vector<long>& VIn, const BigInt& denom)
      {
        std::vector<BigRat> v;
        v.reserve(VIn.size());
        for (vector<long>::const_iterator it=VIn.begin(); it!=VIn.end(); ++it)
          v.push_back(BigRat(BigInt(*it),denom));
        return v;
      }

      std::vector<BigRat> MPZ_classVToBigRatV(const std::vector<mpz_class>& VIn, const BigInt& denom)
      {
        std::vector<BigRat> v;
        v.reserve(VIn.size());
        for (vector<mpz_class>::const_iterator it=VIn.begin(); it!=VIn.end(); ++it)
          v.push_back(BigRat(BigInt(it->get_mpz_t()),denom));
        return v;
      }


      std::vector<long> BigIntVToLongV(const std::vector<BigInt>& VIn)
      {
        std::vector<long> v;
        long n;
        v.reserve(VIn.size());
        for (vector<BigInt>::const_iterator it=VIn.begin(); it!=VIn.end(); ++it)
          if (IsConvertible(n, *it))  v.push_back(n);
          else CoCoA_ERROR(ERR::BadConvert, "normaliz: BigIntVToLongV");
        return v;
      }


      std::vector<mpz_class> BigIntVToMPZ_classV(const std::vector<BigInt>& VIn)
      {
        std::vector<mpz_class> v;
        v.reserve(VIn.size());
        for (vector<BigInt>::const_iterator it=VIn.begin(); it!=VIn.end(); ++it)
          v.push_back(mpz_class(mpzref(*it)));
        return v;
      }


      void ConvertFromNormaliz(std::vector<std::vector<BigInt> >& v, const std::vector<std::vector<long> >& l)
      {
        v.clear();      // not exception safe
        v.reserve(l.size());
        //      transform(l.begin(), l.end(), v.begin(), LongVToBigIntV);
        for (vector<vector<long> >::const_iterator it=l.begin(); it!=l.end(); ++it)
          v.push_back(LongVToBigIntV(*it));
      }    


      void ConvertFromNormaliz(std::vector<std::vector<BigInt> >& v, const std::vector<std::vector<mpz_class> >& l)
      {
        v.clear();      // not exception safe
        v.reserve(l.size());
        //      transform(l.begin(), l.end(), v.begin(), LongVToBigIntV);
        for (vector<vector<mpz_class> >::const_iterator it=l.begin(); it!=l.end(); ++it)
          v.push_back(MPZ_classVToBigIntV(*it));
      }    


      std::vector<std::vector<long> >  ReturnVecVecLong(const std::vector<std::vector<BigInt> >& v)
      {
        std::vector<std::vector<long> > l;
        //      transform(v.begin(), v.end(), l.begin(), BigIntVToLongV);
        try
        {
            for (vector<vector<BigInt> >::const_iterator it=v.begin(); it!=v.end(); ++it)
              l.push_back(BigIntVToLongV(*it));
            //      std::clog << l << std::endl;
        } catch (...) {l.clear();}
        return l;
      }


      map< libnormaliz::InputType, vector<vector<long> > >  ReturnMapVecVecLong(const map< libnormaliz::InputType, vector<vector<BigInt> > > m)
      {
        map< libnormaliz::InputType, vector<vector<long> > > ret;
        for (map<libnormaliz::InputType, vector<vector<BigInt> > >::const_iterator it=m.begin(); it!=m.end(); ++it)
          ret.insert(make_pair((*it).first, ReturnVecVecLong((*it).second)));
        return ret;
      }
      
      
      std::vector<std::vector<mpz_class> >  ReturnVecVecMPZ_class(const std::vector<std::vector<BigInt> >& v)
      {
        std::vector<std::vector<mpz_class> > l;
        //      transform(v.begin(), v.end(), l.begin(), BigIntVToLongV);
        for (vector<vector<BigInt> >::const_iterator it=v.begin(); it!=v.end(); ++it)
          l.push_back(BigIntVToMPZ_classV(*it));
        //      std::clog << l << std::endl;
        return l;
      }


      map< libnormaliz::InputType, vector<vector<mpz_class> > >  ReturnMapVecVecMPZ_class(const map< libnormaliz::InputType, vector<vector<BigInt> > > m)
      {
        map< libnormaliz::InputType, vector<vector<mpz_class> > > ret;
        for (map<libnormaliz::InputType, vector<vector<BigInt> > >::const_iterator it=m.begin(); it!=m.end(); ++it)
          ret.insert(make_pair((*it).first, ReturnVecVecMPZ_class((*it).second)));
        return ret;
      }


      std::vector<std::vector<BigInt> > PPVectorToVecVecBigInt(const PPVector& ppv)
      {
        vector<BigInt> tmp;
        const long n =  len(ppv);
        vector<vector<BigInt> > v(n);
        for (long i=0; i<n; ++i)
        {
          BigExponents(tmp,PP(ppv[i]));
          v[i]=tmp;
        }
        return v;
      }

      void VecVecBigIntToPPVector(PPVector& ppv, const std::vector<std::vector<BigInt> >& M)
      {
        PPMonoid ppm = PPM(ppv);
        const long n =  len(M);
        for (long i=0; i < n; ++i)
        {
          ppv.myPushBack(PPMonoidElem(ppm, M[i]));
        }
      }

    } // end of anonymous namespace

    std::vector<std::vector<BigInt> > MatrixToVecVecBigInt(ConstMatrixView M)
    {
      const ErrorInfo ErrMesg("Matrix entries must be integer", "MatrixToVecVecBigInt (cone ctor)");

      vector<vector<BigInt> > v;
      for (long i=0; i<NumRows(M); ++i)
      {
        v.push_back(vector<BigInt>());
        for (long j=0; j<NumCols(M); ++j)
          v[i].push_back(ConvertTo<BigInt>(M(i,j), ErrMesg));
      }
      return v;
    }





  } // namespace Normaliz
} // namespace CoCoA

// RCS header/log in the next few lines
// $Header: /Volumes/Home_1/cocoa/cvs-repository/CoCoALib-0.99/src/AlgebraicCore/ExternalLibs-Normaliz.C,v 1.23 2014/05/12 14:33:51 bigatti Exp $
// $Log: ExternalLibs-Normaliz.C,v $
// Revision 1.23  2014/05/12 14:33:51  bigatti
// -- updated from Christof Soeger
//
// Revision 1.22  2014/05/09 14:56:30  bigatti
// -- new fn by Christof Soeger
//
// Revision 1.21  2014/01/30 09:56:49  bigatti
// -- removed DOS newlines
//
// Revision 1.20  2014/01/29 17:38:35  bigatti
// -- added HilbertSeries (by Christof Soeger)
// -- added multiplicity (by Christof Soeger)
//
// Revision 1.19  2014/01/28 09:44:42  abbott
// Tidier impl of MatrixToVecVecBigInt using new ConvertTo syntax.
//
// Revision 1.18  2013/07/12 14:53:32  abbott
// Improved indentation.
//
// Revision 1.17  2013/03/15 17:49:17  abbott
// Minor cleaning; removed an out-of-date TODO comment.
//
// Revision 1.16  2012/10/08 13:52:02  bigatti
// -- more cleaning and updates by Christof Soeger
//
// Revision 1.15  2012/10/05 10:17:04  bigatti
// by Christof Soeger
// * Made the NewCone functions to constructors of cone.
// * Introduced some abbreviation in the method names:
//  IntegralClosure -> IntClosure
//  MonomialIdeal -> MonIdeal
//  Normaliz -> Nmz  (the prefix for CoCoA5 functions)
// * New function IntClosureMonIdeal
//
// Revision 1.14  2012/09/28 13:56:36  abbott
// Cleaned up code with Christof's help -- now cleaner & more symmetrical.
// Also fixed a subtle bug where mySmallConeIsGood was not updated as it should have been.
//
// Revision 1.13  2012/08/03 16:31:22  bigatti
// -- changed: procedural --> functional (by C.Soeger)
//
// Revision 1.12  2012/07/25 12:46:27  bigatti
// -- merged with C.Soeger changes for NormalizComputation (init from map, etc)
//
// Revision 1.11  2012/07/19 17:12:05  abbott
// Added NewCone -- unified pseudo-ctor so user does not have to choose between long and BigInt.
//
// Revision 1.10  2012/06/19 14:44:41  bigatti
// -- changed Ht1 --> Deg1, changed def of HVector (by C.Soeger)
//
// Revision 1.9  2011/11/07 11:09:32  bigatti
// -- new ctors taking ConstRefRingElem
// -- added PPVector NormalToricRing(const PPVector& ppv)
// -- removed void HilbertBasis(std::vector<std::vector<BigInt> >& v, const cone& c)
//
// Revision 1.8  2011/10/04 13:03:02  bigatti
// -- new logo for gui
//
// Revision 1.7  2011/09/30 12:55:44  bigatti
// -- introduced namespace "Normaliz" and removed Normaliz from function names
// -- input of Normaliz functions in CoCoA-5 is now a matrix instead of
//    vector<vector<BigInt>>
//
// Revision 1.6  2011/08/14 15:52:17  abbott
// Changed ZZ into BigInt (phase 1: just the library sources).
//
// Revision 1.5  2011/07/20 15:31:21  bigatti
// -- fixed InputType in (undefined) functions
//
// Revision 1.4  2011/07/20 13:49:37  bigatti
// -- added "Normaliz" postfix to Normaliz function calls
//
// Revision 1.3  2011/07/20 12:45:12  bigatti
// -- new normaliz interface (not yet public)
//
// Revision 1.2  2011/02/17 16:50:04  bigatti
// -- getting ready for new official veson on Normaliz: added HVector, removed Triangulation
//
// Revision 1.1  2010/11/05 14:27:18  bigatti
// -- was TmpNormaliz**
//
// Revision 1.6  2010/10/12 11:22:54  bigatti
// -- TmpNormaliz.H simplified:
//    now cone is a smart pointer to ConeBase
//    and the concrete classes are entirely in the .C file
// -- added Ht1Elements, SupportHyperplanes, Triangulation
// -- added some text in ex-Normaliz1 and 2
//
// Revision 1.5  2010/10/08 13:40:37  bigatti
// -- moved inclusion of libnormaliz.cpp and instantiation of types
//    into TmpNormalizTypes.C
//
// Revision 1.4  2010/10/08 10:39:32  bigatti
// -- extended interface for normaliz
//

#endif
