//   Copyright (c) 2010 Giovanni Lagorio (lagorio@disi.unige.it)
//
//   This file is part of the source of CoCoALib, the CoCoA Library.
//
//   CoCoALib is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.
//
//   CoCoALib is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with CoCoALib.  If not, see <http://www.gnu.org/licenses/>.

#include "CoCoA/IntOperations.H"
#include "CoCoA/MatrixOperations.H"
#include "CoCoA/QuotientRing.H"
#include "CoCoA/RingQQ.H"
#include "CoCoA/RingZZ.H"
#include "CoCoA/SparsePolyRing.H"
#include "CoCoA/convert.H"
#include "CoCoA/error.H"


#include <limits>
#include <iostream>
#include <exception>
#include <cassert>
#include <errno.h>
#include <set>
#include <boost/scope_exit.hpp>
#include <boost/foreach.hpp>
#include <boost/algorithm/string/case_conv.hpp>

//#include "CoCoA/library.H" // already included by AST->Parser->Interpreter
#include "Interpreter.H"
#include "BuiltInFunctions.H"
#include "CoCoALibSupplement.H"
#include "OnlineHelp.H"

namespace CoCoA {

namespace InterpreterNS {

using namespace std;
using namespace boost;
using namespace CoCoA::AST;
using namespace CoCoA::LexerNS;
using namespace CoCoA::ParserNS;

intrusive_ptr<BoolValue> BoolValue::trueValue(new BoolValue(true)), BoolValue::falseValue(new BoolValue(false));
intrusive_ptr<BigIntValue> BigIntValue::zero(new BigIntValue(BigInt(0))), BigIntValue::one(new BigIntValue(BigInt(1))), BigIntValue::minusOne(new BigIntValue(BigInt(-1)));
intrusive_ptr<StringValue> StringValue::empty(new StringValue(""));

namespace
{
	ring ringForList(const intrusive_ptr<const ListValue> list) {
		bool foundBigRat = false;
		const ListValue::ContainerType::size_type size = list->size();
		for(ListValue::ContainerType::size_type a=0; a<size; ++a) {
			const intrusive_ptr<RightValue> e = list->getValue(a);
			if (const intrusive_ptr<RingElemValue> ringElem = dynamic_pointer_cast<RingElemValue>(e))
				return owner(ringElem->theRingElem);
			else if (dynamic_pointer_cast<BigRatValue>(e))
				foundBigRat = true;
		}
		if (foundBigRat)
			return RingQQ();
		return RingZZ();
	}

	vector<RingElem> toRingElemList(const intrusive_ptr<const ListValue> list, const intrusive_ptr<const Expression> exp, const ring &ring) {
		const ListValue::ContainerType::size_type size = list->size();
		vector<RingElem> result;
		for(ListValue::ContainerType::size_type a=0; a<size; ++a) {
			const intrusive_ptr<RightValue> e = list->getValue(a);
			if (const intrusive_ptr<RingElemValue> ringElem = dynamic_pointer_cast<RingElemValue>(e))
				result.push_back(ringElem->theRingElem);
			else if (const intrusive_ptr<BigIntValue> N = dynamic_pointer_cast<BigIntValue>(e))
				result.push_back(RingElem(ring, N->theInteger));
			else if (const intrusive_ptr<BigRatValue> qq = dynamic_pointer_cast<BigRatValue>(e))
				result.push_back(RingElem(ring, qq->theRational));
			else
				throw RuntimeException("A list containing RINGELEM, INT or RAT values is required", exp);
		}
		return result;
	}
}

vector<RingElem> RuntimeEnvironment::evalArgAsRingElemList(const Argument &arg, const ring &ring) {
	return toRingElemList(this->evalArgAs<ListValue>(arg), arg.exp, ring);
}

vector<RingElem> RuntimeEnvironment::evalArgAsRingElemList(const Argument &arg) {  
	const intrusive_ptr<const ListValue> list = this->evalArgAs<ListValue>(arg);
	return toRingElemList(list, arg.exp, ringForList(list));
}

intrusive_ptr<Value> LeftValue::indexedByBigInt(intrusive_ptr<BigIntValue> index, const CharPointer &targetExpBegin, const CharPointer &targetExpEnd, const CharPointer &indexExpBegin, const CharPointer &indexExpEnd) {
	return new IntegerIndexedAccess(this, index->theInteger, targetExpBegin, targetExpEnd, indexExpBegin, indexExpEnd);
}

intrusive_ptr<Value> LeftValue::indexedByString(intrusive_ptr<StringValue> index, const CharPointer &targetExpBegin, const CharPointer &targetExpEnd, const CharPointer &indexExpBegin, const CharPointer &indexExpEnd) {
	return new StringIndexedAccess(this, index->theString, targetExpBegin, targetExpEnd, indexExpBegin, indexExpEnd);
}

intrusive_ptr<Value> RightValue::indexedByBigInt(intrusive_ptr<BigIntValue> /* index */, const CharPointer &targetExpBegin, const CharPointer &targetExpEnd, const CharPointer & /* indexExpBegin */, const CharPointer & /* indexExpEnd */) {
	throw NonIntegerIndexableException(this, targetExpBegin, targetExpEnd);
}

intrusive_ptr<Value> RightValue::indexedByString(intrusive_ptr<StringValue> /* index */, const CharPointer &targetExpBegin, const CharPointer &targetExpEnd, const CharPointer & /* indexExpBegin */, const CharPointer & /* indexExpEnd */) {
	throw NonStringIndexableException(this, targetExpBegin, targetExpEnd);
}

intrusive_ptr<Value> RecordValue::indexedByString(intrusive_ptr<StringValue> str, const CharPointer & /* targetExpBegin */, const CharPointer & /* targetExpEnd */, const CharPointer &indexExpBegin, const CharPointer &indexExpEnd) {
	const intrusive_ptr<Value> v = this->getField(str->theString);
	if (!v)
		throw FieldNotFoundException(str->theString, indexExpBegin, indexExpEnd, this);
	return v;
}

intrusive_ptr<Value> ListValue::indexedByBigInt(intrusive_ptr<BigIntValue> N, const CharPointer & /* targetExpBegin */, const CharPointer & /* targetExpEnd */, const CharPointer &indexExpBegin, const CharPointer &indexExpEnd) {
	long l;
	if (!IsConvertible(l, N->theInteger) || l<=0 || static_cast<ContainerType::size_type>(l)>this->size())
		throw IndexOutOfRangeException(N->theInteger, this->size(), indexExpBegin, indexExpEnd);
	return this->getValue(l-1);
}

intrusive_ptr<Value> StringValue::indexedByBigInt(intrusive_ptr<BigIntValue> N, const CharPointer & /* targetExpBegin */, const CharPointer & /* targetExpEnd */, const CharPointer &indexExpBegin, const CharPointer &indexExpEnd) {
	long l;
	if (!IsConvertible(l, N->theInteger) || l<=0 || static_cast<string::size_type>(l)>this->theString.length())
		throw IndexOutOfRangeException(N->theInteger, this->theString.length(), indexExpBegin, indexExpEnd);
	return new StringValue(string(1, this->theString[l-1]));
}

intrusive_ptr<Value> IntMapValue::indexedByBigInt(intrusive_ptr<BigIntValue> N, const CharPointer & /* targetExpBegin */, const CharPointer & /* targetExpEnd */, const CharPointer &indexExpBegin, const CharPointer &indexExpEnd) {
	long l;
	if (!IsConvertible(l, N->theInteger))
		throw RuntimeException("Index is not a machine-integer", indexExpBegin, indexExpEnd);
	if (const intrusive_ptr<RightValue> result = this->getValue(l))
		return result;
	throw RuntimeException("Invalid index", indexExpBegin, indexExpEnd);
}

intrusive_ptr<Value> MatrixValue::indexedByBigInt(intrusive_ptr<BigIntValue> N, const CharPointer & /* targetExpBegin */, const CharPointer & /* targetExpEnd */, const CharPointer &indexExpBegin, const CharPointer &indexExpEnd) {
	long l;
	if (!IsConvertible(l, N->theInteger))
		throw RuntimeException("Index is not a machine-integer", indexExpBegin, indexExpEnd);
	const size_t row = static_cast<size_t>(l);
	if (l>=1 && row<=this->numRows()) // note: cannot use row>=1 because size_t is unsigned
		return new MatrixRowValue(this, row-1);
	throw RuntimeException("Invalid row index", indexExpBegin, indexExpEnd);
}

intrusive_ptr<Value> MatrixRowValue::indexedByBigInt(intrusive_ptr<BigIntValue> N, const CharPointer & /* targetExpBegin */, const CharPointer & /* targetExpEnd */, const CharPointer &indexExpBegin, const CharPointer &indexExpEnd) {
	long l;
	if (!IsConvertible(l, N->theInteger))
		throw RuntimeException("Index is not a machine-integer", indexExpBegin, indexExpEnd);
	const size_t col = static_cast<size_t>(l);
	if (l>=1 && col<=this->matrix->numColumns()) // note: cannot use col>=1 because size_t is unsigned
		return new RingElemValue(this->matrix->theMatrix(this->nRow, col-1));
	throw RuntimeException("Invalid column index", indexExpBegin, indexExpEnd);
}

intrusive_ptr<Value> ModuleElemValue::indexedByBigInt(intrusive_ptr<BigIntValue> N, const CharPointer & /* targetExpBegin */, const CharPointer & /* targetExpEnd */, const CharPointer &indexExpBegin, const CharPointer &indexExpEnd) {
	long l;
  long nc = NumCompts(this->theModuleElem);
	if (!IsConvertible(l, N->theInteger) || l<=0 || l>nc)
		throw IndexOutOfRangeException(N->theInteger, nc, indexExpBegin, indexExpEnd);
	return Value::from((this->theModuleElem)[l-1]);
}

IntegerIndexedAccess::IntegerIndexedAccess(const intrusive_ptr<LeftValue> targetLV, const BigInt &index,  const CharPointer &targetExpBegin, const CharPointer &targetExpEnd, const CharPointer &indexExpBegin, const CharPointer &indexExpEnd) :
	IndexedAccess(targetLV, targetExpBegin, targetExpEnd, indexExpBegin, indexExpEnd),
	index(index)
{
	intrusive_ptr<RightValue> target = this->targetLV->asRightValue();
	if (!target->canBeIndexedByBigInt())
		throw NonIntegerIndexableException(target, targetExpBegin, targetExpEnd);
}

SnapshotFrame::SnapshotFrame(const Frame &frame) :
	invocationExp(frame.invocationExp),
	block(frame.block)
{
	assert(this->invocationExp);
	assert(this->block);
}

RuntimeException::RuntimeException(const string &reason, const CharPointer &from, const CharPointer &to, const Frame * const) :
	ExceptionWithSourcePosition(reason, from, to, false)
{}

RuntimeException::RuntimeException(const string &reason, intrusive_ptr<const ParsedObject> po, const Frame * const) :
	ExceptionWithSourcePosition(reason, po->getBegin(), po->getEnd(), false)
{}

RuntimeException::RuntimeException(const string &reason, const Token &token, const Frame * const) :
	ExceptionWithSourcePosition(reason, token, false)
{}

PackageValue::PackageValue(const intrusive_ptr<const PackageStatement> pkgDecl) :
		pkgDecl(pkgDecl),
		pkgName(pkgDecl->name),
		prefix(pkgDecl->name+".")
{}

void RightValue::describe(intrusive_ptr<OutputStreamValue> out) const {
	out->print("A value of type ")->println(this->getType());
}

void BigIntValue::describe(intrusive_ptr<OutputStreamValue> out) const {
	if (this->theInteger==BigInt(42))
		out->println("The Answer to the Ultimate Question of Life, the Universe and Everything");
	else
		this->RightValue::describe(out);
}

void BuiltInFunction::describe(intrusive_ptr<OutputStreamValue> out) const {
	out->println("A built-in function");
}

void RingHomValue::describe(intrusive_ptr<OutputStreamValue> out) const {
	out->println(this);
}

void UserDefinedFunction::describe(intrusive_ptr<OutputStreamValue> out) const {
	out->println(this->fnDecl->getBegin().stringTo(this->fnDecl->getEnd()));
}

void PackageValue::describe(intrusive_ptr<OutputStreamValue> out) const
{
  const intrusive_ptr<const PackageStatement> pkgDecl = this->pkgDecl;
  const string &pkgName = pkgDecl->name;
  if (this->pkgDecl->memberNames.empty())
  {
    out->print("The package ")->print(pkgName)->println(" is empty.\n");
    return;
  }
  set<string> exports;
  if (pkgDecl->exportedNames.empty())
  { out->print("The package ")->print(pkgName)->print(" has no exported names.\n"); }
  else
  {
    out->print("The package ")->print(pkgName)->print(" exports the following names:\n");
    // Put names into a set so that they are printed in alphabetical order.
    BOOST_FOREACH(const Token &t, pkgDecl->exportedNames)
      exports.insert(t.lexeme());
    BOOST_FOREACH(const string &s, exports)
      out->print("* ")->println(s);
  }
  if (pkgDecl->memberNames.size() > exports.size())
  {
    out->print("\nThe package ")->print(pkgName)->print(" also has the following non-exported members:\n");
    BOOST_FOREACH(const string &s, pkgDecl->memberNames)
      if (exports.find(s) == exports.end()) // print s only if it not an export
        out->print("* ")->print(pkgName)->print(".")->println(s);
  }
}

OutputFileStreamValue::OutputFileStreamValue(intrusive_ptr<StringValue> filename, intrusive_ptr<const Expression> sourceExp) :
	CppOutputStreamValue(stream, true, false),
	stream(filename->theString.c_str())
{
	if (!stream.is_open())
		throw RuntimeException("Cannot open file \""+filename->theString+"\" for writing", sourceExp);
}

intrusive_ptr<OutputStreamValue> RuntimeEnvironment::getOutputStream() {
	return this->standardOutput;
}

TaggedValue::TaggedValue(boost::intrusive_ptr<RightValue> value, const std::string &tag) :
	type(TypeValue::tagType(tag)),
	value(value),
	tag(tag)
{}

map<string, intrusive_ptr<TypeValue> > TypeValue::taggedTypes;

intrusive_ptr<TypeValue> TypeValue::tagType(const string &tag) {
	map<string, intrusive_ptr<TypeValue> >::const_iterator it = taggedTypes.find(tag);
	if (it!=taggedTypes.end())
		return it->second;
	intrusive_ptr<TypeValue> newTaggedType(new TypeValue("TAGGED(\""+tag+"\")", -1));
	taggedTypes.insert(make_pair(tag, newTaggedType));
	return newTaggedType;
}

ostream &TypeValue::dumpAsString(ostream &out) const {
	return out << this->name;
}


ostream &RingElemValue::dumpAsString(ostream &out) const {
	return out << this->theRingElem;
}

ostream &IdealValue::dumpAsString(ostream &out) const {
	return out << this->theIdeal;
}

ostream &ModuleValue::dumpAsString(ostream &out) const {
	return out << this->theModule;
}

ostream &ModuleElemValue::dumpAsString(ostream &out) const {
	return out << this->theModuleElem;
}

ostream &JBMillValue::dumpAsString(ostream &out) const {
	// return out << this->theJBMill;
	return out << "yeah this is a first test of the output of the JBMilll";
}

ostream &RingValue::dumpAsString(ostream &out) const {
	return out << this->theRing;
}

ostream &IntMapValue::dumpAsString(ostream &out) const {
	out << "{";
	bool first=true;
	for(MapType::const_iterator pos = this->map.begin(); pos!=this->map.end(); ++pos) {
		if (first)
			first = false;
		else
			out << ", ";
		out << pos->first << " -> " << pos->second;
		pos->second->dumpRefCountAsString(out);
	}
	return out << "}";
}

intrusive_ptr<TypeValue> BoolValue::type(new TypeValue("BOOL", TypeValue::DISPATCH_INDEX_OF_BOOL));
intrusive_ptr<TypeValue> FunctionValue::type(new TypeValue("FUNCTION", TypeValue::DISPATCH_INDEX_OF_FUNCTION));
intrusive_ptr<TypeValue> ListValue::type(new TypeValue("LIST", TypeValue::DISPATCH_INDEX_OF_LIST));
intrusive_ptr<TypeValue> BigIntValue::type(new TypeValue("INT", TypeValue::DISPATCH_INDEX_OF_BIGINT));
intrusive_ptr<TypeValue> BigRatValue::type(new TypeValue("RAT", TypeValue::DISPATCH_INDEX_OF_BIGRAT));
intrusive_ptr<TypeValue> RecordValue::type(new TypeValue("RECORD", TypeValue::DISPATCH_INDEX_OF_RECORD));
intrusive_ptr<TypeValue> TypeValue::type(new TypeValue("TYPE", TypeValue::DISPATCH_INDEX_OF_TYPE));
intrusive_ptr<TypeValue> StringValue::type(new TypeValue("STRING", TypeValue::DISPATCH_INDEX_OF_STRING));
intrusive_ptr<TypeValue> VoidValue::type(new TypeValue("VOID", TypeValue::DISPATCH_INDEX_OF_VOID));
intrusive_ptr<TypeValue> ErrorValue::type(new TypeValue("ERROR", TypeValue::DISPATCH_INDEX_OF_ERROR));
intrusive_ptr<TypeValue> OutputStreamValue::type(new TypeValue("OSTREAM", TypeValue::DISPATCH_INDEX_OF_OSTREAM));
intrusive_ptr<TypeValue> ZZModValue::type(new TypeValue("ZMOD", TypeValue::DISPATCH_INDEX_OF_ZMOD));
intrusive_ptr<TypeValue> RingElemValue::type(new TypeValue("RINGELEM", TypeValue::DISPATCH_INDEX_OF_RINGELEM));
intrusive_ptr<TypeValue> RatFunValue::type(new TypeValue("RATFUN", TypeValue::DISPATCH_INDEX_OF_RATFUN));
intrusive_ptr<TypeValue> IdealValue::type(new TypeValue("IDEAL", TypeValue::DISPATCH_INDEX_OF_IDEAL));
intrusive_ptr<TypeValue> ModuleValue::type(new TypeValue("MODULE", TypeValue::DISPATCH_INDEX_OF_MODULE));
intrusive_ptr<TypeValue> ModuleElemValue::type(new TypeValue("MODULEELEM", TypeValue::DISPATCH_INDEX_OF_MODULEELEM));
intrusive_ptr<TypeValue> MatrixValue::type(new TypeValue("MAT", TypeValue::DISPATCH_INDEX_OF_MAT));
intrusive_ptr<TypeValue> RingValue::type(new TypeValue("RING", TypeValue::DISPATCH_INDEX_OF_RING));
intrusive_ptr<TypeValue> PackageValue::type(new TypeValue("PACKAGE", TypeValue::DISPATCH_INDEX_OF_PACKAGE));
intrusive_ptr<TypeValue> IntMapValue::type(new TypeValue("INTMAP", TypeValue::DISPATCH_INDEX_OF_INTMAP));
intrusive_ptr<TypeValue> RingHomValue::type(new TypeValue("RINGHOM", TypeValue::DISPATCH_INDEX_OF_RINGHOM));
intrusive_ptr<TypeValue> MatrixRowValue::type(new TypeValue("MATRIXROW", TypeValue::DISPATCH_INDEX_OF_MATRIXROW));

intrusive_ptr<TypeValue> JBMillValue::type(new TypeValue("JBMILL", TypeValue::DISPATCH_INDEX_OF_JBMILL));

intrusive_ptr<TypeValue> TaggedValue::getType() const {
	return this->type;
}

intrusive_ptr<TypeValue> BoolValue::getType() const {
	return type;
}

intrusive_ptr<TypeValue> FunctionValue::getType() const {
	return type;
}

intrusive_ptr<TypeValue> ListValue::getType() const {
	return type;
}

intrusive_ptr<TypeValue> BigIntValue::getType() const {
	return type;
}

intrusive_ptr<TypeValue> BigRatValue::getType() const {
	return type;
}

intrusive_ptr<TypeValue> RecordValue::getType() const {
	return type;
}

intrusive_ptr<TypeValue> StringValue::getType() const {
	return type;
}

intrusive_ptr<TypeValue> VoidValue::getType() const {
	return type;
}

intrusive_ptr<TypeValue> ErrorValue::getType() const {
	return type;
}

intrusive_ptr<TypeValue> OutputStreamValue::getType() const {
	return type;
}

intrusive_ptr<TypeValue> TypeValue::getType() const {
	return type;
}

intrusive_ptr<TypeValue> PackageValue::getType() const {
	return type;
}

intrusive_ptr<TypeValue> RingElemValue::getType() const {
	return type;
}

intrusive_ptr<TypeValue> IdealValue::getType() const {
	return type;
}

intrusive_ptr<TypeValue> ModuleValue::getType() const {
	return type;
}

intrusive_ptr<TypeValue> ModuleElemValue::getType() const {
	return type;
}

intrusive_ptr<TypeValue> JBMillValue::getType() const {
	return type;
}

intrusive_ptr<TypeValue> RingValue::getType() const {
	return type;
}

intrusive_ptr<TypeValue> IntMapValue::getType() const {
	return type;
}

intrusive_ptr<TypeValue> RingHomValue::getType() const {
	return type;
}

intrusive_ptr<TypeValue> MatrixValue::getType() const {
	return type;
}

intrusive_ptr<TypeValue> MatrixRowValue::getType() const {
	return type;
}

vector<RingValue::SymbolPair> RingValue::allSymbolValues() {
	vector<SymbolPair> result;
	createAllSymbolValues(this->theRing, result);
	return result;
}

void RingValue::createAllSymbolValues(const ring &r, std::vector<SymbolPair> &allIndets) {
  vector<symbol> syms(symbols(r));
  BOOST_FOREACH(const symbol &s, syms)
    allIndets.push_back(make_pair(s, new RingElemValue(RingElem(r, s))));
}

void RingValue::removeInjectedIndeterminates(RuntimeEnvironment *runtimeEnv) {
	Frame * const tlFrame = runtimeEnv->getTopLevelFrame();
	BOOST_FOREACH(int slot, this->injectedSlots) {
		VariableSlot &vs = tlFrame->varSlots[slot];
		vs.value = 0;
		vs.unprotect();
	}
	this->injectedSlots.clear();
}

void RingValue::injectIndeterminates(RuntimeEnvironment *runtimeEnv, const intrusive_ptr<const RingDefinition> ringDefinition, const string &ringName) {
	assert(this->injectedSlots.empty());
	string protectedNames, msgPrefix, msgSuffix;
	Frame * const tlFrame = runtimeEnv->getTopLevelFrame();
	vector<SymbolPair> allIndets(this->allSymbolValues());
	BOOST_FOREACH(const SymbolPair &pair, allIndets) {
			const symbol &sym = pair.first;
			const string hd = head(sym);
			if (hd==ringName)
				runtimeEnv->interpreter->errorReporter->reportWarning("The name of an indeterminate corresponds to the name of the ring", ringDefinition->getBegin(), ringDefinition->getEnd(), WS_NORMAL);
			int slot = runtimeEnv->slotFor(hd);
			if (slot<0)
				continue;
			if (tlFrame->varSlots[slot].isProtected()) {
				if (protectedNames.length()==0) {
					protectedNames = hd;
					msgPrefix = " ";
					msgSuffix = " is ";
				} else {
					protectedNames += ", ";
					protectedNames += hd;
					msgPrefix = "s ";
					msgSuffix = " are ";
				}
			}
	}
	if (protectedNames.length())
		throw RuntimeException("Cannot use the ring because the name"+msgPrefix+protectedNames+msgSuffix+"protected", ringDefinition);
	int indetIndex=0;
	map<string, intrusive_ptr<IntMapValue> > head2map;
	BOOST_FOREACH(const SymbolPair &pair, allIndets) {
		const symbol &sym = pair.first;
		const int nSubscripts = NumSubscripts(sym);
		const intrusive_ptr<RingElemValue> indeterminate(pair.second);
		const string hd(head(sym));
		//cout << "sym=" << sym << ", hd=" << hd << ", nSubscripts = " << nSubscripts << ", indeterminate=" << indeterminate << endl;
		if (nSubscripts==0)
			this->injectedSlots.push_back(runtimeEnv->setTopLevelVar(hd, indeterminate, VariableSlot::VSF_SystemProtected));
		else {
			map<string, intrusive_ptr<IntMapValue> >::const_iterator it = head2map.find(hd);
			intrusive_ptr<IntMapValue> map;
			if (it==head2map.end()) {
				map = new IntMapValue();
				head2map.insert(make_pair(hd, map));
			} else
				map = it->second;
			this->injectedSlots.push_back(runtimeEnv->setTopLevelVar(hd, map, VariableSlot::VSF_SystemProtected));
			for(int a=0; a<(nSubscripts-1); ++a) {
				IntMapValue::KeyType i = subscript(sym, a);
				intrusive_ptr<IntMapValue> v = dynamic_pointer_cast<IntMapValue>(map->getValue(i));
				if (!v) {
					v = new IntMapValue();
					map->addValue(i, v);
				}
				map = v;
			}
			map->addValue(subscript(sym, nSubscripts-1), indeterminate);
		}
		++indetIndex;
	}
}

namespace {
	class FakeExpression : public Expression {
	private:
		const intrusive_ptr<RightValue> value;
		intrusive_ptr<Value> implEval(RuntimeEnvironment *) const {
			return this->value;
		}
		bool skipDebugging() const {
			return true;
		}
	public:
		explicit FakeExpression(const intrusive_ptr<RightValue> value, const CharPointer &beginSourcePosition, const CharPointer & endSourcePosition) :
			Expression(beginSourcePosition, endSourcePosition),
			value(value)
		{}
		bool isLeftValue() const { return false; }
		ostream &dumpAsString(ostream &out) const { assert(false); return out; }
		void accept(ParsedObjectVisitor *) { assert(false); }
	};

}

intrusive_ptr<RightValue> TaggedValue::unaryMinus(const CharPointer &opPosition, RuntimeEnvironment * const runtimeEnv) {
	runtimeEnv->interpreter->reportWarning("Performing this operation requires an implicit untagging; use untagged() to avoid this warning", opPosition, opPosition);
	return this->value->unaryMinus(opPosition, runtimeEnv);
}

intrusive_ptr<RightValue> TaggedValue::clone() {
	return new TaggedValue(this->value->clone(), this->tag);
}

ostream &TaggedValue::dumpAsString(ostream &out) const {
	out << "tagged(";
	this->value->dumpAsString(out);
	intrusive_ptr<StringValue> s = new StringValue(this->tag);
	return out << ", " << s << ")";
}

ostream &ErrorValue::dumpAsString(ostream &out) const {
	out << "error(";
	intrusive_ptr<StringValue> s = new StringValue(this->message);
	return out << s << ")";
}

ostream &OutputStreamValue::dumpAsString(ostream &out) const {
	return out << "<out-stream>";
}

intrusive_ptr<RightValue> OutputStreamValue::close(RuntimeEnvironment *, intrusive_ptr<const Expression> sourceExp) {
	if (!this->canBeClosed)
		throw RuntimeException("This stream cannot be closed", sourceExp);
	if (this->isClosed)
		throw RuntimeException("This stream is already closed", sourceExp);
	this->isClosed = true;
	// we don't actually close anything here, because that's a subclass responsibility
	return VoidValue::theInstance;
}

intrusive_ptr<RightValue> OutputStringStreamValue::close(RuntimeEnvironment *runtimeEnv, intrusive_ptr<const Expression> sourceExp) {
	OutputStreamValue::close(runtimeEnv, sourceExp);
	return new StringValue(this->ss.str());
}

intrusive_ptr<RightValue> OutputFileStreamValue::close(RuntimeEnvironment *runtimeEnv, intrusive_ptr<const Expression> sourceExp) {
	OutputStreamValue::close(runtimeEnv, sourceExp);
	this->stream.close();
	return VoidValue::theInstance;
}

namespace {
	string stripPkgName(const string &s) {
		string::size_type dotPos = s.find('.', 0);
		assert(dotPos!=string::npos);
		return s.substr(dotPos+1);
	}

	string getPkgNameWithDot(const string &s) {
		static const string::size_type l = ParserNS::Parser::TopLevelPackageName.length();
		if (s.substr(0, l)==ParserNS::Parser::TopLevelPackageName)
			return "";
		string::size_type dotPos = s.find('.', 0);
		assert(dotPos!=string::npos);
		return s.substr(0, dotPos+1);
	}
}

void printTaggedValue(RuntimeEnvironment *runtimeEnv, intrusive_ptr<TaggedValue> v, intrusive_ptr<OutputStreamValue> out, intrusive_ptr<const Expression> sourceExp) {
	const string recordName(getPkgNameWithDot(v->tag)+"PrintTagged");
	int slot = runtimeEnv->slotFor(recordName);
	if (slot<0) {
		runtimeEnv->interpreter->reportWarning("Cannot find \""+recordName+"\", so I'm implicitly untagging the value", sourceExp);
implicit_untagging:
		out->print(runtimeEnv, v->untagged(), sourceExp);
		return;
	}
	intrusive_ptr<RecordValue> record = dynamic_pointer_cast<RecordValue>(runtimeEnv->getTopLevelFrame()->varSlots[slot].value);
	if (!record) {
		runtimeEnv->interpreter->reportWarning("The variable \""+recordName+"\" is not a record, so I'm implicitly untagging the value", sourceExp);
		goto implicit_untagging;
	}
	const string fieldName(stripPkgName(v->tag));
	intrusive_ptr<RightValue> fieldValue = record->getField(fieldName);
	if (!fieldValue) {
		runtimeEnv->interpreter->reportWarning("The variable \""+recordName+"\" does not contain a field named \""+fieldName+"\", so I'm implicitly untagging the value", sourceExp);
		goto implicit_untagging;
	}
	intrusive_ptr<FunctionValue> printingFun = dynamic_pointer_cast<FunctionValue>(fieldValue);
	const string funName(recordName+"."+fieldName);
	if (!printingFun) {
		runtimeEnv->interpreter->reportWarning(funName+" does not contain a function, so I'm implicitly untagging the value", sourceExp);
		goto implicit_untagging;
	}
	if (!printingFun->canBeCalledWith(2)) {
		runtimeEnv->interpreter->reportWarning("The function "+funName+" cannot receive two arguments, so I'm implicitly untagging the value", sourceExp);
		goto implicit_untagging;
	}
	if (const intrusive_ptr<UserDefinedFunction> udf = dynamic_pointer_cast<UserDefinedFunction>(printingFun)) {
		for(int a=0; a<=1; ++a)
			if (udf->fnDecl->params[a].byRef) {
				runtimeEnv->interpreter->reportWarning("The function "+funName+" expects a by-ref argument, so I'm implicitly untagging the value", sourceExp);
				goto implicit_untagging;
			}
	}
	vector<Argument> args;
	args.push_back(Argument(false, new FakeExpression(out, sourceExp->getBegin(), sourceExp->getEnd()), true));
	args.push_back(Argument(false, new FakeExpression(v->untagged(), sourceExp->getBegin(), sourceExp->getEnd()), true));
	intrusive_ptr<InvocationExpression> fakeInvoke =
			new InvocationExpression(
					new FakeExpression(printingFun, sourceExp->getBegin(), sourceExp->getEnd()),
					boost::shared_ptr<Token>(),
					args,
					sourceExp->getEnd(),
					""
			);
	intrusive_ptr<Value> retValue = fakeInvoke->eval(runtimeEnv);
	if (!dynamic_pointer_cast<VoidValue>(retValue))
		throw RuntimeException("Printing-procedures must not return any value", sourceExp);
}

void OutputStreamValue::print(RuntimeEnvironment *runtimeEnv, intrusive_ptr<RightValue> v,  intrusive_ptr<const Expression> sourceExp) {
	if (this->isClosed)
		throw RuntimeException("Cannot print to a closed stream", sourceExp);
	if (const intrusive_ptr<StringValue> s = dynamic_pointer_cast<StringValue>(v))
		this->print(s->theString);
	else if (const intrusive_ptr<TaggedValue> tv = dynamic_pointer_cast<TaggedValue>(v))
		printTaggedValue(runtimeEnv, tv, this, sourceExp);
	else
		this->print(v);
}

void CppOutputStreamValue::flush() {
	out.flush();
}

intrusive_ptr<OutputStreamValue> CppOutputStreamValue::print(const string &s) {
	this->out << s;
	return this;
}

intrusive_ptr<OutputStreamValue> CppOutputStreamValue::print(intrusive_ptr<const RightValue> v) {
	this->out << v;
	return this;
}

intrusive_ptr<OutputStreamValue> CppOutputStreamValue::newline() {
	this->out << endl;
	return this;
}

bool TaggedValue::needsToBeCopiedBeforeChanges() const {
	return this->value->needsToBeCopiedBeforeChanges();
}

FramePointer::FramePointer(Frame *const frame) :
	frame(frame),
	id(frame ? frame->id : 0)
{
}

Frame *FramePointer::toCheckedPointer() const {
#ifdef VERBOSE_RUNTIME_DEBUG
	//cout << "toCheckedPointer this->frame=" << this->frame << endl;
	//if (this->frame)
	//	cout << "expected id=" << this->id << ", found " << this->frame->id << endl;
#endif // #ifdef VERBOSE_RUNTIME_DEBUG
	return this->frame ? (this->frame->id==this->id ? this->frame : 0) : 0;
}

Frame *FramePointer::toNonNullCheckedPointer(intrusive_ptr<const Identifier> expId) const {
	Frame *f = this->toCheckedPointer();
	if (!f)
		throw DeadEnviromentException(expId);
	return f;
}

intrusive_ptr<StaticEnv> CheckNamesVisitor::addEnvironmentForIterationVar(const string &identifier, const CharPointer &from, const CharPointer &to) {
	StaticEnv::VarData vd = this->env->lookup(identifier);
	if (vd.depth!=StaticEnv::IN_THE_WILD)
		this->errorReporter->reportWarning("This name hides an outer one", from, to);
	this->env = new StaticEnv(this->env);
	this->env->add(identifier, 0, 0, true, false, false);
	return this->env;
}

void CheckNamesVisitor::visit(IdInExpSuchThatExp &idInExpSuchThatExp) {
	idInExpSuchThatExp.exp1->accept(this);
	idInExpSuchThatExp.staticEnv = this->addEnvironmentForIterationVar(idInExpSuchThatExp.identifier, idInExpSuchThatExp.tokIdentifier.getBegin(), idInExpSuchThatExp.tokIdentifier.getEnd());
	idInExpSuchThatExp.exp2->accept(this);
	this->popEnv();
}

void CheckNamesVisitor::visit(ExpSuchThatIdInExpAndExp &expSuchThatIdInExpAndExp) {
	expSuchThatIdInExpAndExp.exp2->accept(this);
	expSuchThatIdInExpAndExp.staticEnv = this->addEnvironmentForIterationVar(expSuchThatIdInExpAndExp.identifier, expSuchThatIdInExpAndExp.tokIdentifier.getBegin(), expSuchThatIdInExpAndExp.tokIdentifier.getEnd());
	expSuchThatIdInExpAndExp.exp1->accept(this);
	if (expSuchThatIdInExpAndExp.optionalExp3)
		expSuchThatIdInExpAndExp.optionalExp3->accept(this);
	this->popEnv();
}

void CheckNamesVisitor::visit(FieldAccessExpression &fieldAccessExp) {
	if (const intrusive_ptr<Identifier> identifier = dynamic_pointer_cast<Identifier>(fieldAccessExp.targetExp)) {
		StaticEnv::VarData vd = this->env->lookup(identifier->identifier);
		if (vd.depth==StaticEnv::IN_THE_WILD) {
			string alias;
			if (this->aliases->lookup(identifier->identifier, alias)) {
				identifier->identifier = alias;
				assert(vd.index<0);
				assert(!vd.isCapturedValue);
				assert(!vd.isIterationVar);
				assert(!vd.isImplictlyImported);
				vd.depth = StaticEnv::TOP_LEVEL;
			} else {
				if (this->insideFunctions)
					this->reportError(VariableNotFoundException::errorMessage(identifier->identifier, identifier->arity, this->env), identifier->getBegin(), identifier->getEnd());
				else
					vd.depth = StaticEnv::TOP_LEVEL; // outside fn-procs, anything goes...
			}
		} else {
			string alias;
			if (this->aliases->lookup(identifier->identifier, alias))
				this->errorReporter->reportWarning("I'm using the local/imported variable here, but note that there is an alias, with the same name, for the package "+alias, identifier->getBegin(), identifier->getEnd(), WS_PEDANTIC);
		}
		identifier->varData = vd;
	} else
		fieldAccessExp.targetExp->accept(this);
}

void CheckNamesVisitor::visit(Identifier &identifier) {
	StaticEnv::VarData vd = this->env->lookup(identifier.identifier);
	if (vd.depth==StaticEnv::IN_THE_WILD) {
		if (this->runtimeEnvironment->isRegisteredType(identifier.identifier))
			vd.depth = StaticEnv::TOP_LEVEL;
		else {
			if (this->insideFunctions)
				this->reportError(VariableNotFoundException::errorMessage(identifier.identifier, identifier.arity, this->env), identifier.getBegin(), identifier.getEnd());
			else {
				vd.depth = StaticEnv::TOP_LEVEL; // outside fn-procs, anything goes...
				vd.isImplictlyImported = true; // this is important for non-fnproc package member declarations
			}
		}
	}
	identifier.varData = vd;
}

void CheckNamesVisitor::reportError(const string &msg, const CharPointer &from, const CharPointer &to) {
	this->foundErrors = true;
	this->errorReporter->reportError(msg, from, to);
}

void CheckNamesVisitor::visit(ForStatement &forStmt) {
	forStmt.beginExp->accept(this);
	if (forStmt.stepExp)
		forStmt.stepExp->accept(this);
	forStmt.endExp->accept(this);
	forStmt.staticEnv = this->addEnvironmentForIterationVar(forStmt.identifier, forStmt.tokIdentifier.getBegin(), forStmt.tokIdentifier.getEnd());
	forStmt.statements->accept(this);
	this->popEnv();
}

void CheckNamesVisitor::visit(ForeachStatement &foreachStmt) {
	foreachStmt.inExp->accept(this);
	foreachStmt.staticEnv = this->addEnvironmentForIterationVar(foreachStmt.identifier, foreachStmt.tokIdentifier.getBegin(), foreachStmt.tokIdentifier.getEnd());
	foreachStmt.statements->accept(this);
	this->popEnv();
}

void CheckNamesVisitor::visit(TryStatement &tryStmt) {
	tryStmt.tryStatements->accept(this);
	tryStmt.staticEnv = this->addEnvironmentForIterationVar(tryStmt.identifier, tryStmt.tokIdentifier.getBegin(), tryStmt.tokIdentifier.getEnd());
	tryStmt.uponErrorStatements->accept(this);
	this->popEnv();
}

void CheckNamesVisitor::visit(FunctionDeclaration &funDecl) {
	bool oldInsideFunctions = this->insideFunctions;
	this->insideFunctions = true;
	funDecl.staticEnv = this->env = new StaticEnv(this->env);
	BOOST_FOREACH(Import &i, funDecl.imports) {
		StaticEnv::VarData vd = (i.type==Import::IT_BYVALUE ? this->env->parent : this->env)->lookup(i.name);
		const int depth = vd.depth;
		if (i.type==Import::IT_TOPLEVEL) {
			if (i.implicit && depth>=0)
				this->errorReporter->reportWarning("The name \""+i.name+"\" has been implicitly imported from TopLevel; note that importing the name by-ref could yield a different result (add an explicit import to avoid this message)", funDecl.getBegin(), funDecl.getEnd());
			this->env->add(i.name, StaticEnv::TOP_LEVEL, -1, false, false, i.implicit);
			continue;
		}
		if (depth==StaticEnv::IN_THE_WILD) {
			assert(i.expId);
			this->reportError("Cannot find "+i.name, i.expId->getBegin(), i.expId->getEnd());
			continue;
		}
		if (i.type==Import::IT_BYVALUE) {
			assert(i.expId);
			i.expId->varData = vd;
			assert(!i.implicit);
			this->env->add(i.name, 0, i.byValueIndex, false, true, false);
		} else {
			assert(i.type==Import::IT_BYREF);
			if (vd.isIterationVar) {
				assert(i.expId);
				this->reportError("Iteration variables cannot be imported by-ref", i.expId->getBegin(), i.expId->getEnd());
			} if (vd.isCapturedValue) {
				assert(i.expId);
				this->reportError("Names imported by-value cannot be re-imported by-ref", i.expId->getBegin(), i.expId->getEnd());
			} else {
				assert(!i.implicit);
				this->env->add(i.name, depth, vd.index, false, false, false);
			}
		}
	}
	for(map<string, int>::const_iterator it=funDecl.localNames.begin(); it!=funDecl.localNames.end(); ++it)
		env->add(it->first, 0, it->second, false, false, false);
	funDecl.statements->accept(this);
	this->popEnv();
	this->insideFunctions = oldInsideFunctions;
}

void CheckNamesVisitor::visit(InvocationExpression &invocationExp) {
	ParsedObjectVisitor::visit(invocationExp);
	BOOST_FOREACH(const Argument &arg, invocationExp.args)
		if (arg.byRef)
			if (const intrusive_ptr<Identifier> id = dynamic_pointer_cast<Identifier>(arg.exp)) {
				if (id->varData.isCapturedValue)
					this->reportError("Cannot pass by-ref an imported by value", arg.exp->getBegin(), arg.exp->getEnd());
			}
}

void CheckNamesVisitor::visit(AssignmentStatement &assignmentStmt) {
	ParsedObjectVisitor::visit(assignmentStmt);
	if (const intrusive_ptr<Identifier> id = dynamic_pointer_cast<Identifier>(assignmentStmt.leftExp)) {
		if (id->varData.isCapturedValue)
			this->reportError("Cannot assign to a variable imported by value", assignmentStmt.leftExp->getBegin(), assignmentStmt.leftExp->getEnd());
	}
}

void CheckNamesVisitor::visit(DefineStatement &defineStmt) {
	defineStmt.funDecl->accept(this);
}

void CheckNamesVisitor::visit(AliasStatement &aliasStmt) {
	assert(this->aliases);
	//if (aliasStmt.statements)
	//	this->aliases = new AliasEnv(this->aliases);
	BOOST_FOREACH(const Binding &b, aliasStmt.bindings) {
		const string newAlias(b.identifier->identifier);
		if ( /* aliasStmt.statements || */ this->insidePackage) {
			string useless;
			if (this->aliases->parent->lookup(newAlias, useless))
				this->errorReporter->reportWarning("This alias hides an outer one", b.identifier->getBegin(), b.identifier->getEnd());
		}
		this->aliases->add(newAlias, b.packageName);
	}
	/* if (aliasStmt.statements) {
		aliasStmt.statements->accept(this);
		this->aliases = this->aliases->parent;
		assert(this->aliases);
	} */
}

void CheckNamesVisitor::visit(PackageStatement &p) {
	assert(this->aliases);
	this->insidePackage = true;
	this->aliases = new AliasEnv(this->aliases);
	p.statements->accept(this);
	this->aliases = this->aliases->parent;
	assert(this->aliases);
	this->insidePackage = false;
}

void ResolvePackageNamesVisitor::visit(InvocationExpression &invocationExp) {
	const intrusive_ptr<Identifier> id = dynamic_pointer_cast<Identifier>(invocationExp.targetExp);
	if (!id || !id->varData.isImplictlyImported || this->packageStatement.memberNames.find(id->identifier)==this->packageStatement.memberNames.end())
		invocationExp.targetExp->accept(this);
	else {
		StaticEnv::VarData &vd = id->varData;
		assert(vd.depth == StaticEnv::TOP_LEVEL);
		assert(!vd.isCapturedValue);
		assert(!vd.isIterationVar);
		id->identifier = this->packageStatement.name+"."+id->identifier;
		vd.isImplictlyImported = false;
		//cout << "\n\n### " << id << " --> " << invocationExp.targetExp << " ###\n\n";
	}
	BOOST_FOREACH(const Argument &arg, invocationExp.args)
		arg.exp->accept(this);
}

namespace {
	// based on: http://www.merriampark.com/ld.htm and http://www.merriampark.com/ldcpp.htm
	int levenshteinDistance(string source, string target) {
	  const int n = source.length();
	  const int m = target.length();
	  if (n == 0)
		return m;
	  if (m == 0)
		return n;
	  to_lower(source);
	  to_lower(target);
	  vector<vector<int> > matrix(n+1);
	  for (int i = 0; i <= n; ++i)
		matrix[i].resize(m+1);
	  for (int i = 0; i <= n; ++i)
		matrix[i][0]=i;
	  for (int j = 0; j <= m; ++j)
		matrix[0][j]=j;
	  for (int i = 1; i <= n; ++i) {
		const char s_i = source[i-1];
		for (int j = 1; j <= m; ++j) {
		  const char t_j = target[j-1];
		  const int cost = s_i == t_j ? 0 : 1;
		  const int above = matrix[i-1][j];
		  const int left = matrix[i][j-1];
		  const int diag = matrix[i-1][j-1];
		  int cell = min(above + 1, min(left + 1, diag + cost));
		  // Step 6A: Cover transposition, in addition to deletion,
		  // insertion and substitution. This step is taken from:
		  // Berghel, Hal ; Roach, David : "An Extension of Ukkonen's
		  // Enhanced Dynamic Programming ASM Algorithm"
		  // (http://www.acm.org/~hlb/publications/asm/asm.html)
		  if (i>2 && j>2) {
			int trans=matrix[i-2][j-2]+1;
			if (source[i-2]!=t_j) ++trans;
			if (s_i!=target[j-2]) ++trans;
			if (cell>trans) cell=trans;
		  }
		  matrix[i][j]=cell;
		}
	  }
	  return matrix[n][m];
	}

	inline int maximumDistanceForSimilaryIdentifiers(const string &id) {
		const int len = id.length();
		assert(len>=0);
		if (len<=1)
			return 0; // use only case-insensitivity when looking for one-char identifiers (otherwise all of them would be considered similar)
		if (len<=4)
			return 1; // accept one-char difference with short/medium identifiers
		return 2; // accept two-char difference for all the long identifiers
	}
}

void Interpreter::reportError(const string &msg) {
	this->errorReporter->reportError(msg);
}

void Interpreter::reportError(const string &msg, const CharPointer &from, const CharPointer &to) {
	this->errorReporter->reportError(msg, from, to);
}

intrusive_ptr<const StaticEnv> RuntimeEnvironment::findStaticEnv() {
	Frame *f = this->currentFrame;
//JAA	for(; f>this->frames; --f)
          for(; f>getTopLevelFrame(); --f)
		if (f->userdefinedFun) {
			assert(f->userdefinedFun->fnDecl);
			assert(f->userdefinedFun->fnDecl->staticEnv);
			return f->userdefinedFun->fnDecl->staticEnv;
		}
	return topLevelStaticEnv;
}

Frame *RuntimeEnvironment::pushFrame(const FramePointer &accessLink, intrusive_ptr<const InvocationExpression> invocationExp, intrusive_ptr<const ParsedObject> block, intrusive_ptr<const UserDefinedFunction> userdefinedFun) {
	assert(block);
#ifdef VERBOSE_RUNTIME_DEBUG
	cout << "pushFrame for " << block << endl;
#endif // #ifdef VERBOSE_RUNTIME_DEBUG
	if (this->currentFrame==&frames.back())
          throw RuntimeException("Too many nested scopes", invocationExp ? invocationExp : block);
	Frame * const f = ++this->currentFrame;
	f->id = ++this->nextFrameId;
	assert(f->id); // f->id==0 would mean we've just overflowed!
	f->accessLink = accessLink;
	f->invocationExp = invocationExp;
	f->block = block;
	f->userdefinedFun = userdefinedFun;
#ifdef C5IDE
	f->singleStepOnPop = false;
#endif // #ifdef C5IDE
	return f;
}

Frame *RuntimeEnvironment::pushIterationFrame(intrusive_ptr<const ParsedObject> block) {
	assert(block);
	assert(block->staticEnv);
	Frame *f = this->pushFrame(this->currentFrame, 0, block, 0);
	this->currentFrame->varSlots.push_back(VariableSlot(VoidValue::theInstance, static_cast<VariableSlot::Flags>(VariableSlot::VSF_SystemProtected|VariableSlot::VSF_IterationVariable)));
	return f;
}

void RuntimeEnvironment::popFrame() {
	Frame * const f = this->currentFrame--;
	assert(f > getTopLevelFrame());
#ifdef C5IDE
	if (f->singleStepOnPop)
		this->interpreter->singleStepExecution = true;
#endif // #ifdef C5IDE
	f->id = 0;
	f->accessLink.reset();
	f->varSlots.clear();
	f->invocationExp = 0;
	f->block = 0;
	f->userdefinedFun = 0;
#ifdef VERBOSE_RUNTIME_DEBUG
	if (this->currentFrame==frames)
		cout << "popFrame, back to global\n";
	else
		cout << "popFrame, back to: " << this->currentFrame->block << endl;
#endif // #ifdef VERBOSE_RUNTIME_DEBUG
}

vector<SnapshotFrame> RuntimeEnvironment::takeSnapshot() {
	vector<SnapshotFrame> result;
	const Frame *f = this->currentFrame;
	do {
		if (f->invocationExp) {
			result.push_back(SnapshotFrame(*f));
		}
	} while (--f != getTopLevelFrame());
	return result;
}

void RuntimeEnvironment::registerType(intrusive_ptr<TypeValue> type) {
#ifndef NDEBUG
	bool inserted =
#endif
	this->registeredTypeNames.insert(type->name).second;
	assert(inserted);
	this->setTopLevelVar(type->name, type, VariableSlot::VSF_SystemProtected);
}

intrusive_ptr<ListValue> RuntimeEnvironment::currentTypes() {
	intrusive_ptr<ListValue> result(new ListValue);
	BOOST_FOREACH(const string &s, this->registeredTypeNames) {
		const int slot = this->slotFor(s);
		assert(slot>=0 && static_cast<vector<VariableSlot>::size_type>(slot)<getTopLevelFrame()->varSlots.size());
		result->addValue(intrusive_ptr_cast<TypeValue>(getTopLevelFrame()->varSlots[slot].value));
	}
	typedef pair<string, intrusive_ptr<TypeValue> > pair;
	BOOST_FOREACH(const pair &p, TypeValue::taggedTypes)
		result->addValue(p.second);
	return result;
}

  RuntimeEnvironment::RuntimeEnvironment(Interpreter * const interpreter, intrusive_ptr<OutputStreamValue> standardOutput, long MaxStackSize) :
    frames(MaxStackSize),
    currentFrame(&(this->frames[0])),
	nextFrameId(1),
	standardOutput(standardOutput),
	topLevelStaticEnv(new StaticEnv(this)),
	topLevelAliases(new AliasEnv(0)),
	itSlot(this->setTopLevelVar(Interpreter::IT, VoidValue::theInstance, VariableSlot::VSF_SystemProtected)),
	currentRingSlot(this->setTopLevelVar(Interpreter::CURRENT_RING, VoidValue::theInstance, VariableSlot::VSF_SystemProtected)),
	interpreter(interpreter)
{
	this->initBuiltInFunctions();
	this->registerType(BoolValue::type);
	this->registerType(FunctionValue::type);
	this->registerType(ListValue::type);
	this->registerType(BigIntValue::type);
	this->registerType(BigRatValue::type);
	this->registerType(RecordValue::type);
	this->registerType(TypeValue::type);
	this->registerType(StringValue::type);
	this->registerType(VoidValue::type);
	this->registerType(ErrorValue::type);
	this->registerType(OutputStreamValue::type);
	this->registerType(ZZModValue::type);
	this->registerType(RingElemValue::type);
	this->registerType(RatFunValue::type);
	this->registerType(IdealValue::type);
	this->registerType(JBMillValue::type);
	this->registerType(ModuleValue::type);
	this->registerType(ModuleElemValue::type);
	this->registerType(MatrixValue::type);
	this->registerType(RingValue::type);
	this->registerType(IntMapValue::type);
	this->registerType(RingHomValue::type);
	this->registerType(MatrixRowValue::type);
	this->setTopLevelVar("ZZ", new RingValue(RingZZ()), VariableSlot::VSF_SystemProtected);
	this->setTopLevelVar("QQ", new RingValue(RingQQ()), VariableSlot::VSF_SystemProtected);
	this->setTopLevelVar("R", new RingValue(NewPolyRing(RingQQ(),symbols("x","y","z"))), VariableSlot::VSF_None);
	this->initMaps();
}

intrusive_ptr<ListValue> RuntimeEnvironment::topLevelFunctions() {
	intrusive_ptr<ListValue> result(new ListValue);
	for(map<string, int>::const_iterator it = this->topLevelIdentifiers.begin(); it!=this->topLevelIdentifiers.end(); ++it) {
		assert(it->first.length());
		assert(it->second>=0 && it->second<static_cast<int>(getTopLevelFrame()->varSlots.size()));
		const VariableSlot &vslot = getTopLevelFrame()->varSlots[it->second];
		if (intrusive_ptr<FunctionValue> f = dynamic_pointer_cast<FunctionValue>(vslot.value)) {
			intrusive_ptr<RecordValue> record(new RecordValue);
			result->addValue(record);
			record->setField("name", new StringValue(it->first));
			record->setField("IsExported", BoolValue::fromBool(vslot.hasBeenExported()));
		}
	}
	return result;
}

  long RuntimeEnvironment::ResizeStack(long NewSize, const boost::intrusive_ptr<const AST::Expression> OrigExp)
  {
    if (NewSize < 2)
      throw RuntimeException("Ridiculous stack size", OrigExp);
    const long CurrStackHeight = 1+(currentFrame-&frames[0]);
    if (CurrStackHeight >= NewSize)
      throw RuntimeException("Stack size too small", OrigExp);
    frames.resize(NewSize);
    return CurrStackHeight;
  }

int RuntimeEnvironment::slotFor(const string &id) {
	map<string, int>::const_iterator i=this->topLevelIdentifiers.find(id);
	return i!=this->topLevelIdentifiers.end() ? i->second : -1;
}

int RuntimeEnvironment::setTopLevelVar(const string &id, intrusive_ptr<Value> v, VariableSlot::Flags flags) {
	const int slot = this->slotFor(id);
	if (slot>=0) {
		VariableSlot &vs = this->frames[0].varSlots[slot];
		vs.value = v;
		vs.flags = flags;
		return slot;
	}
	const int newSlot = this->frames[0].varSlots.size();
	this->frames[0].varSlots.push_back(VariableSlot(v, flags));
	this->topLevelIdentifiers.insert(make_pair(id, newSlot));
	return newSlot;
}

  Interpreter::Interpreter(bool warnAboutCocoa5, intrusive_ptr<LineProvider> lineProvider, intrusive_ptr<ErrorReporter> errorReporter,  intrusive_ptr<OutputStreamValue> standardOutput, bool fullCoCoALibError, long MaxStackSize) :
	warnAboutCocoa5(warnAboutCocoa5),
	fullCoCoALibError(fullCoCoALibError),
	lineProvider(lineProvider),
	runtimeEnvironment(this, standardOutput, MaxStackSize),
	controlC(false),
#ifdef C5IDE
	singleStepExecution(false),
	doStepOver(false),
#endif // #ifdef C5IDE
	errorReporter(errorReporter)
#ifdef C5IDE
	, status(IS_WAITING_FOR_COMMAND) // while not 100% true (the call to this->run() will make this happen), it's a safe initialization value
#endif // #ifdef C5IDE
{
	assert(this->lineProvider);
	assert(this->errorReporter);
}

string VariableNotFoundException::errorMessage(const string &varName, int arity, intrusive_ptr<const StaticEnv> env) {
	string message("Cannot find a variable named \"");
	message += varName;
	message += "\" in scope";
	set<string> s;
	bool thereIsAnExactMatch=false;
	env->collectSimilarlyNamedIdentifiers(varName, arity, s, thereIsAnExactMatch);
	if (thereIsAnExactMatch)
		return message+", but there is one in an outside scope. You're probably missing an import statement";
	const int size = s.size();
	if (size==0)
		return message;
	if (size==1) {
		message += ".\nA similarly named variable (that you might need to import) is: \"";
		message += *s.begin();
		message += '\"';
		return message;
	}
	message += ".\nSimilarly named variables (that you might need to import) are: ";
	bool first = true;
	BOOST_FOREACH(const string &id, s) {
		if (first)
			first = false;
		else
			message += ", ";
		message += '\"';
		message += id;
		message += '\"';
	}
	return message;
}

void RecordValue::collectSimilarlyNamedFields(const string &fieldName, set<string> &set) const {
	const int maxDistance = maximumDistanceForSimilaryIdentifiers(fieldName);
	for(MapType::const_iterator pos = this->fields.begin(); pos!=this->fields.end(); ++pos)
		if (levenshteinDistance(fieldName, pos->first)<=maxDistance)
			set.insert(pos->first);
}

vector<string> RecordValue::myFieldNamesStrings() const {
	vector<string> result;
	for(MapType::const_iterator pos = this->fields.begin(); pos!=this->fields.end(); ++pos)
		result.push_back(pos->first);
	return result;
}

intrusive_ptr<ListValue> RecordValue::fieldNames() const {
	intrusive_ptr<ListValue> result(new ListValue);
	for(MapType::const_iterator pos = this->fields.begin(); pos!=this->fields.end(); ++pos)
		result->addValue(new StringValue(pos->first));
	return result;
}

string FieldNotFoundException::errorMessage(const string &fieldName, intrusive_ptr<const RecordValue> rv) {
	string message("Cannot find a field named \"");
	message += fieldName;
	message += '\"';
	if (rv->numberOfFields()==0)
		message += ". Note: the record is actually empty; that is, it has no fields at all";
	else {
		set<string> s;
		rv->collectSimilarlyNamedFields(fieldName, s);
		const int size = s.size();
		if (size==0)
			return message;
		if (size==1) {
			message += ".\nA similarly named (existing) field is: \"";
			message += *s.begin();
			message += '\"';
			return message;
		}
		message += ".\nSimilarly named (existing) fields are: ";
		bool first = true;
		BOOST_FOREACH(const string &id, s) {
			if (first)
				first = false;
			else
				message += ", ";
			message += '\"';
			message += id;
			message += '\"';
		}
	}
	return message;
}

void Interpreter::reportWarning(const string &msg, const intrusive_ptr<const ParsedObject> po) {
	this->errorReporter->reportWarning(msg, po->getBegin(), po->getEnd());
}

void Interpreter::reportWarning(const string &msg, const CharPointer &from, const CharPointer &to) {
	this->errorReporter->reportWarning(msg, from, to);
}

intrusive_ptr<Value> RuntimeEnvironment::evalArg(const Argument &arg, EvalKind evalKind) {
	const bool byRef = evalKind==EVAL_BY_REF;
	if (arg.byRef && !byRef)
		throw CorrespondingParameterNotReferenceException(arg.exp);
	intrusive_ptr<Value> v;
	if (byRef) {
		if (!arg.byRef && !arg.synthetized)
			this->interpreter->reportWarning("You should use \"Ref\" when passing arguments by reference", arg.exp);
		v = arg.exp->eval(this);
		if (!dynamic_pointer_cast<LeftValue>(v))
			throw RuntimeException("Only left-values can be passed by reference", arg.exp);
	} else
		v = arg.exp->evalAs<RightValue>(this);
	return v;
}

const string Interpreter::IT("It");
const string Interpreter::CURRENT_RING("CurrentRing");

namespace // anonymous
{

  void checkProtection(VariableSlot *vs, const intrusive_ptr<const Identifier> id)
  {
    const string& name = id->identifier;
    if (vs->isProtected())
    {
      const string& reason(vs->getProtectionReason());
      if (!reason.empty())
        throw ProtectedVariableException(name, "Cannot set \""+name+"\" (protected: "+reason+")", id);
      if (vs->isSystemProtected())
        throw ProtectedVariableException(name, "Cannot set \""+name+"\" (system-protected variable)", id);
      throw ProtectedVariableException(name, "Cannot set \""+name+"\" (user-protected variable)", id);
    }
    if (vs->isPackage())
    {
      throw ProtectedVariableException(name, "Cannot set \""+name+"\" (package-exported variable)", id);
    }
  }

} // end of anonymous namespace

bool TypeValue::isProperSubtypeOf(int otherDI) const {
	switch (this->dispatchIndex) {
	case DISPATCH_INDEX_OF_BIGRAT:
		return otherDI==DISPATCH_INDEX_OF_RINGELEM;
	case DISPATCH_INDEX_OF_BIGINT:
		return otherDI==DISPATCH_INDEX_OF_BIGRAT || otherDI==DISPATCH_INDEX_OF_RINGELEM;
	case DISPATCH_INDEX_OF_RINGHOM:
		return otherDI==DISPATCH_INDEX_OF_FUNCTION;
	case DISPATCH_INDEX_OF_RECORD:
	case DISPATCH_INDEX_OF_BOOL:
	case DISPATCH_INDEX_OF_VOID:
	case DISPATCH_INDEX_OF_STRING:
	case DISPATCH_INDEX_OF_LIST:
	case DISPATCH_INDEX_OF_FUNCTION:
	case DISPATCH_INDEX_OF_TYPE:
	case DISPATCH_INDEX_OF_ERROR:
	case DISPATCH_INDEX_OF_OSTREAM:
	case DISPATCH_INDEX_OF_ZMOD:
	case DISPATCH_INDEX_OF_RATFUN:
	case DISPATCH_INDEX_OF_MODULEELEM:
	case DISPATCH_INDEX_OF_IDEAL:
	case DISPATCH_INDEX_OF_MODULE:
	case DISPATCH_INDEX_OF_MAT:
	case DISPATCH_INDEX_OF_RING:
	case DISPATCH_INDEX_OF_PACKAGE:
	case DISPATCH_INDEX_OF_RINGELEM:
	case DISPATCH_INDEX_OF_INTMAP:
	case DISPATCH_INDEX_OF_MATRIXROW:
	case DISPATCH_INDEX_OF_JBMILL:
		return false;
	default:
		assert(false);
	}
	return false;
}

intrusive_ptr<const RightValue> TypeValue::convert(intrusive_ptr<const RightValue> from, intrusive_ptr<const RightValue> to) {
	assert(from);
	assert(to);
	assert(from->getType()->isProperSubtypeOf(to->getType()->dispatchIndex));
	switch (to->getType()->dispatchIndex) {
		case DISPATCH_INDEX_OF_BIGRAT:
                  return new BigRatValue(BigRat(intrusive_ptr_cast<const BigIntValue>(from)->theInteger,1));
		case DISPATCH_INDEX_OF_RINGELEM:
			if (from->getType()->dispatchIndex==DISPATCH_INDEX_OF_BIGINT)
				return new RingElemValue(RingElem(owner(intrusive_ptr_cast<const RingElemValue>(to)->theRingElem), intrusive_ptr_cast<const BigIntValue>(from)->theInteger));
			return new RingElemValue(RingElem(owner(intrusive_ptr_cast<const RingElemValue>(to)->theRingElem), intrusive_ptr_cast<const BigRatValue>(from)->theRational));
		case DISPATCH_INDEX_OF_FUNCTION:
			return from;
	}
	assert(false);
	return 0;
}

intrusive_ptr<RightValue> RuntimeEnvironment::binaryOperatorDispatch(intrusive_ptr<const RightValue> leftOp, intrusive_ptr<const RightValue> rightOp, DispatchMapType map, const LexerNS::CharPointer &beginOperatorSourcePosition, const LexerNS::CharPointer &endOperatorSourcePosition) {
	this->interpreter->checkForInterrupts(beginOperatorSourcePosition, endOperatorSourcePosition);
	const intrusive_ptr<const TypeValue> leftType(leftOp->getType());
	assert(leftType);
	int leftIndex = leftType->dispatchIndex;
	if (leftIndex<0) {
		const intrusive_ptr<const TaggedValue> tv = dynamic_pointer_cast<const TaggedValue>(leftOp);
		assert(tv);
		this->interpreter->reportWarning("I'm implicitly untagging the left operand; use untagged() to avoid this warning", beginOperatorSourcePosition, endOperatorSourcePosition);
		leftOp = tv->untagged();
		leftIndex = leftOp->getType()->dispatchIndex;
	}
	assert(leftIndex>=0 && leftIndex<TypeValue::NUM_OF_DISPATCHABLE_TYPES);
	const intrusive_ptr<const TypeValue> rightType(rightOp->getType());
	assert(rightType);
	int rightIndex = rightType->dispatchIndex;
	if (rightIndex<0) {
		const intrusive_ptr<const TaggedValue> tv = dynamic_pointer_cast<const TaggedValue>(rightOp);
		assert(tv);
		this->interpreter->reportWarning("I'm implicitly untagging the right operand; use untagged() to avoid this warning", beginOperatorSourcePosition, endOperatorSourcePosition);
		rightOp = tv->untagged();
		rightIndex = rightOp->getType()->dispatchIndex;
	}
	assert(rightIndex>=0 && rightIndex<TypeValue::NUM_OF_DISPATCHABLE_TYPES);
	for(;;) {
		BinaryOpFunc f = map[leftIndex][rightIndex];
		if (f)
			try	{
				return f(this, leftOp, rightOp, beginOperatorSourcePosition, endOperatorSourcePosition);
			} catch (const ErrorInfo& err) {
				this->announceCLE(err);
				throw RuntimeException(err.what(), beginOperatorSourcePosition, endOperatorSourcePosition);
			}
		if (leftType->isProperSubtypeOf(rightIndex) && map[rightIndex][rightIndex]) {
			try	{
        leftOp = TypeValue::convert(leftOp, rightOp);
			} catch (const ErrorInfo& err) {
				this->announceCLE(err);
				throw RuntimeException(err.what(), beginOperatorSourcePosition, endOperatorSourcePosition);
			}
			assert(leftOp->getType()->dispatchIndex == rightIndex);
			leftIndex = rightIndex;
			continue;
		}
		if (rightType->isProperSubtypeOf(leftIndex) && map[leftIndex][leftIndex]) {
      try {
        rightOp = TypeValue::convert(rightOp, leftOp);
			} catch (const ErrorInfo& err) {
				this->announceCLE(err);
				throw RuntimeException(err.what(), beginOperatorSourcePosition, endOperatorSourcePosition);
			}
			assert(rightOp->getType()->dispatchIndex == leftIndex);
			rightIndex = leftIndex;
			continue;
		}
		const string leftType = leftOp->getType()->name;
		const string rightType = rightOp->getType()->name;
		throw RuntimeException("I don't know how to evaluate operator "+beginOperatorSourcePosition.stringTo(endOperatorSourcePosition)+" between "+leftType+" and "+rightType, beginOperatorSourcePosition, endOperatorSourcePosition);
	}
}

void RuntimeEnvironment::announceCLE(const ErrorInfo& err) {
	if (this->interpreter->fullCoCoALibError) {
		ostringstream os;
		ANNOUNCE(os, err);
		this->standardOutput->print(os.str());
	}
}

template<typename LeftType, typename RightType>
intrusive_ptr<RightValue> opEqual(RuntimeEnvironment *, const intrusive_ptr<const RightValue> LeftTypeOp, const intrusive_ptr<const RightValue> RightTypeOp, const CharPointer &, const CharPointer &) {
	return BoolValue::fromBool(theValue(intrusive_ptr_cast<const LeftType>(LeftTypeOp)) == theValue(intrusive_ptr_cast<const RightType>(RightTypeOp)));
}

template<typename LeftType, typename RightType>
intrusive_ptr<RightValue> opNotEqual(RuntimeEnvironment *, const intrusive_ptr<const RightValue> LeftTypeOp, const intrusive_ptr<const RightValue> RightTypeOp, const CharPointer &, const CharPointer &) {
	return BoolValue::fromBool(theValue(intrusive_ptr_cast<const LeftType>(LeftTypeOp)) != theValue(intrusive_ptr_cast<const RightType>(RightTypeOp)));
}

template<typename LeftType, typename RightType>
intrusive_ptr<RightValue> opLessThan(RuntimeEnvironment *, const intrusive_ptr<const RightValue> LeftTypeOp, const intrusive_ptr<const RightValue> RightTypeOp, const CharPointer &, const CharPointer &) {
	return BoolValue::fromBool(theValue(intrusive_ptr_cast<const LeftType>(LeftTypeOp)) < theValue(intrusive_ptr_cast<const RightType>(RightTypeOp)));
}

template<typename LeftType, typename RightType>
intrusive_ptr<RightValue> opLessOrEqual(RuntimeEnvironment *, const intrusive_ptr<const RightValue> LeftTypeOp, const intrusive_ptr<const RightValue> RightTypeOp, const CharPointer &, const CharPointer &) {
	return BoolValue::fromBool(theValue(intrusive_ptr_cast<const LeftType>(LeftTypeOp)) <= theValue(intrusive_ptr_cast<const RightType>(RightTypeOp)));
}

template<typename LeftType, typename RightType>
intrusive_ptr<RightValue> opGreaterOrEqual(RuntimeEnvironment *, const intrusive_ptr<const RightValue> LeftTypeOp, const intrusive_ptr<const RightValue> RightTypeOp, const CharPointer &, const CharPointer &) {
	return BoolValue::fromBool(theValue(intrusive_ptr_cast<const LeftType>(LeftTypeOp)) >= theValue(intrusive_ptr_cast<const RightType>(RightTypeOp)));
}

template<typename LeftType, typename RightType>
intrusive_ptr<RightValue> opGreaterThan(RuntimeEnvironment *, const intrusive_ptr<const RightValue> LeftTypeOp, const intrusive_ptr<const RightValue> RightTypeOp, const CharPointer &, const CharPointer &) {
	return BoolValue::fromBool(theValue(intrusive_ptr_cast<const LeftType>(LeftTypeOp)) > theValue(intrusive_ptr_cast<const RightType>(RightTypeOp)));
}

template<typename LeftType, typename RightType>
intrusive_ptr<RightValue> opPlus(RuntimeEnvironment *, const intrusive_ptr<const RightValue> LeftTypeOp, const intrusive_ptr<const RightValue> RightTypeOp, const CharPointer &, const CharPointer &) {
	return Value::simplifiedValueFrom(theValue(intrusive_ptr_cast<const LeftType>(LeftTypeOp)) + theValue(intrusive_ptr_cast<const RightType>(RightTypeOp)));
}

template<typename LeftType, typename RightType>
intrusive_ptr<RightValue> opMinus(RuntimeEnvironment *, const intrusive_ptr<const RightValue> LeftTypeOp, const intrusive_ptr<const RightValue> RightTypeOp, const CharPointer &, const CharPointer &) {
	return Value::simplifiedValueFrom(theValue(intrusive_ptr_cast<const LeftType>(LeftTypeOp)) - theValue(intrusive_ptr_cast<const RightType>(RightTypeOp)));
}

template<typename LeftType, typename RightType>
intrusive_ptr<RightValue> opStar(RuntimeEnvironment *, const intrusive_ptr<const RightValue> LeftTypeOp, const intrusive_ptr<const RightValue> RightTypeOp, const CharPointer &, const CharPointer &) {
	return Value::simplifiedValueFrom(theValue(intrusive_ptr_cast<const LeftType>(LeftTypeOp)) * theValue(intrusive_ptr_cast<const RightType>(RightTypeOp)));
}

template<typename LeftType, typename RightType>
intrusive_ptr<RightValue> opSlash(RuntimeEnvironment *, const intrusive_ptr<const RightValue> LeftTypeOp, const intrusive_ptr<const RightValue> RightTypeOp, const CharPointer &, const CharPointer &) {
	return Value::simplifiedValueFrom(theValue(intrusive_ptr_cast<const LeftType>(LeftTypeOp)) / theValue(intrusive_ptr_cast<const RightType>(RightTypeOp)));
}

intrusive_ptr<RightValue> opLessThan_RingElem_RingElem(RuntimeEnvironment *, const intrusive_ptr<const RightValue> leftOp, const intrusive_ptr<const RightValue> rightOp, const CharPointer &beginOperatorSourcePosition, const CharPointer &endOperatorSourcePosition) {
	intrusive_ptr<const RingElemValue> a = intrusive_ptr_cast<const RingElemValue>(leftOp);
	intrusive_ptr<const RingElemValue> b = intrusive_ptr_cast<const RingElemValue>(rightOp);
  if (IsOrderedDomain(owner(a->theRingElem)))
    return BoolValue::fromBool(a->theRingElem < b->theRingElem);
  if (IsMonomial(a->theRingElem) && IsMonomial(b->theRingElem)
      && IsOne(LC(a->theRingElem)) && IsOne(LC(b->theRingElem)))
    return BoolValue::fromBool(LPP(a->theRingElem) < LPP(b->theRingElem));
  throw RuntimeException("RINGELEM can be compared only if in ordered domain or if power-products", beginOperatorSourcePosition, endOperatorSourcePosition);
}


intrusive_ptr<RightValue> opLessOrEqual_RingElem_RingElem(RuntimeEnvironment *, const intrusive_ptr<const RightValue> leftOp, const intrusive_ptr<const RightValue> rightOp, const CharPointer &beginOperatorSourcePosition, const CharPointer &endOperatorSourcePosition) {
	intrusive_ptr<const RingElemValue> a = intrusive_ptr_cast<const RingElemValue>(leftOp);
	intrusive_ptr<const RingElemValue> b = intrusive_ptr_cast<const RingElemValue>(rightOp);
  if (IsOrderedDomain(owner(a->theRingElem)))
    return BoolValue::fromBool(a->theRingElem <= b->theRingElem);
  if (IsMonomial(a->theRingElem) && IsMonomial(b->theRingElem)
      && IsOne(LC(a->theRingElem)) && IsOne(LC(b->theRingElem)))
    return BoolValue::fromBool(LPP(a->theRingElem) <= LPP(b->theRingElem));
  throw RuntimeException("RINGELEM can be compared only if in ordered domain or if power-products", beginOperatorSourcePosition, endOperatorSourcePosition);
}


intrusive_ptr<RightValue> opGreaterThan_RingElem_RingElem(RuntimeEnvironment *rv, const intrusive_ptr<const RightValue> leftOp, const intrusive_ptr<const RightValue> rightOp, const CharPointer &beginOperatorSourcePosition, const CharPointer &endOperatorSourcePosition) {
  return opLessThan_RingElem_RingElem(rv, rightOp, leftOp, beginOperatorSourcePosition, endOperatorSourcePosition);
}


intrusive_ptr<RightValue> opGreaterOrEqual_RingElem_RingElem(RuntimeEnvironment *rv, const intrusive_ptr<const RightValue> leftOp, const intrusive_ptr<const RightValue> rightOp, const CharPointer &beginOperatorSourcePosition, const CharPointer &endOperatorSourcePosition) {
  return opLessOrEqual_RingElem_RingElem(rv, rightOp, leftOp, beginOperatorSourcePosition, endOperatorSourcePosition);
}

//----- opPower_ ---------------------------------------------------------
intrusive_ptr<RightValue> opPower_RingElem_BigInt(RuntimeEnvironment *, const intrusive_ptr<const RightValue> leftOp, const intrusive_ptr<const RightValue> rightOp, const CharPointer &, const CharPointer &) {
	intrusive_ptr<const BigIntValue> exponent = intrusive_ptr_cast<const BigIntValue>(rightOp);
	return new RingElemValue(power(intrusive_ptr_cast<const RingElemValue>(leftOp)->theRingElem, exponent->theInteger));
}

intrusive_ptr<RightValue> opPower_BigInt_BigInt(RuntimeEnvironment *, const intrusive_ptr<const RightValue> leftOp, const intrusive_ptr<const RightValue> rightOp, const CharPointer &, const CharPointer &) {
	intrusive_ptr<const BigIntValue> exponent = intrusive_ptr_cast<const BigIntValue>(rightOp);
	if (exponent->theInteger<0)
          return Value::simplifiedValueFrom(power(BigRat(intrusive_ptr_cast<const BigIntValue>(leftOp)->theInteger,1), exponent->theInteger));
	return new BigIntValue(power(intrusive_ptr_cast<const BigIntValue>(leftOp)->theInteger, exponent->theInteger));
}

intrusive_ptr<RightValue> opPower_BigRat_BigInt(RuntimeEnvironment *, const intrusive_ptr<const RightValue> leftOp, const intrusive_ptr<const RightValue> rightOp, const CharPointer &, const CharPointer &) {
	return new BigRatValue(power(intrusive_ptr_cast<const BigRatValue>(leftOp)->theRational, intrusive_ptr_cast<const BigIntValue>(rightOp)->theInteger));
}

intrusive_ptr<RightValue> opPower_Matrix_BigInt(RuntimeEnvironment *, const intrusive_ptr<const RightValue> leftOp, const intrusive_ptr<const RightValue> rightOp, const CharPointer &, const CharPointer &) {
	intrusive_ptr<const BigIntValue> exponent = intrusive_ptr_cast<const BigIntValue>(rightOp);
	return Value::from(power(intrusive_ptr_cast<const MatrixValue>(leftOp)->theMatrix, exponent->theInteger));
}

intrusive_ptr<RightValue> opPower_Ideal_BigInt(RuntimeEnvironment *, const intrusive_ptr<const RightValue> leftOp, const intrusive_ptr<const RightValue> rightOp, const CharPointer &, const CharPointer &) {
	intrusive_ptr<const BigIntValue> exponent = intrusive_ptr_cast<const BigIntValue>(rightOp);
	return Value::from(power(intrusive_ptr_cast<const IdealValue>(leftOp)->theIdeal, exponent->theInteger));
}

// intrusive_ptr<RightValue> opPower_Ring_BigInt(RuntimeEnvironment *, const intrusive_ptr<const RightValue> leftOp, const intrusive_ptr<const RightValue> rightOp, const CharPointer &beginOperatorSourcePosition, const CharPointer &endOperatorSourcePosition) {
// 	intrusive_ptr<const BigIntValue> exponent = intrusive_ptr_cast<const BigIntValue>(rightOp);
//   long n;
//   if (!IsConvertible(n, exponent->theInteger) || n<0)
//     throw RuntimeException("Function values cannot be compared", beginOperatorSourcePosition, endOperatorSourcePosition);
// 	return Value::from(NewFreeModule(intrusive_ptr_cast<const RingValue>(leftOp)->theRing, n));
// }

//----- opSlash_ ---------------------------------------------------------

intrusive_ptr<RightValue> opSlash_BigInt_BigInt(RuntimeEnvironment *, const intrusive_ptr<const RightValue> leftOp, const intrusive_ptr<const RightValue> rightOp, const CharPointer &, const CharPointer &) {
	return Value::simplifiedValueFrom(BigRat(intrusive_ptr_cast<const BigIntValue>(leftOp)->theInteger, intrusive_ptr_cast<const BigIntValue>(rightOp)->theInteger));
}

intrusive_ptr<RightValue> opSlash_Ring_Ideal(RuntimeEnvironment *, const intrusive_ptr<const RightValue> leftOp, const intrusive_ptr<const RightValue> rightOp, const CharPointer &, const CharPointer &) {
	return Value::from(NewQuotientRing(intrusive_ptr_cast<const RingValue>(leftOp)->theRing, intrusive_ptr_cast<const IdealValue>(rightOp)->theIdeal));
}

intrusive_ptr<RightValue> opEqual_Function_Function(RuntimeEnvironment *, const intrusive_ptr<const RightValue>, const intrusive_ptr<const RightValue>, const CharPointer &beginOperatorSourcePosition, const CharPointer &endOperatorSourcePosition) {
	throw RuntimeException("Function values cannot be compared", beginOperatorSourcePosition, endOperatorSourcePosition);
}

intrusive_ptr<RightValue> opNotEqual_Function_Function(RuntimeEnvironment *runtimeEnv, const intrusive_ptr<const RightValue> leftOp, const intrusive_ptr<const RightValue> rightOp, const CharPointer &beginOperatorSourcePosition, const CharPointer &endOperatorSourcePosition) {
	return BoolValue::fromBool(!intrusive_ptr_cast<BoolValue>(opEqual_Function_Function(runtimeEnv, leftOp, rightOp, beginOperatorSourcePosition, endOperatorSourcePosition))->theBool);
}

intrusive_ptr<RightValue> opEqual_List_List(RuntimeEnvironment *runtimeEnv, const intrusive_ptr<const RightValue> leftOp, const intrusive_ptr<const RightValue> rightOp, const CharPointer &beginOperatorSourcePosition, const CharPointer &endOperatorSourcePosition) {
	intrusive_ptr<const ListValue> l1 = intrusive_ptr_cast<const ListValue>(leftOp);
	intrusive_ptr<const ListValue> l2 = intrusive_ptr_cast<const ListValue>(rightOp);
	const ListValue::ContainerType::size_type size = l1->size();
	if (size!=l2->size())
		return BoolValue::falseValue;
	for(ListValue::ContainerType::size_type a=0; a<size; ++a) {
		intrusive_ptr<BoolValue> b = intrusive_ptr_cast<BoolValue>(runtimeEnv->binaryOperatorDispatch(l1->getValue(a), l2->getValue(a), RuntimeEnvironment::opEqualMap, beginOperatorSourcePosition, endOperatorSourcePosition));
		if (!b->theBool)
			return b;
	}
	return BoolValue::trueValue;
}

intrusive_ptr<RightValue> opPlus_List_List(RuntimeEnvironment *runtimeEnv, const intrusive_ptr<const RightValue> leftOp, const intrusive_ptr<const RightValue> rightOp, const CharPointer &beginOperatorSourcePosition, const CharPointer &endOperatorSourcePosition) {
	intrusive_ptr<const ListValue> l1 = intrusive_ptr_cast<const ListValue>(leftOp);
	intrusive_ptr<const ListValue> l2 = intrusive_ptr_cast<const ListValue>(rightOp);
	intrusive_ptr<ListValue> result = new ListValue();
	const ListValue::ContainerType::size_type size = l1->size();
	if (size!=l2->size())
		throw RuntimeException("Cannot sum lists of different sizes", beginOperatorSourcePosition, endOperatorSourcePosition);
	for(ListValue::ContainerType::size_type a=0; a<size; ++a)
		result->addValue(runtimeEnv->binaryOperatorDispatch(
								l1->getValue(a),
								l2->getValue(a),
								RuntimeEnvironment::opPlusMap,
								beginOperatorSourcePosition,
								endOperatorSourcePosition));
	return result;
}

intrusive_ptr<RightValue> opMinus_List_List(RuntimeEnvironment *runtimeEnv, const intrusive_ptr<const RightValue> leftOp, const intrusive_ptr<const RightValue> rightOp, const CharPointer &beginOperatorSourcePosition, const CharPointer &endOperatorSourcePosition) {
	intrusive_ptr<const ListValue> l1 = intrusive_ptr_cast<const ListValue>(leftOp);
	intrusive_ptr<const ListValue> l2 = intrusive_ptr_cast<const ListValue>(rightOp);
	intrusive_ptr<ListValue> result = new ListValue();
	const ListValue::ContainerType::size_type size = l1->size();
	if (size!=l2->size())
		throw RuntimeException("Cannot subtract lists of different sizes", beginOperatorSourcePosition, endOperatorSourcePosition);
	for(ListValue::ContainerType::size_type a=0; a<size; ++a)
		result->addValue(runtimeEnv->binaryOperatorDispatch(
								l1->getValue(a),
								l2->getValue(a),
								RuntimeEnvironment::opMinusMap,
								beginOperatorSourcePosition,
								endOperatorSourcePosition));
	return result;
}

intrusive_ptr<RightValue> opNotEqual_List_List(RuntimeEnvironment *runtimeEnv, const intrusive_ptr<const RightValue> leftOp, const intrusive_ptr<const RightValue> rightOp, const CharPointer &beginOperatorSourcePosition, const CharPointer &endOperatorSourcePosition) {
	return BoolValue::fromBool(!intrusive_ptr_cast<BoolValue>(opEqual_List_List(runtimeEnv, leftOp, rightOp, beginOperatorSourcePosition, endOperatorSourcePosition))->theBool);
}

intrusive_ptr<RightValue> opEqual_Record_Record(RuntimeEnvironment *runtimeEnv, const intrusive_ptr<const RightValue> leftOp, const intrusive_ptr<const RightValue> rightOp, const CharPointer &beginOperatorSourcePosition, const CharPointer &endOperatorSourcePosition) {
	intrusive_ptr<const RecordValue> r1 = intrusive_ptr_cast<const RecordValue>(leftOp);
	intrusive_ptr<const RecordValue> r2 = intrusive_ptr_cast<const RecordValue>(rightOp);
	const RecordValue::MapType::size_type numberOfFields = r1->numberOfFields();
	if (numberOfFields!=r2->numberOfFields())
		return BoolValue::falseValue;
	RecordValue::MapType::const_iterator end = r1->end();
	int a=0;
	for(RecordValue::MapType::const_iterator begin = r1->begin(); begin!=end; ++begin, ++a) {
		intrusive_ptr<const RightValue> v2 = r2->getField((*begin).first);
		if (!v2)
			return BoolValue::falseValue;
		intrusive_ptr<BoolValue> b = intrusive_ptr_cast<BoolValue>(runtimeEnv->binaryOperatorDispatch((*begin).second, v2, RuntimeEnvironment::opEqualMap, beginOperatorSourcePosition, endOperatorSourcePosition));
		if (!b->theBool)
			return b;
	}
	return BoolValue::trueValue;
}

intrusive_ptr<RightValue> opNotEqual_Record_Record(RuntimeEnvironment *runtimeEnv, const intrusive_ptr<const RightValue> leftOp, const intrusive_ptr<const RightValue> rightOp, const CharPointer &beginOperatorSourcePosition, const CharPointer &endOperatorSourcePosition) {
	return BoolValue::fromBool(!intrusive_ptr_cast<BoolValue>(opEqual_Record_Record(runtimeEnv, leftOp, rightOp, beginOperatorSourcePosition, endOperatorSourcePosition))->theBool);
}

intrusive_ptr<RightValue> opStar_BigIntBigRatRE_List(RuntimeEnvironment *runtimeEnv, const intrusive_ptr<const RightValue> leftOp, const intrusive_ptr<const RightValue> rightOp, const CharPointer &beginOperatorSourcePosition, const CharPointer &endOperatorSourcePosition) {
	intrusive_ptr<const ListValue> list = intrusive_ptr_cast<const ListValue>(rightOp);
	intrusive_ptr<ListValue> result = new ListValue();
	const ListValue::ContainerType::size_type size = list->size();
	for(ListValue::ContainerType::size_type a=0; a<size; ++a)
		result->addValue(runtimeEnv->binaryOperatorDispatch(
								leftOp,
								list->getValue(a),
								RuntimeEnvironment::opStarMap,
								beginOperatorSourcePosition,
								endOperatorSourcePosition));
	return result;
}

intrusive_ptr<RightValue> opStar_List_BigIntBigRatRE(RuntimeEnvironment *runtimeEnv, const intrusive_ptr<const RightValue> leftOp, const intrusive_ptr<const RightValue> rightOp, const CharPointer &beginOperatorSourcePosition, const CharPointer &endOperatorSourcePosition) {
	intrusive_ptr<const ListValue> list = intrusive_ptr_cast<const ListValue>(leftOp);
	intrusive_ptr<ListValue> result = new ListValue();
	const ListValue::ContainerType::size_type size = list->size();
	for(ListValue::ContainerType::size_type a=0; a<size; ++a)
		result->addValue(runtimeEnv->binaryOperatorDispatch(
								list->getValue(a),
								rightOp,
								RuntimeEnvironment::opStarMap,
								beginOperatorSourcePosition,
								endOperatorSourcePosition));
	return result;
}

intrusive_ptr<RightValue> opSlash_List_BigIntBigRatRE(RuntimeEnvironment *runtimeEnv, const intrusive_ptr<const RightValue> leftOp, const intrusive_ptr<const RightValue> rightOp, const CharPointer &beginOperatorSourcePosition, const CharPointer &endOperatorSourcePosition) {
	intrusive_ptr<const ListValue> list = intrusive_ptr_cast<const ListValue>(leftOp);
	intrusive_ptr<ListValue> result = new ListValue();
	const ListValue::ContainerType::size_type size = list->size();
	for(ListValue::ContainerType::size_type a=0; a<size; ++a)
		result->addValue(runtimeEnv->binaryOperatorDispatch(
								list->getValue(a),
								rightOp,
								RuntimeEnvironment::opSlashMap,
								beginOperatorSourcePosition,
								endOperatorSourcePosition));
	return result;
}


intrusive_ptr<RightValue> opColon_Ideal_Ideal(RuntimeEnvironment * /*runtimeEnv*/, const intrusive_ptr<const RightValue> leftOp, const intrusive_ptr<const RightValue> rightOp, const CharPointer & /*beginOperatorSourcePosition*/, const CharPointer & /*endOperatorSourcePosition*/) {
	const intrusive_ptr<const IdealValue> l1 = intrusive_ptr_cast<const IdealValue>(leftOp);
	const intrusive_ptr<const IdealValue> l2 = intrusive_ptr_cast<const IdealValue>(rightOp);
	return Value::from(colon(l1->theIdeal, l2->theIdeal));
}

RuntimeEnvironment::DispatchMapType RuntimeEnvironment::opPlusMap;
RuntimeEnvironment::DispatchMapType RuntimeEnvironment::opMinusMap;
RuntimeEnvironment::DispatchMapType RuntimeEnvironment::opStarMap;
RuntimeEnvironment::DispatchMapType RuntimeEnvironment::opSlashMap;
RuntimeEnvironment::DispatchMapType RuntimeEnvironment::opLessThanMap;
RuntimeEnvironment::DispatchMapType RuntimeEnvironment::opLessOrEqualMap;
RuntimeEnvironment::DispatchMapType RuntimeEnvironment::opEqualMap;
RuntimeEnvironment::DispatchMapType RuntimeEnvironment::opNotEqualMap;
RuntimeEnvironment::DispatchMapType RuntimeEnvironment::opGreaterOrEqualMap;
RuntimeEnvironment::DispatchMapType RuntimeEnvironment::opGreaterThanMap;
RuntimeEnvironment::DispatchMapType RuntimeEnvironment::opPowerMap;
RuntimeEnvironment::DispatchMapType RuntimeEnvironment::opModMap;
RuntimeEnvironment::DispatchMapType RuntimeEnvironment::opColonMap;

void RuntimeEnvironment::initMaps() {
	static bool initialized=false;
	if (initialized)
		return;
	initialized = true;

  //------ STRING ----------------------------------------
	opPlusMap[TypeValue::DISPATCH_INDEX_OF_STRING][TypeValue::DISPATCH_INDEX_OF_STRING] = opPlus<StringValue, StringValue>;
	opLessThanMap[TypeValue::DISPATCH_INDEX_OF_STRING][TypeValue::DISPATCH_INDEX_OF_STRING] = opLessThan<StringValue, StringValue>;
	opLessOrEqualMap[TypeValue::DISPATCH_INDEX_OF_STRING][TypeValue::DISPATCH_INDEX_OF_STRING] = opLessOrEqual<StringValue, StringValue>;
	opEqualMap[TypeValue::DISPATCH_INDEX_OF_STRING][TypeValue::DISPATCH_INDEX_OF_STRING] = opEqual<StringValue, StringValue>;
	opNotEqualMap[TypeValue::DISPATCH_INDEX_OF_STRING][TypeValue::DISPATCH_INDEX_OF_STRING] = opNotEqual<StringValue, StringValue>;
	opGreaterOrEqualMap[TypeValue::DISPATCH_INDEX_OF_STRING][TypeValue::DISPATCH_INDEX_OF_STRING] = opGreaterOrEqual<StringValue, StringValue>;
	opGreaterThanMap[TypeValue::DISPATCH_INDEX_OF_STRING][TypeValue::DISPATCH_INDEX_OF_STRING] = opGreaterThan<StringValue, StringValue>;

  //------ BIGINT ----------------------------------------
	opPlusMap[TypeValue::DISPATCH_INDEX_OF_BIGINT][TypeValue::DISPATCH_INDEX_OF_BIGINT] = opPlus<BigIntValue, BigIntValue>;
	opMinusMap[TypeValue::DISPATCH_INDEX_OF_BIGINT][TypeValue::DISPATCH_INDEX_OF_BIGINT] = opMinus<BigIntValue, BigIntValue>;
	opStarMap[TypeValue::DISPATCH_INDEX_OF_BIGINT][TypeValue::DISPATCH_INDEX_OF_BIGINT] = opStar<BigIntValue, BigIntValue>;
	opSlashMap[TypeValue::DISPATCH_INDEX_OF_BIGINT][TypeValue::DISPATCH_INDEX_OF_BIGINT] = opSlash_BigInt_BigInt;
	opLessThanMap[TypeValue::DISPATCH_INDEX_OF_BIGINT][TypeValue::DISPATCH_INDEX_OF_BIGINT] = opLessThan<BigIntValue, BigIntValue>;
	opLessOrEqualMap[TypeValue::DISPATCH_INDEX_OF_BIGINT][TypeValue::DISPATCH_INDEX_OF_BIGINT] = opLessOrEqual<BigIntValue, BigIntValue>;
	opEqualMap[TypeValue::DISPATCH_INDEX_OF_BIGINT][TypeValue::DISPATCH_INDEX_OF_BIGINT] = opEqual<BigIntValue, BigIntValue>;
	opNotEqualMap[TypeValue::DISPATCH_INDEX_OF_BIGINT][TypeValue::DISPATCH_INDEX_OF_BIGINT] = opNotEqual<BigIntValue, BigIntValue>;
	opGreaterOrEqualMap[TypeValue::DISPATCH_INDEX_OF_BIGINT][TypeValue::DISPATCH_INDEX_OF_BIGINT] = opGreaterOrEqual<BigIntValue, BigIntValue>;
	opGreaterThanMap[TypeValue::DISPATCH_INDEX_OF_BIGINT][TypeValue::DISPATCH_INDEX_OF_BIGINT] = opGreaterThan<BigIntValue, BigIntValue>;
	opPowerMap[TypeValue::DISPATCH_INDEX_OF_BIGINT][TypeValue::DISPATCH_INDEX_OF_BIGINT] = opPower_BigInt_BigInt;

  //------ BIGRAT ----------------------------------------
	opPlusMap[TypeValue::DISPATCH_INDEX_OF_BIGRAT][TypeValue::DISPATCH_INDEX_OF_BIGRAT] = opPlus<BigRatValue, BigRatValue>;
	opMinusMap[TypeValue::DISPATCH_INDEX_OF_BIGRAT][TypeValue::DISPATCH_INDEX_OF_BIGRAT] = opMinus<BigRatValue, BigRatValue>;
	opStarMap[TypeValue::DISPATCH_INDEX_OF_BIGRAT][TypeValue::DISPATCH_INDEX_OF_BIGRAT] = opStar<BigRatValue, BigRatValue>;
	opSlashMap[TypeValue::DISPATCH_INDEX_OF_BIGRAT][TypeValue::DISPATCH_INDEX_OF_BIGRAT] = opSlash<BigRatValue, BigRatValue>;
	opLessThanMap[TypeValue::DISPATCH_INDEX_OF_BIGRAT][TypeValue::DISPATCH_INDEX_OF_BIGRAT] = opLessThan<BigRatValue, BigRatValue>;
	opLessOrEqualMap[TypeValue::DISPATCH_INDEX_OF_BIGRAT][TypeValue::DISPATCH_INDEX_OF_BIGRAT] = opLessOrEqual<BigRatValue, BigRatValue>;
	opEqualMap[TypeValue::DISPATCH_INDEX_OF_BIGRAT][TypeValue::DISPATCH_INDEX_OF_BIGRAT] = opEqual<BigRatValue, BigRatValue>;
	opNotEqualMap[TypeValue::DISPATCH_INDEX_OF_BIGRAT][TypeValue::DISPATCH_INDEX_OF_BIGRAT] = opNotEqual<BigRatValue, BigRatValue>;
	opGreaterOrEqualMap[TypeValue::DISPATCH_INDEX_OF_BIGRAT][TypeValue::DISPATCH_INDEX_OF_BIGRAT] = opGreaterOrEqual<BigRatValue, BigRatValue>;
	opGreaterThanMap[TypeValue::DISPATCH_INDEX_OF_BIGRAT][TypeValue::DISPATCH_INDEX_OF_BIGRAT] = opGreaterThan<BigRatValue, BigRatValue>;
	opPowerMap[TypeValue::DISPATCH_INDEX_OF_BIGRAT][TypeValue::DISPATCH_INDEX_OF_BIGINT] = opPower_BigRat_BigInt;

	opEqualMap[TypeValue::DISPATCH_INDEX_OF_FUNCTION][TypeValue::DISPATCH_INDEX_OF_FUNCTION] = opEqual_Function_Function;
	opNotEqualMap[TypeValue::DISPATCH_INDEX_OF_FUNCTION][TypeValue::DISPATCH_INDEX_OF_FUNCTION] = opNotEqual_Function_Function;

  //------ LIST ----------------------------------------
	opEqualMap[TypeValue::DISPATCH_INDEX_OF_LIST][TypeValue::DISPATCH_INDEX_OF_LIST] = opEqual_List_List;
	opNotEqualMap[TypeValue::DISPATCH_INDEX_OF_LIST][TypeValue::DISPATCH_INDEX_OF_LIST] = opNotEqual_List_List;
	opStarMap[TypeValue::DISPATCH_INDEX_OF_BIGINT][TypeValue::DISPATCH_INDEX_OF_LIST] = opStar_BigIntBigRatRE_List;
	opStarMap[TypeValue::DISPATCH_INDEX_OF_BIGRAT][TypeValue::DISPATCH_INDEX_OF_LIST] = opStar_BigIntBigRatRE_List;
	opStarMap[TypeValue::DISPATCH_INDEX_OF_RINGELEM][TypeValue::DISPATCH_INDEX_OF_LIST] = opStar_BigIntBigRatRE_List;
	opStarMap[TypeValue::DISPATCH_INDEX_OF_LIST][TypeValue::DISPATCH_INDEX_OF_BIGINT] = opStar_List_BigIntBigRatRE;
	opStarMap[TypeValue::DISPATCH_INDEX_OF_LIST][TypeValue::DISPATCH_INDEX_OF_BIGRAT] = opStar_List_BigIntBigRatRE;
	opStarMap[TypeValue::DISPATCH_INDEX_OF_LIST][TypeValue::DISPATCH_INDEX_OF_RINGELEM] = opStar_List_BigIntBigRatRE;
	opSlashMap[TypeValue::DISPATCH_INDEX_OF_LIST][TypeValue::DISPATCH_INDEX_OF_BIGINT] = opSlash_List_BigIntBigRatRE;
	opSlashMap[TypeValue::DISPATCH_INDEX_OF_LIST][TypeValue::DISPATCH_INDEX_OF_BIGRAT] = opSlash_List_BigIntBigRatRE;
	opSlashMap[TypeValue::DISPATCH_INDEX_OF_LIST][TypeValue::DISPATCH_INDEX_OF_RINGELEM] = opSlash_List_BigIntBigRatRE;
	opPlusMap[TypeValue::DISPATCH_INDEX_OF_LIST][TypeValue::DISPATCH_INDEX_OF_LIST] = opPlus_List_List;
	opMinusMap[TypeValue::DISPATCH_INDEX_OF_LIST][TypeValue::DISPATCH_INDEX_OF_LIST] = opMinus_List_List;

  //------ MAT ----------------------------------------
	opStarMap[TypeValue::DISPATCH_INDEX_OF_MAT][TypeValue::DISPATCH_INDEX_OF_MAT] = opStar<MatrixValue, MatrixValue>;
	opStarMap[TypeValue::DISPATCH_INDEX_OF_BIGINT][TypeValue::DISPATCH_INDEX_OF_MAT] = opStar<BigIntValue, MatrixValue>;
	opStarMap[TypeValue::DISPATCH_INDEX_OF_BIGRAT][TypeValue::DISPATCH_INDEX_OF_MAT] = opStar<BigRatValue, MatrixValue>;
	opStarMap[TypeValue::DISPATCH_INDEX_OF_RINGELEM][TypeValue::DISPATCH_INDEX_OF_MAT] = opStar<RingElemValue, MatrixValue>;
	opStarMap[TypeValue::DISPATCH_INDEX_OF_MAT][TypeValue::DISPATCH_INDEX_OF_BIGINT] = opStar<MatrixValue, BigIntValue>;
	opStarMap[TypeValue::DISPATCH_INDEX_OF_MAT][TypeValue::DISPATCH_INDEX_OF_BIGRAT] = opStar<MatrixValue, BigRatValue>;
	opStarMap[TypeValue::DISPATCH_INDEX_OF_MAT][TypeValue::DISPATCH_INDEX_OF_RINGELEM] = opStar<MatrixValue, RingElemValue>;
	opPlusMap[TypeValue::DISPATCH_INDEX_OF_MAT][TypeValue::DISPATCH_INDEX_OF_MAT] = opPlus<MatrixValue, MatrixValue>;
	opMinusMap[TypeValue::DISPATCH_INDEX_OF_MAT][TypeValue::DISPATCH_INDEX_OF_MAT] = opMinus<MatrixValue, MatrixValue>;
	opEqualMap[TypeValue::DISPATCH_INDEX_OF_MAT][TypeValue::DISPATCH_INDEX_OF_MAT] = opEqual<MatrixValue, MatrixValue>;
	opNotEqualMap[TypeValue::DISPATCH_INDEX_OF_MAT][TypeValue::DISPATCH_INDEX_OF_MAT] = opNotEqual<MatrixValue, MatrixValue>;
	opPowerMap[TypeValue::DISPATCH_INDEX_OF_MAT][TypeValue::DISPATCH_INDEX_OF_BIGINT] = opPower_Matrix_BigInt;
	opSlashMap[TypeValue::DISPATCH_INDEX_OF_MAT][TypeValue::DISPATCH_INDEX_OF_BIGINT] = opSlash<MatrixValue, BigIntValue>;
	opSlashMap[TypeValue::DISPATCH_INDEX_OF_MAT][TypeValue::DISPATCH_INDEX_OF_BIGRAT] = opSlash<MatrixValue, BigRatValue>;
	opSlashMap[TypeValue::DISPATCH_INDEX_OF_MAT][TypeValue::DISPATCH_INDEX_OF_RINGELEM] = opSlash<MatrixValue, RingElemValue>;

  //------ RECORD ----------------------------------------
	opEqualMap[TypeValue::DISPATCH_INDEX_OF_RECORD][TypeValue::DISPATCH_INDEX_OF_RECORD] = opEqual_Record_Record;
	opNotEqualMap[TypeValue::DISPATCH_INDEX_OF_RECORD][TypeValue::DISPATCH_INDEX_OF_RECORD] = opNotEqual_Record_Record;

  //------ BOOL ----------------------------------------
	opEqualMap[TypeValue::DISPATCH_INDEX_OF_BOOL][TypeValue::DISPATCH_INDEX_OF_BOOL] = opEqual<BoolValue, BoolValue>;
	opNotEqualMap[TypeValue::DISPATCH_INDEX_OF_BOOL][TypeValue::DISPATCH_INDEX_OF_BOOL] = opNotEqual<BoolValue, BoolValue>;

  //------ TYPE ----------------------------------------
	opEqualMap[TypeValue::DISPATCH_INDEX_OF_TYPE][TypeValue::DISPATCH_INDEX_OF_TYPE] = opEqual<TypeValue, TypeValue>;
	opNotEqualMap[TypeValue::DISPATCH_INDEX_OF_TYPE][TypeValue::DISPATCH_INDEX_OF_TYPE] = opNotEqual<TypeValue, TypeValue>;

  //------ RINGELEM ----------------------------------------
	opPlusMap[TypeValue::DISPATCH_INDEX_OF_RINGELEM][TypeValue::DISPATCH_INDEX_OF_RINGELEM] = opPlus<RingElemValue, RingElemValue>;
	opMinusMap[TypeValue::DISPATCH_INDEX_OF_RINGELEM][TypeValue::DISPATCH_INDEX_OF_RINGELEM] = opMinus<RingElemValue, RingElemValue>;
	opStarMap[TypeValue::DISPATCH_INDEX_OF_RINGELEM][TypeValue::DISPATCH_INDEX_OF_RINGELEM] = opStar<RingElemValue, RingElemValue>;
	opSlashMap[TypeValue::DISPATCH_INDEX_OF_RINGELEM][TypeValue::DISPATCH_INDEX_OF_RINGELEM] = opSlash<RingElemValue, RingElemValue>;
  //	opLessThanMap[TypeValue::DISPATCH_INDEX_OF_RINGELEM][TypeValue::DISPATCH_INDEX_OF_RINGELEM] = opLessThan<RingElemValue, RingElemValue>;
  //	opLessOrEqualMap[TypeValue::DISPATCH_INDEX_OF_RINGELEM][TypeValue::DISPATCH_INDEX_OF_RINGELEM] = opLessOrEqual<RingElemValue, RingElemValue>;
	opLessThanMap[TypeValue::DISPATCH_INDEX_OF_RINGELEM][TypeValue::DISPATCH_INDEX_OF_RINGELEM] = opLessThan_RingElem_RingElem;
	opLessOrEqualMap[TypeValue::DISPATCH_INDEX_OF_RINGELEM][TypeValue::DISPATCH_INDEX_OF_RINGELEM] = opLessOrEqual_RingElem_RingElem;
	opEqualMap[TypeValue::DISPATCH_INDEX_OF_RINGELEM][TypeValue::DISPATCH_INDEX_OF_RINGELEM] = opEqual<RingElemValue, RingElemValue>;
	opNotEqualMap[TypeValue::DISPATCH_INDEX_OF_RINGELEM][TypeValue::DISPATCH_INDEX_OF_RINGELEM] = opNotEqual<RingElemValue, RingElemValue>;
  //	opGreaterOrEqualMap[TypeValue::DISPATCH_INDEX_OF_RINGELEM][TypeValue::DISPATCH_INDEX_OF_RINGELEM] = opGreaterOrEqual<RingElemValue, RingElemValue>;
  //	opGreaterThanMap[TypeValue::DISPATCH_INDEX_OF_RINGELEM][TypeValue::DISPATCH_INDEX_OF_RINGELEM] = opGreaterThan<RingElemValue, RingElemValue>;
	opGreaterOrEqualMap[TypeValue::DISPATCH_INDEX_OF_RINGELEM][TypeValue::DISPATCH_INDEX_OF_RINGELEM] = opGreaterOrEqual_RingElem_RingElem;
	opGreaterThanMap[TypeValue::DISPATCH_INDEX_OF_RINGELEM][TypeValue::DISPATCH_INDEX_OF_RINGELEM] = opGreaterThan_RingElem_RingElem;
	opPowerMap[TypeValue::DISPATCH_INDEX_OF_RINGELEM][TypeValue::DISPATCH_INDEX_OF_BIGINT] = opPower_RingElem_BigInt;

  //------ MODULEELEM ----------------------------------------
	opStarMap[TypeValue::DISPATCH_INDEX_OF_BIGINT][TypeValue::DISPATCH_INDEX_OF_MODULEELEM] = opStar<BigIntValue, ModuleElemValue>;
	opStarMap[TypeValue::DISPATCH_INDEX_OF_BIGRAT][TypeValue::DISPATCH_INDEX_OF_MODULEELEM] = opStar<BigRatValue, ModuleElemValue>;
	opStarMap[TypeValue::DISPATCH_INDEX_OF_RINGELEM][TypeValue::DISPATCH_INDEX_OF_MODULEELEM] = opStar<RingElemValue, ModuleElemValue>;
	opStarMap[TypeValue::DISPATCH_INDEX_OF_MODULEELEM][TypeValue::DISPATCH_INDEX_OF_BIGINT] = opStar<ModuleElemValue, BigIntValue>;
	opStarMap[TypeValue::DISPATCH_INDEX_OF_MODULEELEM][TypeValue::DISPATCH_INDEX_OF_BIGRAT] = opStar<ModuleElemValue, BigRatValue>;
	opStarMap[TypeValue::DISPATCH_INDEX_OF_MODULEELEM][TypeValue::DISPATCH_INDEX_OF_RINGELEM] = opStar<ModuleElemValue, RingElemValue>;
	opPlusMap[TypeValue::DISPATCH_INDEX_OF_MODULEELEM][TypeValue::DISPATCH_INDEX_OF_MODULEELEM] = opPlus<ModuleElemValue, ModuleElemValue>;
	opMinusMap[TypeValue::DISPATCH_INDEX_OF_MODULEELEM][TypeValue::DISPATCH_INDEX_OF_MODULEELEM] = opMinus<ModuleElemValue, ModuleElemValue>;
	opEqualMap[TypeValue::DISPATCH_INDEX_OF_MODULEELEM][TypeValue::DISPATCH_INDEX_OF_MODULEELEM] = opEqual<ModuleElemValue, ModuleElemValue>;
	opNotEqualMap[TypeValue::DISPATCH_INDEX_OF_MODULEELEM][TypeValue::DISPATCH_INDEX_OF_MODULEELEM] = opNotEqual<ModuleElemValue, ModuleElemValue>;
	opSlashMap[TypeValue::DISPATCH_INDEX_OF_MODULEELEM][TypeValue::DISPATCH_INDEX_OF_BIGINT] = opSlash<ModuleElemValue, BigIntValue>;
	opSlashMap[TypeValue::DISPATCH_INDEX_OF_MODULEELEM][TypeValue::DISPATCH_INDEX_OF_BIGRAT] = opSlash<ModuleElemValue, BigRatValue>;
	opSlashMap[TypeValue::DISPATCH_INDEX_OF_MODULEELEM][TypeValue::DISPATCH_INDEX_OF_RINGELEM] = opSlash<ModuleElemValue, RingElemValue>;

  //------ RING ----------------------------------------
	opEqualMap[TypeValue::DISPATCH_INDEX_OF_RING][TypeValue::DISPATCH_INDEX_OF_RING] = opEqual<RingValue, RingValue>;
	opNotEqualMap[TypeValue::DISPATCH_INDEX_OF_RING][TypeValue::DISPATCH_INDEX_OF_RING] = opNotEqual<RingValue, RingValue>;
	opSlashMap[TypeValue::DISPATCH_INDEX_OF_RING][TypeValue::DISPATCH_INDEX_OF_IDEAL] = opSlash_Ring_Ideal;
  //	opPowerMap[TypeValue::DISPATCH_INDEX_OF_RING][TypeValue::DISPATCH_INDEX_OF_BIGINT] = opPower_Ring_BigInt;

  //------ MODULE ----------------------------------------
	opEqualMap[TypeValue::DISPATCH_INDEX_OF_MODULE][TypeValue::DISPATCH_INDEX_OF_MODULE] = opEqual<ModuleValue, ModuleValue>;
	opNotEqualMap[TypeValue::DISPATCH_INDEX_OF_MODULE][TypeValue::DISPATCH_INDEX_OF_MODULE] = opNotEqual<ModuleValue, ModuleValue>;
  //	opSlashMap[TypeValue::DISPATCH_INDEX_OF_RING][TypeValue::DISPATCH_INDEX_OF_IDEAL] = opSlash_Ring_Ideal;

  //------ IDEAL ----------------------------------------
	opPlusMap[TypeValue::DISPATCH_INDEX_OF_IDEAL][TypeValue::DISPATCH_INDEX_OF_IDEAL] = opPlus<IdealValue, IdealValue>;
	opStarMap[TypeValue::DISPATCH_INDEX_OF_IDEAL][TypeValue::DISPATCH_INDEX_OF_RINGELEM] = opStar<IdealValue, RingElemValue>;
	opStarMap[TypeValue::DISPATCH_INDEX_OF_RINGELEM][TypeValue::DISPATCH_INDEX_OF_IDEAL] = opStar<RingElemValue, IdealValue>;
	opStarMap[TypeValue::DISPATCH_INDEX_OF_IDEAL][TypeValue::DISPATCH_INDEX_OF_IDEAL] = opStar<IdealValue, IdealValue>;
	opPowerMap[TypeValue::DISPATCH_INDEX_OF_IDEAL][TypeValue::DISPATCH_INDEX_OF_BIGINT] = opPower_Ideal_BigInt;
	opEqualMap[TypeValue::DISPATCH_INDEX_OF_IDEAL][TypeValue::DISPATCH_INDEX_OF_IDEAL] = opEqual<IdealValue, IdealValue>;
	opNotEqualMap[TypeValue::DISPATCH_INDEX_OF_IDEAL][TypeValue::DISPATCH_INDEX_OF_IDEAL] = opNotEqual<IdealValue, IdealValue>;
	opColonMap[TypeValue::DISPATCH_INDEX_OF_IDEAL][TypeValue::DISPATCH_INDEX_OF_IDEAL] = opColon_Ideal_Ideal;
}

ostream &VariableName::dumpAsString(ostream &out) const {
	return out << "<" << this->expId->identifier << " @ " <<
			static_cast<void *>(this->framePointer.toCheckedPointer()) << "[" << this->slotIndex << "]>";
}

ostream &FieldAccess::dumpAsString(std::ostream &out) const {
	return out << "<access to field " << this->fieldName << " of " << this->targetLV << ">";
}

ostream &IntegerIndexedAccess::dumpAsString(std::ostream &out) const {
	return out << "<access to index " << this->index << " of " << this->targetLV << ">";
}

ostream &StringIndexedAccess::dumpAsString(std::ostream &out) const {
	return out << "<indexed-access to field " << this->fieldName << " of " << this->targetLV << ">";
}

intrusive_ptr<RecordValue> FieldAccess::targetAsRecordValue() {
	intrusive_ptr<RightValue> target = this->targetLV->asRightValue();
	intrusive_ptr<RecordValue> rv = dynamic_pointer_cast<RecordValue>(target);
	if (!rv)
		throw WrongTypeException(RecordValue::type->name, target->getType()->name, this->fieldAccessExp->targetExp);
	return rv;
}

intrusive_ptr<RecordValue> StringIndexedAccess::targetAsRecordValue() {
	intrusive_ptr<RightValue> target = this->targetLV->asRightValue();
	intrusive_ptr<RecordValue> rv = dynamic_pointer_cast<RecordValue>(target);
	if (!rv)
		throw WrongTypeException(
				RecordValue::type->name,
				target->getType()->name,
				this->targetExpBegin,
				this->targetExpEnd);
	return rv;
}

intrusive_ptr<RightValue> IntegerIndexedAccess::targetAsListOrMatrixOrMatrixRow(int &which) {
	const intrusive_ptr<RightValue> target = this->targetLV->asRightValue();
	if (dynamic_pointer_cast<ListValue>(target)) {
		which = 1;
		return target;
	}
	if (dynamic_pointer_cast<MatrixValue>(target)) {
		which = 2;
		return target;
	}
	if (dynamic_pointer_cast<MatrixRowValue>(target)) {
		which = 3;
		return target;
	}
	throw WrongTypeException(
			ListValue::type->name + ", " + MatrixValue::type->name + " or " + MatrixRowValue::type->name,
			target->getType()->name,
			this->targetExpBegin,
			this->targetExpEnd);
}

ListValue::ContainerType::size_type IntegerIndexedAccess::getIndexFromBigInt(const BigInt &N, intrusive_ptr<const ListValue> list) const {
	long l;
	if (!IsConvertible(l, N) || l<=0 || static_cast<ListValue::ContainerType::size_type>(l)>list->size())
		throw IndexOutOfRangeException(N, list->size(), this->indexExpBegin, this->indexExpEnd);
	return l-1;
}

string::size_type IntegerIndexedAccess::getIndexFromBigInt(const BigInt &N, intrusive_ptr<const StringValue> str) const {
	long l;
	if (!IsConvertible(l, N) || l<=0 || static_cast<string::size_type>(l)>str->theString.size())
		throw IndexOutOfRangeException(N, str->theString.size(), this->indexExpBegin, this->indexExpEnd);
	return l-1;
}

size_t IntegerIndexedAccess::getIndexFromBigInt(const BigInt &N, intrusive_ptr<const MatrixRowValue> mrv) const {
	long l;
	if (!IsConvertible(l, N) || l<=0 || static_cast<size_t>(l)>mrv->matrix->numColumns())
		throw IndexOutOfRangeException(N, mrv->matrix->numColumns(), this->indexExpBegin, this->indexExpEnd);
	return l-1;
}

intrusive_ptr<RightValue> IntegerIndexedAccess::asRightValue() {
	return this->targetLV->asRightValue()->indexedByBigInt(new BigIntValue(this->index), this->targetExpBegin, this->targetExpEnd, this->indexExpBegin, this->indexExpEnd)->asRightValue();
}

void IntegerIndexedAccess::assign(boost::intrusive_ptr<RightValue> value, const CharPointer &valueExpBegin, const CharPointer &valueExpEnd, RuntimeEnvironment *runtimeEnv) {
	int which;
	intrusive_ptr<RightValue> rv = this->targetAsListOrMatrixOrMatrixRow(which);
	switch (which) {
	case 1 /* ListValue */ : {
			const intrusive_ptr<ListValue> list = intrusive_ptr_cast<ListValue>(rv);
			list->setValue(this->getIndexFromBigInt(index, list), value);
		}
		break;
	case 2 /* MatrixValue */ : {
			throw RuntimeException("Matrix rows cannot be assigned", this->targetExpBegin, this->indexExpEnd);
		}
		break;
	case 3 /* MatrixRowValue */ : {
			try	{
				const intrusive_ptr<MatrixRowValue> mrv = intrusive_ptr_cast<MatrixRowValue>(rv);
				if (const intrusive_ptr<RingElemValue> ringElem = dynamic_pointer_cast<RingElemValue>(value))
					SetEntry(mrv->matrix->theMatrix, mrv->nRow, this->getIndexFromBigInt(index, mrv), ringElem->theRingElem);
				else if (const intrusive_ptr<BigIntValue> N = dynamic_pointer_cast<BigIntValue>(value))
					SetEntry(mrv->matrix->theMatrix, mrv->nRow, this->getIndexFromBigInt(index, mrv), N->theInteger);
				else if (const intrusive_ptr<BigRatValue> qq = dynamic_pointer_cast<BigRatValue>(value))
					SetEntry(mrv->matrix->theMatrix, mrv->nRow, this->getIndexFromBigInt(index, mrv), qq->theRational);
				else
					throw RuntimeException(value->getType()->name+" cannot be assigned to matrix cells", valueExpBegin, valueExpEnd);
			} catch (const ErrorInfo& err) {
				runtimeEnv->announceCLE(err);
				throw RuntimeException(err.what(), valueExpBegin, valueExpEnd);
			}
		}
		break;
	default:
		assert(false);
	}
}

intrusive_ptr<VariableName> IndexedAccess::getBase() {
	return this->targetLV->getBase();
}

bool IndexedAccess::assignmentNeedsOwnership() const {
	return true;
}

intrusive_ptr<RightValue> FieldAccess::asRightValue() {
	intrusive_ptr<RecordValue> rv = this->targetAsRecordValue();
	intrusive_ptr<RightValue> fieldValue = rv->getField(this->fieldName);
	if (!fieldValue)
		throw FieldNotFoundException(this->fieldName, this->fieldAccessExp->tokName.getBegin(), this->fieldAccessExp->tokName.getEnd(), rv);
	return fieldValue;
}

intrusive_ptr<RightValue> StringIndexedAccess::asRightValue() {
	intrusive_ptr<RecordValue> rv = this->targetAsRecordValue();
	intrusive_ptr<RightValue> fieldValue = rv->getField(this->fieldName);
	if (!fieldValue)
		throw FieldNotFoundException(this->fieldName,  this->indexExpBegin, this->indexExpEnd, rv);
	return fieldValue;
}

void StringIndexedAccess::assign(boost::intrusive_ptr<RightValue> value, const CharPointer & /*valueExpBegin*/, const CharPointer & /*valueExpEnd*/, RuntimeEnvironment * /*runtimeEnv*/) {
	intrusive_ptr<RecordValue> rv = this->targetAsRecordValue();
	rv->setField(this->fieldName, value);
}

void FieldAccess::assign(boost::intrusive_ptr<RightValue> value, const CharPointer & /*valueExpBegin*/, const CharPointer & /*valueExpEnd*/, RuntimeEnvironment * /*runtimeEnv*/) {
	intrusive_ptr<RecordValue> rv = this->targetAsRecordValue();
	rv->setField(this->fieldName, value);
}

intrusive_ptr<VariableName> FieldAccess::getBase() {
	return this->targetLV->getBase();
}

bool FieldAccess::assignmentNeedsOwnership() const {
	return true;
}

ostream &ReferenceVariable::dumpAsString(ostream &out) const {
	return out << "<reference to " << this->framePointer.toCheckedPointer() << "[" << this->slotIndex << "]>";
}

VariableSlot *ReferenceVariable::referencedSlot() const {
	return &(this->framePointer.toNonNullCheckedPointer(this->expId)->varSlots[this->slotIndex]);
}

intrusive_ptr<RightValue> ReferenceVariable::asRightValue() {
#ifdef VERBOSE_RUNTIME_DEBUG
	cout << "ReferenceVariable::asRightValue expId=" << this->expId << endl;
#endif // #ifdef VERBOSE_RUNTIME_DEBUG
	try {
		return this->referencedSlot()->value->asRightValue();
	} catch (const InterruptException &) {
		throw;
	} catch (const RuntimeException &) {
		this->throwIncompatibleChanges();
		return VoidValue::theInstance; // useless, but it keeps the compiler happy ;-)
	}
}

void ReferenceVariable::throwIncompatibleChanges() const {
	throw RuntimeException(
			"Reference variable \""+this->expId->identifier+"\" references an undefined variable or its underlying structure has changed in an incompatible way",
			this->expId);
}

void ReferenceVariable::assign(intrusive_ptr<RightValue> value, const CharPointer &valueExpBegin, const CharPointer &valueExpEnd, RuntimeEnvironment *runtimeEnv) {
	try {
		VariableSlot *vs = this->referencedSlot();
		checkProtection(vs, this->expId);
		return intrusive_ptr_cast<LeftValue>(vs->value)->assign(value, valueExpBegin, valueExpEnd, runtimeEnv);
	} catch (const ProtectedVariableException &pe) {
		throw ProtectedVariableException(pe.protectedVarName,
				"Reference variable "+this->expId->identifier+" cannot be assigned because "+pe.protectedVarName +
								", its (directly or indirectly) referenced variable, is protected",
				this->expId);
	} catch (const InterruptException &) {
		throw;
	} catch (const RuntimeException &) {
		this->throwIncompatibleChanges();
	}
}

intrusive_ptr<VariableName> ReferenceVariable::getBase() {
	return this->referencedLV()->getBase();
}

intrusive_ptr<VariableName> VariableName::getBase() {
	return this;
}

intrusive_ptr<RightValue> VariableName::asRightValue() {
	Frame * const f = this->framePointer.toNonNullCheckedPointer(this->expId);
	if (this->slotIndex<0) {
		assert(f==this->runtimeEnvironment->getTopLevelFrame());
		this->slotIndex = this->runtimeEnvironment->slotFor(this->expId->identifier);
		if (this->slotIndex<0)
			throw VariableNotFoundException(this->expId, this->runtimeEnvironment);
	}
#ifdef VERBOSE_RUNTIME_DEBUG
	cout << "VariableName::asRightValue for " << this->expId->identifier << ", this->slotIndex=" << this->slotIndex
			<< ", f->varSlots.size()=" << f->varSlots.size() << endl;
#endif // #ifdef VERBOSE_RUNTIME_DEBUG
	assert(static_cast<vector<VariableSlot>::size_type>(this->slotIndex)<f->varSlots.size());
	VariableSlot &vs = f->varSlots[this->slotIndex];
	if (!vs.value) { // this can only happen when a package has been reloaded (and some of the previous version members are not defined anymore) or an indeterminate has been removed (by Use-ing another ring)
		assert(f == this->runtimeEnvironment->getTopLevelFrame());
		throw VariableNotFoundException(this->expId, this->runtimeEnvironment);
	}
#ifdef VERBOSE_RUNTIME_DEBUG
	cout << "f->varSlots[this->slotIndex].value=" << f->varSlots[this->slotIndex].value
			<< ", f->varSlots[this->slotIndex].value->asRightValue()=" << f->varSlots[this->slotIndex].value->asRightValue() << endl;
#endif // #ifdef VERBOSE_RUNTIME_DEBUG
	return f->varSlots[this->slotIndex].value->asRightValue();
}

bool VariableName::assignmentNeedsOwnership() const {
	return false;
}

bool ReferenceVariable::assignmentNeedsOwnership() const {
	return this->referencedLV()->assignmentNeedsOwnership();
}

void VariableName::assign(boost::intrusive_ptr<RightValue> value, const CharPointer & /*valueExpBegin*/, const CharPointer & /*valueExpEnd*/, RuntimeEnvironment * /*runtimeEnv*/) {
	Frame *f = this->framePointer.toNonNullCheckedPointer(this->expId);
	if (this->slotIndex<0) {
		assert(f==this->runtimeEnvironment->getTopLevelFrame());
		this->slotIndex = this->runtimeEnvironment->slotFor(this->expId->identifier);
		if (this->slotIndex<0) {
			this->slotIndex = this->runtimeEnvironment->setTopLevelVar(this->expId->identifier, value);
			return;
		}
	}
	VariableSlot *vs = &(f->varSlots[this->slotIndex]);
	checkProtection(vs, this->expId);
	vs->value = value;
}

ostream &operator<<(ostream &out, const Value * const value) {
	return value->dumpAsString(out);
}

ostream &BigIntValue::dumpAsString(std::ostream &out) const {
	return out << this->theInteger;
}

intrusive_ptr<RightValue> BigIntValue::unaryMinus(const CharPointer &, RuntimeEnvironment * const) {
	return intrusive_ptr<BigIntValue>(new BigIntValue(-this->theInteger));
}

intrusive_ptr<RightValue> BigRatValue::unaryMinus(const CharPointer &, RuntimeEnvironment * const) {
	return intrusive_ptr<BigRatValue>(new BigRatValue(-this->theRational));
}

intrusive_ptr<RightValue> RingElemValue::unaryMinus(const CharPointer &, RuntimeEnvironment * const) {
	return intrusive_ptr<RingElemValue>(new RingElemValue(-this->theRingElem));
}

intrusive_ptr<RightValue> MatrixValue::unaryMinus(const CharPointer &, RuntimeEnvironment * const) {
	return intrusive_ptr<MatrixValue>(new MatrixValue(-this->theMatrix));
}

// ??? JAA cannot figure this one out currently :-(   ???
// intrusive_ptr<RightValue> ListValue::unaryMinus(const CharPointer &, RuntimeEnvironment * const) {
//   intrusive_ptr<const ListValue> l1 = this->theList;
// 	intrusive_ptr<ListValue> result = new ListValue();
//         typedef ListValue::ContainerType::size_type INDEX;
// 	const INDEX size = l1->size();
// 	for(INDEX i=0; i<size; ++i)
// 		result->addValue(runtimeEnv->binaryOperatorDispatch(
// 								l1->getValue(i),
// 								RuntimeEnvironment::opMinusMap,
// 								beginOperatorSourcePosition,
// 								endOperatorSourcePosition));
// 	return result;
// }

intrusive_ptr<RightValue> RightValue::unaryMinus(const CharPointer &op, RuntimeEnvironment * const) {
	throw RuntimeException("I don't know how to \"negate\" a "+this->getType()->name, op, op);
}

ostream &BigRatValue::dumpAsString(std::ostream &out) const {
	return out << this->theRational;
}

ostream &RecordValue::dumpAsString(std::ostream &out) const {
	out << "record[";
	bool first=true;
	for(MapType::const_iterator pos = this->fields.begin(); pos!=this->fields.end(); ++pos) {
		if (first)
			first = false;
		else
			out << ", ";
		out << pos->first << " := " << pos->second;
		pos->second->dumpRefCountAsString(out);
	}
	return out << "]";
}

RecordValue::MapType::size_type RecordValue::numberOfFields() const {
	return this->fields.size();
}

RecordValue::MapType::const_iterator RecordValue::begin() const {
	return this->fields.begin();
}

RecordValue::MapType::const_iterator RecordValue::end() const {
	return this->fields.end();
}

ostream &BoolValue::dumpAsString(std::ostream &out) const {
	return out << (this->theBool ? "true":"false");
}

ostream &StringValue::dumpAsString(std::ostream &out) const {
	out << '\"';
	BOOST_FOREACH(char c, this->theString) {
		switch (c) {
		case '\n':
			out << "\\n";
			break;
		case '\r':
			out << "\\r";
			break;
		case '\t':
			out << "\\t";
			break;
		case '\a':
			out << "\\a";
			break;
		case '\'':
			out << "\\\'";
			break;
		case '\"':
			out << "\\\"";
			break;
		case '\\':
			out << "\\\\";
			break;
		default:
			out << c;
		}
	}
	return out << '\"';
}

ostream &VoidValue::dumpAsString(std::ostream &out) const {
	return out << "<void>";
}

intrusive_ptr<VoidValue> VoidValue::theInstance(new VoidValue());

intrusive_ptr<RightValue> RecordValue::clone() {
	intrusive_ptr<RecordValue> copy(new RecordValue);
	for(MapType::const_iterator pos = this->fields.begin(); pos!=this->fields.end(); ++pos)
		copy->fields.insert(make_pair(pos->first, pos->second->clone()));
	return copy;
}

intrusive_ptr<RightValue> ListValue::clone() {
	intrusive_ptr<ListValue> copy(new ListValue(this->size()));
	BOOST_FOREACH(intrusive_ptr<RightValue> v, this->container) {
		copy->addValue(v->clone());
	}
	return copy;
}

intrusive_ptr<RightValue> MatrixValue::clone() {
	const matrix &m = this->theMatrix;
	const size_t nRows = NumRows(m);
	const size_t nCols = NumCols(m);
	matrix clone = matrix(m->myZeroClone(BaseRing(m), nRows, nCols));
	for(size_t a=0; a<nRows; ++a)
		for(size_t b=0; b<nCols; ++b)
			SetEntry(clone, a, b, m(a, b));
	return new MatrixValue(clone);
}

bool ListValue::needsToBeCopiedBeforeChanges() const {
	if (this->getRefCount()>1)
		return true;
	for(ContainerType::const_iterator i = this->container.begin(); i!=this->container.end(); ++i)
		if ((*i)->needsToBeCopiedBeforeChanges())
			return true;
	return false;
}

bool MatrixValue::needsToBeCopiedBeforeChanges() const {
	return this->getRefCount()>1;
}

intrusive_ptr<RightValue> ListValue::getValue(ContainerType::size_type index) const {
	assert(index<=this->container.size());
	return this->container[index];
}

void ListValue::setValue(ContainerType::size_type index, intrusive_ptr<RightValue> newValue) {
	assert(index<=this->container.size());
	this->container[index] = newValue;
}

void ListValue::addValue(intrusive_ptr<RightValue> newValue) {
	this->container.push_back(newValue);
}

void ListValue::insertValue(ContainerType::size_type position, intrusive_ptr<RightValue> newValue) {
	this->container.insert(this->container.begin()+position, newValue);
}

void ListValue::removeValue(ContainerType::size_type position) {
	this->container.erase(this->container.begin()+position);
}

ListValue::ContainerType::size_type ListValue::size() const {
	return this->container.size();
}

ostream &MatrixValue::dumpAsString(ostream &out) const {
	return out << this->theMatrix;
}

ostream &MatrixRowValue::dumpAsString(ostream &out) const {
	return out << "<matrix-row>";
}

ostream &ListValue::dumpAsString(ostream &out) const {
	out << "[";
	bool first = true;
	for(ContainerType::const_iterator i = this->container.begin(); i!=this->container.end(); ++i) {
		if (first)
			first = false;
		else
			out << ", ";
		out << *i;
		(*i)->dumpRefCountAsString(out);
	}
	return out << "]";
}

intrusive_ptr<RightValue> RecordValue::getField(const string &fieldName) const {
	MapType::const_iterator pos = this->fields.find(fieldName);
	return (pos!=this->fields.end()) ? pos->second : 0;
}

void RecordValue::setField(const string &fieldName, intrusive_ptr<RightValue> newValue) {
	MapType::iterator pos = this->fields.find(fieldName);
	if (pos!=this->fields.end())
		pos->second = newValue;
	else
		this->fields.insert(make_pair(fieldName, newValue));
}

bool RecordValue::needsToBeCopiedBeforeChanges() const {
	if (this->getRefCount()>1)
		return true;
	for(MapType::const_iterator pos = this->fields.begin(); pos!=this->fields.end(); ++pos)
		if (pos->second->needsToBeCopiedBeforeChanges())
			return true;
	return false;
}

ostream &FunctionValue::dumpAsString(ostream &out) const {
	return out << "<function>";
}

ostream &RingHomValue::dumpAsString(ostream &out) const {
	return out << this->theRingHom;
}

string WrongNumberOfArgumentsException::formatExpectedArgs(int found, int min, int max) {
	assert(min<=max);
	string msg("Wrong number of arguments; found: "+boost::lexical_cast<std::string>(found)+", expecting: ");
	if (min==max)
		return msg+lexical_cast<string>(min);
	return msg+lexical_cast<string>(min) + ".." + lexical_cast<string>(max);
}

UserDefinedFunction::UserDefinedFunction(RuntimeEnvironment *runtimeEnv, const intrusive_ptr<const FunctionDeclaration> fnDecl) :
	fnDecl(fnDecl),
	fp(runtimeEnv->getCurrentFrame())
{
	assert(this->fnDecl);
	BOOST_FOREACH(const Import &import, fnDecl->imports)
		if (import.type==Import::IT_BYVALUE) {
			assert(this->capturedValues.size()==static_cast<std::vector<boost::intrusive_ptr<RightValue> >::size_type>(import.byValueIndex));
			this->capturedValues.push_back(import.expId->evalAs<RightValue>(runtimeEnv));
		}
}

Frame *RuntimeEnvironment::getCurrentFunctionFrame() {
	Frame *f = this->currentFrame;
	assert(f>getTopLevelFrame());
	for(;;) {
		if (f->userdefinedFun)
			return f;
		--f;
		assert(f>getTopLevelFrame());
	}
}

Frame *RuntimeEnvironment::getCurrentFunctionFrameOrNull() {
	Frame *f = this->currentFrame;
	for(;;) {
          if (f==getTopLevelFrame())
			return 0;
		if (f->userdefinedFun)
			return f;
		--f;
	}
}

intrusive_ptr<RightValue> UserDefinedFunction::eval(const intrusive_ptr<const InvocationExpression> invocationExpression, RuntimeEnvironment *runtimeEnv) const {
	const vector<Param>::size_type nParameters = this->fnDecl->params.size();
	const vector<Param>::size_type nMandatoryParameters = static_cast<const vector<Param>::size_type>(this->fnDecl->nMandatoryParameters);
	intrusive_ptr<ListValue> ellipsisValues(new ListValue());
	const bool calledWithEllipsis = invocationExpression->ellipsis;
	const bool receivesEllipsis = this->fnDecl->thereIsEllipsis;
	if (calledWithEllipsis) {
		const Frame * const f = runtimeEnv->getCurrentFunctionFrame();
		assert(f->varSlots.size());
		const VariableSlot &vs = f->varSlots.front();
		assert(vs.isArgs());
		ellipsisValues = intrusive_ptr_cast<ListValue>(vs.value);
		if (!receivesEllipsis) {
			const ListValue::ContainerType::size_type nArgs = ellipsisValues->size();
			if (nArgs<nMandatoryParameters || nArgs>nParameters)
				throw WrongNumberOfArgumentsException(invocationExpression, nArgs, nMandatoryParameters, nParameters);
			BOOST_FOREACH(const Param &param, this->fnDecl->params)
				if (param.byRef)
					throw RuntimeException("Cannot use ellipsis on functions/procedures expecting by-ref parameters", invocationExpression);
		}
	} else { // called with actual arguments (that is, not "...")
		const vector<Argument>::size_type nArgs = invocationExpression->args.size();
		if (receivesEllipsis) {
			BOOST_FOREACH(const Argument &arg, invocationExpression->args)
				if (arg.byRef)
					throw RuntimeException("Cannot pass by-ref arguments to functions/procedures expecting a variable number of arguments", invocationExpression);
		} else
			if (nArgs<nMandatoryParameters || nArgs>nParameters)
				throw WrongNumberOfArgumentsException(invocationExpression, nArgs, nMandatoryParameters, nParameters);
	}
	vector<VariableSlot> slots;
	if (!receivesEllipsis) {
		if (calledWithEllipsis) {
			const ListValue::ContainerType::size_type nArgs = ellipsisValues->size();
			for(vector<Param>::size_type a=0; a<nParameters; ++a) {
				if (a<nArgs) {
					const intrusive_ptr<RightValue> v = ellipsisValues->getValue(a);
					assert(v);
					slots.push_back(VariableSlot(v, VariableSlot::VSF_None));
				} else {
					assert(this->fnDecl->params[a].opt);
					slots.push_back(VariableSlot(VoidValue::theInstance, VariableSlot::VSF_None));
				}
			}
		} else { // don't receive ellipsis and is not called with ellipsis
			const vector<Argument>::size_type nArgs = invocationExpression->args.size();
			for(vector<Param>::size_type a=0; a<nParameters; ++a) {
				const Param &param = this->fnDecl->params[a];
				if (a<nArgs)
					slots.push_back(VariableSlot(runtimeEnv->evalArg(invocationExpression->args[a], param.byRef ? RuntimeEnvironment::EVAL_BY_REF:RuntimeEnvironment::EVAL_BY_VALUE), VariableSlot::VSF_None));
				else {
					assert(param.opt);
					slots.push_back(VariableSlot(VoidValue::theInstance, VariableSlot::VSF_None));
				}
			}
		}
	} else { // receives ellipsis
		slots.push_back(VariableSlot(ellipsisValues, VariableSlot::VSF_Args));
		if (!calledWithEllipsis) {
			const vector<Argument>::size_type nArgs = invocationExpression->args.size();
			for(vector<Argument>::size_type a=0; a<nArgs; ++a)
				ellipsisValues->addValue(runtimeEnv->evalArg(invocationExpression->args[a], RuntimeEnvironment::EVAL_BY_VALUE)->asRightValue());
		}
	}
	Frame *frame = runtimeEnv->pushFrame(this->fp, invocationExpression, this->fnDecl, this);
	BOOST_SCOPE_EXIT( (runtimeEnv) ) {
		runtimeEnv->popFrame();
	} BOOST_SCOPE_EXIT_END
	const std::map<std::string, int>::size_type localSize = this->fnDecl->localNames.size();
	assert(slots.size()<=localSize);
	while (slots.size()<localSize)
		slots.push_back(VariableSlot(VoidValue::theInstance, VariableSlot::VSF_None));
	frame->varSlots.swap(slots);
#ifdef VERBOSE_RUNTIME_DEBUG
	{
		cout << "\n======\nUserDefinedFunction::eval, frames->varSlots:";
		int i = 0;
		BOOST_FOREACH(const VariableSlot &vs, frame->varSlots)
			cout << "[" << i++ << "].value = " << vs.value << endl;
		cout << "======\n";
	}
#endif // #ifdef VERBOSE_RUNTIME_DEBUG
	try {
		this->fnDecl->statements->execute(runtimeEnv);
		if (this->fnDecl->thereAreReturnsWithExpr) {
			const CharPointer p = this->fnDecl->statements->getEnd();
			throw RuntimeException("A function must explicitly Return a value", p, p);
		}
	} catch (const Return &ret) {
		return ret.value;
	} catch (const Break &) {
		assert(false);
	} catch (const Continue &) {
		assert(false);
	} catch (RuntimeException &re) {
		if (re.snapshot.size()==0)
			re.snapshot = runtimeEnv->takeSnapshot();
		throw;
	}
	return VoidValue::theInstance;
}

intrusive_ptr<RightValue> RingHomValue::eval(const intrusive_ptr<const InvocationExpression> invocationExpression, RuntimeEnvironment *runtimeEnv) const {
	intrusive_ptr<RightValue> argValue;
	CharPointer argBegin(invocationExpression->getBegin());
	CharPointer argEnd(invocationExpression->getEnd());
	if (invocationExpression->ellipsis) {
		argBegin = invocationExpression->tokEllipsis->getBegin();
		argEnd = invocationExpression->tokEllipsis->getEnd();
		const Frame * const f = runtimeEnv->getCurrentFunctionFrame();
		assert(f->varSlots.size());
		const VariableSlot &vs = f->varSlots.front();
		assert(vs.isArgs());
		intrusive_ptr<ListValue> ellipsisValues(intrusive_ptr_cast<ListValue>(vs.value));
		const ListValue::ContainerType::size_type nArgs = ellipsisValues->size();
		if (nArgs!=1)
			throw WrongNumberOfArgumentsException(invocationExpression, nArgs, 1);
		argValue = ellipsisValues->getValue(0)->asRightValue();
	} else {
		const Argument &arg = invocationExpression->args.front();
		argBegin = arg.exp->getBegin();
		argEnd = arg.exp->getEnd();
		const vector<Argument>::size_type nArgs = invocationExpression->args.size();
		if (nArgs!=1)
			throw WrongNumberOfArgumentsException(invocationExpression, nArgs, 1);
		argValue = intrusive_ptr_cast<RightValue>(runtimeEnv->evalArg(arg, RuntimeEnvironment::EVAL_BY_VALUE));
	}
	try	{
		if (intrusive_ptr<RingElemValue> poly = dynamic_pointer_cast<RingElemValue>(argValue))
			return new RingElemValue(this->theRingHom(poly->theRingElem));
		if (intrusive_ptr<BigIntValue> N = dynamic_pointer_cast<BigIntValue>(argValue))
			return new RingElemValue(this->theRingHom(N->theInteger));
		if (intrusive_ptr<BigRatValue> q = dynamic_pointer_cast<BigRatValue>(argValue))
			return new RingElemValue(this->theRingHom(q->theRational));
		else if (intrusive_ptr<RingHomValue> hom = dynamic_pointer_cast<RingHomValue>(argValue))
			return new RingHomValue(this->theRingHom(hom->theRingHom));
		throw WrongTypeException(
				RingElemValue::type->name + " or " +
				BigIntValue::type->name + " or " +
				BigRatValue::type->name + " or " +
        RingHomValue::type->name,
				argValue->getType()->name,
				argBegin,
				argEnd);
	} catch (const ErrorInfo& err) {
		runtimeEnv->announceCLE(err);
		throw RuntimeException(err.what(), invocationExpression->getBegin(), invocationExpression->getEnd());
	}
}

intrusive_ptr<RightValue> BuiltInFunction::eval(const intrusive_ptr<const InvocationExpression> invocationExpression, RuntimeEnvironment *runtimeEnv) const {
#ifdef C5IDE
	Interpreter::InterpreterStatus prevStatus = runtimeEnv->interpreter->UpdateStatusStartingBuiltIn();
	BOOST_SCOPE_EXIT( (runtimeEnv)(prevStatus) ) {
		runtimeEnv->interpreter->UpdateStatusEndingBuiltIn(prevStatus);
	} BOOST_SCOPE_EXIT_END
#endif // #ifdef C5IDE
	try {
		if (invocationExpression->ellipsis) {
			const Frame * const f = runtimeEnv->getCurrentFunctionFrame();
			assert(f->varSlots.size());
			const VariableSlot &vs = f->varSlots.front();
			assert(vs.isArgs());
			intrusive_ptr<ListValue> ellipsisValues(intrusive_ptr_cast<ListValue>(vs.value));
			ListValue::ContainerType::size_type nArgs = ellipsisValues->size();
			vector<Argument> args;
			CharPointer begin(invocationExpression->tokEllipsis->getBegin());
			CharPointer end(invocationExpression->tokEllipsis->getEnd());
			for(ListValue::ContainerType::size_type a=0; a<nArgs; ++a)
				args.push_back(Argument(false, new FakeExpression(ellipsisValues->getValue(a), begin, end), true));
      boost::shared_ptr<Token> noEllipsis;
			const intrusive_ptr<const InvocationExpression> fakeInvocationExpression(new InvocationExpression(invocationExpression->targetExp, noEllipsis, args, invocationExpression->getEnd(), invocationExpression->packageName));
			return this->function(fakeInvocationExpression, runtimeEnv);
		}
		return this->function(invocationExpression, runtimeEnv);
	} catch (const ErrorInfo& err) {
		runtimeEnv->announceCLE(err);
		throw RuntimeException(err.what(), invocationExpression);
	}
}

void LeftValue::obtainOwnership() {
	const intrusive_ptr<VariableName> base = this->getBase();
	Frame * const f = base->framePointer.toNonNullCheckedPointer(base->expId);
	if (base->slotIndex<0) {
		assert(f==base->runtimeEnvironment->getTopLevelFrame());
		base->slotIndex = base->runtimeEnvironment->slotFor(base->expId->identifier);
		if (base->slotIndex<0)
			return;
	}
	assert(static_cast<vector<VariableSlot>::size_type>(base->slotIndex)<f->varSlots.size());
	assert(f->varSlots[base->slotIndex].value);
	VariableSlot &varSlot = f->varSlots[base->slotIndex];
	RightValue * const target = intrusive_ptr_cast<RightValue>(varSlot.value).get();
	if (target->needsToBeCopiedBeforeChanges())
		varSlot.value = target->clone();
}

ListValue::ContainerType::size_type ListValue::checkIndex(const intrusive_ptr<const RightValue> v, const intrusive_ptr<const Expression> originalExp) const {
	if (const intrusive_ptr<const BigIntValue> NV = dynamic_pointer_cast<const BigIntValue>(v)) {
		const BigInt N(NV->theInteger);
		long l;
		if (!IsConvertible(l, N) || l<=0 || static_cast<ContainerType::size_type>(l)>this->size())
			throw IndexOutOfRangeException(N, this->size(), originalExp);
		return l-1;
	}
	throw WrongTypeException(
			BigIntValue::type->name,
			v->getType()->name,
			originalExp);
}

namespace {
	void printErrors(intrusive_ptr<ErrorReporter> errorReporter, intrusive_ptr<OutputStreamValue> output) {
		assert(errorReporter);
		const int nErrors = errorReporter->getErrorCount();
		const int nWarnings = errorReporter->getWarningCount();
		if (nErrors || nWarnings)
			output->print("Got ")->print(lexical_cast<string>(nErrors))->print(" error(s) and ")->print(lexical_cast<string>(nWarnings))->print(" warning(s).\n");
	}

	void execute(intrusive_ptr<Statement> s, RuntimeEnvironment *re, intrusive_ptr<ErrorReporter> errorReporter) {
		assert(re);
#ifdef VERBOSE_RUNTIME_DEBUG
		cout << "Parsed statement = " << s << "\n";
#endif // #ifdef VERBOSE_RUNTIME_DEBUG
		CheckNamesVisitor cnv(re, errorReporter);
		s->accept(&cnv);
#ifdef VERBOSE_RUNTIME_DEBUG
		DumpAsTreeVisitor v(cout);
		s->accept(&v);
#endif // #ifdef VERBOSE_RUNTIME_DEBUG
		if (!errorReporter->getErrorCount())
			s->execute(re);
	}
}

#ifdef C5IDE
int Interpreter::run(IDE::Console *console) {
#else // #ifdef C5IDE
int Interpreter::run() {
#endif // #ifdef C5IDE
#ifdef C5IDE
	assert(this->status==IS_WAITING_FOR_COMMAND);
	loadPackages(console);
#endif // #ifdef C5IDE
	try {
		intrusive_ptr<Lexer> lexer = intrusive_ptr<Lexer>(new Lexer(this->errorReporter, this->lineProvider, this->warnAboutCocoa5, true));
		intrusive_ptr<Parser> parser(new Parser(lexer));
		for(;;) {
#ifdef C5IDE
			this->status = IS_WAITING_FOR_COMMAND;
#endif // #ifdef C5IDE
			intrusive_ptr<Statement> s;
			try	{
				this->errorReporter->resetErrorCounts();
				lexer->startingTopLevelStatement();
				Token t = lexer->getToken(parser->getStatus());
				if (t.getType()==TT_EOF)
					return EXIT_SUCCESS;
				lexer->ungetToken(t);
				s = parser->parseTopLevelStatement();
				if (!s || this->errorReporter->getErrorCount())
					continue;
			} catch (const UnexpectedTokenException &ete) {
				lexer->reportError(ete.reason, ete.found.getBegin(), ete.found.getEnd(), false);
				if (ete.found.getType()==TT_EOF)
					return EXIT_FAILURE;
				if (ete.needsRecovery)
					parser->tryToRecover(ete.from);
				continue;
			} catch (const ExceptionWithSourcePosition &ewsp) {
				lexer->reportError(ewsp.reason, ewsp.from, ewsp.to, false);
				if (ewsp.needsRecovery)
					parser->tryToRecover(ewsp.from);
				continue;
			} catch (const BaseException &e) {
				lexer->reportError(e.reason);
				continue;
			}
			assert(s && this->errorReporter->getErrorCount()==0);
			try {
#ifdef C5IDE
				this->controlC = false;
				this->status = IS_RUNNING;
				this->runtimeEnvironment.getOutputStream()->flush(); // this allows the GUI thread to update the status label
#endif // #ifdef C5IDE
					execute(s, &(this->runtimeEnvironment), this->errorReporter);
			} catch (const Ciao &) {
#ifdef C5IDE
				this->status = IS_ENDED;
#endif // #ifdef C5IDE
				return EXIT_SUCCESS;
			} catch (const RuntimeException &re) {
				this->errorReporter->reportError(re);
			}
		}
	} catch (const BaseException &e) {
		cout << "***ERROR*** UNCAUGHT Interpreter BaseException exc-reason=" << e.reason << endl;
	} catch (const ErrorInfo& err) {
		cout << "***ERROR***  UNCAUGHT CoCoA error" << endl;
		ANNOUNCE(cout, err);
	} catch (const std::exception& exc) {
		cout << "***ERROR***  UNCAUGHT std::exception: " << exc.what() << endl;
	} catch(...) {
		cout << "***ERROR***  UNCAUGHT UNKNOWN EXCEPTION" << endl;
	}
	return EXIT_FAILURE;
}

void Interpreter::readAndExecute(const std::string &filename, bool calledFromMain, bool immediateExecution, long FromLine, long FromChar, long ToLine, long ToChar) {
	struct stat statbuf;
	if (stat(filename.c_str(), &statbuf)) {
		if (errno==ENOENT)
			throw BaseException("Cannot find a file named \""+filename+"\"");
		throw BaseException("Cannot read \""+filename+"\" information; system-error: "+strerror(errno));
	}
	FilePair pair(statbuf.st_dev, statbuf.st_ino);
	if (!this->sourcedFiles.insert(pair).second)
		throw BaseException("Files cannot be recursively Source-d");
	Interpreter &This = *this;
	BOOST_SCOPE_EXIT( (&This)(&pair) ) {
		This.sourcedFiles.erase(pair);
	} BOOST_SCOPE_EXIT_END
	try {
          boost::intrusive_ptr<LineProvider> SourceText;
          if (FromLine < 0) SourceText = new FileLineProvider(filename);
          else SourceText = new FileRegionLineProvider(filename, FromLine,FromChar, ToLine,ToChar);
		intrusive_ptr<Lexer> lexer = new Lexer(this->errorReporter, SourceText, this->warnAboutCocoa5, false);
		intrusive_ptr<Parser> parser(new Parser(lexer));
		vector<intrusive_ptr<Statement> > statements;
		for(;;) {
			intrusive_ptr<Statement> s;
			try	{
				lexer->startingTopLevelStatement();
				Token t = lexer->getToken(parser->getStatus());
				if (t.getType()==TT_EOF)
					break;
				lexer->ungetToken(t);
				s = parser->parseTopLevelStatement();
			} catch (const UnexpectedTokenException &ete) {
				lexer->reportError(ete.reason, ete.found.getBegin(), ete.found.getEnd(), false);
				this->errorReporter->outputStream->newline();
				if (ete.found.getType()==TT_EOF) {
					if (calledFromMain) {
						printErrors(this->errorReporter, this->runtimeEnvironment.getOutputStream());
						return;
					}
					throw BaseException("Unexpected End-Of-Input while reading \""+filename+"\"");
				}
				if (ete.needsRecovery)
					parser->tryToRecover(ete.from);
			} catch (const ExceptionWithSourcePosition &ewsp) {
				lexer->reportError(ewsp.reason, ewsp.from, ewsp.to, false);
				this->errorReporter->outputStream->newline();
				if (ewsp.needsRecovery)
					parser->tryToRecover(ewsp.from);
			}
			if (s) {
				if (immediateExecution && this->errorReporter->getErrorCount()==0) {
					try {
						execute(s, &(this->runtimeEnvironment), this->errorReporter);
					} catch (const InterruptException &) {
						throw;
					} catch (const RuntimeException &re) {
						this->errorReporter->reportError(re);
						this->errorReporter->outputStream->newline();
					}
				} else
					statements.push_back(s);
			}
		}
		try {
			BOOST_FOREACH(intrusive_ptr<Statement> s, statements) {
				if (this->errorReporter->getErrorCount())
					break;
				execute(s, &(this->runtimeEnvironment), this->errorReporter);
			}
		} catch (const InterruptException &) {
			throw;
		} catch (const RuntimeException &re) {
			this->errorReporter->reportError(re);
			this->errorReporter->outputStream->newline();
		}
	} catch (BaseException &) {
		throw;
	} catch (const std::exception &e) {
		throw BaseException(string(e.what())+"\nWhile trying to read and execute \""+filename+"\"");
	}
	if (calledFromMain)
		printErrors(this->errorReporter, this->runtimeEnvironment.getOutputStream());
	else if (this->errorReporter->getErrorCount())
		throw BaseException("Read and execution of the source file \""+filename+"\" failed");
}

bool UserDefinedFunction::canBeCalledWith(int nArg) {
	const intrusive_ptr<const FunctionDeclaration> fnDecl = this->fnDecl;
	return fnDecl->thereIsEllipsis || (nArg>=fnDecl->nMandatoryParameters && static_cast<vector<Param>::size_type>(nArg)<=fnDecl->params.size());
}

bool BuiltInFunction::canBeCalledWith(int nArg) {
	return this->arityCheck(nArg);
}

bool RingHomValue::canBeCalledWith(int nArg) {
	return nArg==1;
}

intrusive_ptr<VariableName> PackageValue::toVariableName(RuntimeEnvironment *runtimeEnv, const string &memberName, const Token &tokMemberName) {
	const int index = runtimeEnv->slotFor(this->prefix+memberName);
	if (index<0)
		throw RuntimeException("Package "+this->pkgName+" doesn't have a member named "+memberName, tokMemberName);
	return new VariableName(runtimeEnv, runtimeEnv->getTopLevelFrame(), index, new Identifier(tokMemberName));
}

void RuntimeEnvironment::collectSimilarlyNamedIdentifiers(const string &id, int arity, set<string> &set, bool &thereIsAnExactMatch) const {
	const int maxDistance = maximumDistanceForSimilaryIdentifiers(id);
	for(map<string, int>::const_iterator pos = this->topLevelIdentifiers.begin(); pos!=this->topLevelIdentifiers.end(); ++pos) {
		if (pos->first==id) {
                  if (!getTopLevelFrame()->varSlots[pos->second].value)
				continue;
			thereIsAnExactMatch = true;
			return;
		}
		if (levenshteinDistance(id, pos->first)<=maxDistance) {
			if (arity>=0) {
                          const VariableSlot &vs = getTopLevelFrame()->varSlots[pos->second];
				intrusive_ptr<Value> v = vs.value;
				if (!v)
					continue;
				if (vs.isPackage()) {
					const intrusive_ptr<const VariableName> vn = intrusive_ptr_cast<VariableName>(v);
					const int index = vn->expId->varData.index;
					assert(vn->expId->varData.depth==StaticEnv::TOP_LEVEL);
					assert(index>=0 && static_cast<vector<VariableSlot>::size_type>(index)<getTopLevelFrame()->varSlots.size());
					v = getTopLevelFrame()->varSlots[index].value;
					if (!v)
						continue;
				}
				if (const intrusive_ptr<FunctionValue> fv = dynamic_pointer_cast<FunctionValue>(v)) {
					if (!fv->canBeCalledWith(arity))
						continue;
				} else
					continue;
			}
			set.insert(pos->first);
		}
	}
}

//??? template<typename T>  ???
 boost::intrusive_ptr<RecordValue> Value::from(const factorization<RingElem>& f)
 {
   intrusive_ptr<RecordValue> facsC5(new RecordValue);
   facsC5->setField("factors", Value::from(f.myFactors()));
   facsC5->setField("multiplicities", Value::from(f.myMultiplicities()));
   facsC5->setField("RemainingFactor", Value::from(f.myRemainingFactor()));
   return facsC5;
 }

 boost::intrusive_ptr<RecordValue> Value::from(const factorization<BigInt>& f)
 {
   intrusive_ptr<RecordValue> facsC5(new RecordValue);
   facsC5->setField("factors", Value::from(f.myFactors()));
   facsC5->setField("multiplicities", Value::from(f.myMultiplicities()));
   facsC5->setField("RemainingFactor", Value::from(f.myRemainingFactor()));
   return facsC5;
 }

 boost::intrusive_ptr<RecordValue> Value::from(const HPSeries &s)
 {
   intrusive_ptr<RecordValue> r(new RecordValue);
   r->setField("num", Value::from(num(s)));
   r->setField("DenFactors", Value::from(DenFactors(s)));
   return r;
 }
 
 boost::intrusive_ptr<ListValue> Value::from(const degree &x)
 { return Value::from(DegreeToVec(x)); }

}
 // namespace InterpreterNS

namespace AST {

using namespace std;
using namespace boost;
using namespace CoCoA::InterpreterNS;

IntLiteral::IntLiteral(const Token &t) :
	LiteralExpression(t.getBegin(), t.getEnd()),
	theInteger(BigIntValue::from(BigInt(t.lexeme())))
{
}

FloatLiteral::FloatLiteral(const Token &t) :
		LiteralExpression(t.getBegin(), t.getEnd()),
		theRational(BigRatValue::simplifiedValueFrom(buildTheRational(t)))
{
}

StringLiteral::StringLiteral(const CharPointer &beginSourcePosition, const CharPointer & endSourcePosition, const string &unescapedString) :
	LiteralExpression(beginSourcePosition, endSourcePosition),
	theString(StringValue::from(unescapedString))
{
}

BoolLiteral::BoolLiteral(const CharPointer &beginSourcePosition, const CharPointer & endSourcePosition, TokenType tokenType) :
	LiteralExpression(beginSourcePosition, endSourcePosition),
	theBool(BoolValue::from(tokenType==LexerNS::TT_TRUE))
{
	assert(tokenType==LexerNS::TT_TRUE || tokenType==LexerNS::TT_FALSE);
}

int InvocationExpression::checkNumberOfArgs(int n) const {
	const int nArgs = this->args.size();
	if (nArgs!=n)
		throw WrongNumberOfArgumentsException(this, nArgs, n);
	return nArgs;
}

int InvocationExpression::checkNumberOfArgs(int nMin, int nMax) const {
	const int nArgs = this->args.size();
	if (nArgs<nMin || nArgs>nMax)
		throw WrongNumberOfArgumentsException(this, nArgs, nMin, nMax);
	return nArgs;
}

void StaticEnv::collectSimilarlyNamedIdentifiers(const string &id, int arity, set<string> &set, bool &thereIsAnExactMatch) const {
	if (this->runtimeEnvironment) {
		assert(!this->parent);
		this->runtimeEnvironment->collectSimilarlyNamedIdentifiers(id, arity, set, thereIsAnExactMatch);
		return;
	}
	assert(this->parent);
	const int maxDistance = maximumDistanceForSimilaryIdentifiers(id);
	for(IdMap::const_iterator pos = this->identifierMap.begin(); pos!=this->identifierMap.end(); ++pos) {
		if (pos->second.depth<0)
			continue;
		if (pos->first==id) {
			thereIsAnExactMatch = true;
			return;
		}
		if (levenshteinDistance(id, pos->first)<=maxDistance)
			set.insert(pos->first);
	}
	if (!thereIsAnExactMatch)
		this->parent->collectSimilarlyNamedIdentifiers(id, arity, set, thereIsAnExactMatch);
}

namespace {
	intrusive_ptr<Value> dontKnow(const intrusive_ptr<const Expression> e) {
		throw RuntimeException("I don't know (yet) how to evaluate this expression", e->getBegin(), e->getEnd());
	}
}

intrusive_ptr<Value> ColonExpression::implEval(RuntimeEnvironment *runtimeEnv) const {
	return runtimeEnv->binaryOperatorDispatch(
				this->leftExp->evalAs<RightValue>(runtimeEnv),
				this->rightExp->evalAs<RightValue>(runtimeEnv),
				RuntimeEnvironment::opColonMap,
			this->beginOperatorSourcePosition,
			this->endOperatorSourcePosition);
}

intrusive_ptr<Value> ModuloExpression::implEval(RuntimeEnvironment * /*runtimeEnv*/) const {
	return dontKnow(this);
}

intrusive_ptr<Value> BackwardCompatibleExp::implEval(RuntimeEnvironment *runtimeEnv) const {
	return this->innerExp->eval(runtimeEnv);
}

intrusive_ptr<Value> ScopedExpression::implEval(RuntimeEnvironment *) const {
	return dontKnow(this);
}

intrusive_ptr<Value> UnaryPlusExpression::implEval(RuntimeEnvironment *runtimeEnv) const {
	return this->exp->evalAs<RightValue>(runtimeEnv);
}

intrusive_ptr<Value> UnaryMinusExpression::implEval(RuntimeEnvironment * const runtimeEnv) const {
	return this->exp->evalAs<RightValue>(runtimeEnv)->unaryMinus(this->getBegin(), runtimeEnv);
}

intrusive_ptr<Value> IntLiteral::implEval(RuntimeEnvironment *) const {
	return this->theInteger;
}

intrusive_ptr<Value> FloatLiteral::implEval(RuntimeEnvironment *) const {
	return this->theRational;
}

intrusive_ptr<Value> BoolLiteral::implEval(RuntimeEnvironment *) const {
	return this->theBool;
}

intrusive_ptr<Value> StringLiteral::implEval(RuntimeEnvironment *) const {
	return this->theString;
}

intrusive_ptr<Value> SumExpression::implEval(RuntimeEnvironment *runtimeEnv) const {
	return runtimeEnv->binaryOperatorDispatch(
			this->leftExp->evalAs<RightValue>(runtimeEnv),
			this->rightExp->evalAs<RightValue>(runtimeEnv),
			RuntimeEnvironment::opPlusMap,
			this->beginOperatorSourcePosition,
			this->endOperatorSourcePosition);
}

intrusive_ptr<Value> PowerExpression::implEval(RuntimeEnvironment *runtimeEnv) const {
	return runtimeEnv->binaryOperatorDispatch(
			this->leftExp->evalAs<RightValue>(runtimeEnv),
			this->rightExp->evalAs<RightValue>(runtimeEnv),
			RuntimeEnvironment::opPowerMap,
			this->beginOperatorSourcePosition,
			this->endOperatorSourcePosition);
}

intrusive_ptr<Value> IsInExpression::implEval(RuntimeEnvironment *runtimeEnv) const {
	intrusive_ptr<RightValue> v = this->leftExp->evalAs<RightValue>(runtimeEnv);
	intrusive_ptr<RightValue> listOrString = this->rightExp->evalAs<RightValue>(runtimeEnv);
	if (const intrusive_ptr<ListValue> list = dynamic_pointer_cast<ListValue>(listOrString)) {
		ListValue::ContainerType::size_type size = list->size();
		for(ListValue::ContainerType::size_type a=0; a<size; ++a) {
			try {
				intrusive_ptr<BoolValue> bv = intrusive_ptr_cast<BoolValue>(
						runtimeEnv->binaryOperatorDispatch(
								v,
								list->getValue(a),
								RuntimeEnvironment::opEqualMap,
								this->getBegin(),
								this->getEnd()));
				if (bv->theBool)
					return bv;
			} catch (const InterruptException &) {
				throw;
			} catch (const RuntimeException &) {
				/* nop (this just means that we cannot compare v with an element of list) */
			}
		}
		return BoolValue::falseValue;
	} else if (const intrusive_ptr<StringValue> str = dynamic_pointer_cast<StringValue>(listOrString)) {
		intrusive_ptr<StringValue> leftStr = dynamic_pointer_cast<StringValue>(v);
		if (!leftStr)
			throw WrongTypeException(StringValue::type->name, v->getType()->name, this->leftExp);
		return BoolValue::fromBool(str->theString.find(leftStr->theString)!=string::npos);
	} else if (const intrusive_ptr<IdealValue> I = dynamic_pointer_cast<IdealValue>(listOrString)) {
	if (intrusive_ptr<RingElemValue> f = dynamic_pointer_cast<RingElemValue>(v))
          try { return BoolValue::fromBool(IsElem(f->theRingElem, I->theIdeal)); }
          catch (const ErrorInfo& err) { throw RuntimeException(err.what(), beginOperatorSourcePosition, endOperatorSourcePosition); }
	if (intrusive_ptr<BigIntValue> f = dynamic_pointer_cast<BigIntValue>(v))
		return BoolValue::fromBool(IsElem(f->theInteger, I->theIdeal));
	if (intrusive_ptr<BigRatValue> f = dynamic_pointer_cast<BigRatValue>(v))
		return BoolValue::fromBool(IsElem(f->theRational, I->theIdeal));
  throw WrongTypeException(RingElemValue::type->name  + " or " +
                           BigIntValue::type->name  + " or " +
                           BigRatValue::type->name, v->getType()->name, this->leftExp);
	} else if (const intrusive_ptr<ModuleValue> M = dynamic_pointer_cast<ModuleValue>(listOrString)) {
	if (intrusive_ptr<ModuleElemValue> mv = dynamic_pointer_cast<ModuleElemValue>(v))
          try { return BoolValue::fromBool(IsElem(mv->theModuleElem, M->theModule)); }
          catch (const ErrorInfo& err) { throw RuntimeException(err.what(), beginOperatorSourcePosition, endOperatorSourcePosition); }
  throw WrongTypeException(ModuleElemValue::type->name,
                           v->getType()->name, this->leftExp);
	}
	throw WrongTypeException(
			ListValue::type->name + " or " + StringValue::type->name + " or " + 
			IdealValue::type->name + " or " + ModuleValue::type->name,
			listOrString->getType()->name,
			this->rightExp);
}

intrusive_ptr<Value> AndExpression::implEval(RuntimeEnvironment *runtimeEnv) const {
	intrusive_ptr<BoolValue> left = this->leftExp->evalAs<BoolValue>(runtimeEnv);
	if (!left->theBool)
		return left;
	return this->rightExp->evalAs<BoolValue>(runtimeEnv);
}

intrusive_ptr<Value> OrExpression::implEval(RuntimeEnvironment *runtimeEnv) const {
	intrusive_ptr<BoolValue> left = this->leftExp->evalAs<BoolValue>(runtimeEnv);
	if (left->theBool)
		return left;
	return this->rightExp->evalAs<BoolValue>(runtimeEnv);
}

intrusive_ptr<Value> LessThanExpression::implEval(RuntimeEnvironment *runtimeEnv) const {
	return runtimeEnv->binaryOperatorDispatch(
			this->leftExp->evalAs<RightValue>(runtimeEnv),
			this->rightExp->evalAs<RightValue>(runtimeEnv),
			RuntimeEnvironment::opLessThanMap,
			this->beginOperatorSourcePosition,
			this->endOperatorSourcePosition);
}

intrusive_ptr<Value> LessOrEqualExpression::implEval(RuntimeEnvironment *runtimeEnv) const {
	return runtimeEnv->binaryOperatorDispatch(
			this->leftExp->evalAs<RightValue>(runtimeEnv),
			this->rightExp->evalAs<RightValue>(runtimeEnv),
			RuntimeEnvironment::opLessOrEqualMap,
			this->beginOperatorSourcePosition,
			this->endOperatorSourcePosition);
}

intrusive_ptr<Value> EqualExpression::implEval(RuntimeEnvironment *runtimeEnv) const {
	return runtimeEnv->binaryOperatorDispatch(
			this->leftExp->evalAs<RightValue>(runtimeEnv),
			this->rightExp->evalAs<RightValue>(runtimeEnv),
			RuntimeEnvironment::opEqualMap,
			this->beginOperatorSourcePosition,
			this->endOperatorSourcePosition);
}

intrusive_ptr<Value> NotEqualExpression::implEval(RuntimeEnvironment *runtimeEnv) const {
	return runtimeEnv->binaryOperatorDispatch(
			this->leftExp->evalAs<RightValue>(runtimeEnv),
			this->rightExp->evalAs<RightValue>(runtimeEnv),
			RuntimeEnvironment::opNotEqualMap,
			this->beginOperatorSourcePosition,
			this->endOperatorSourcePosition);
}

intrusive_ptr<Value> GreaterOrEqualExpression::implEval(RuntimeEnvironment *runtimeEnv) const {
	return runtimeEnv->binaryOperatorDispatch(
			this->leftExp->evalAs<RightValue>(runtimeEnv),
			this->rightExp->evalAs<RightValue>(runtimeEnv),
			RuntimeEnvironment::opGreaterOrEqualMap,
			this->beginOperatorSourcePosition,
			this->endOperatorSourcePosition);
}

intrusive_ptr<Value> GreaterThanExpression::implEval(RuntimeEnvironment *runtimeEnv) const {
	return runtimeEnv->binaryOperatorDispatch(
			this->leftExp->evalAs<RightValue>(runtimeEnv),
			this->rightExp->evalAs<RightValue>(runtimeEnv),
			RuntimeEnvironment::opGreaterThanMap,
			this->beginOperatorSourcePosition,
			this->endOperatorSourcePosition);
}

intrusive_ptr<Value> SubtractionExpression::implEval(RuntimeEnvironment *runtimeEnv) const {
	return runtimeEnv->binaryOperatorDispatch(
			this->leftExp->evalAs<RightValue>(runtimeEnv),
			this->rightExp->evalAs<RightValue>(runtimeEnv),
			RuntimeEnvironment::opMinusMap,
			this->beginOperatorSourcePosition,
			this->endOperatorSourcePosition);
}

intrusive_ptr<Value> ProductExpression::implEval(RuntimeEnvironment *runtimeEnv) const {
	return runtimeEnv->binaryOperatorDispatch(
			this->leftExp->evalAs<RightValue>(runtimeEnv),
			this->rightExp->evalAs<RightValue>(runtimeEnv),
			RuntimeEnvironment::opStarMap,
			this->beginOperatorSourcePosition,
			this->endOperatorSourcePosition);
}

intrusive_ptr<Value> DivisionExpression::implEval(RuntimeEnvironment *runtimeEnv) const {
	return runtimeEnv->binaryOperatorDispatch(
			this->leftExp->evalAs<RightValue>(runtimeEnv),
			this->rightExp->evalAs<RightValue>(runtimeEnv),
			RuntimeEnvironment::opSlashMap,
			this->beginOperatorSourcePosition,
			this->endOperatorSourcePosition);
}

intrusive_ptr<Value> SummationExpression::implEval(RuntimeEnvironment *runtimeEnv) const {
	vector<Token>::const_iterator opIt = this->operators.begin();
	vector<intrusive_ptr<Expression> >::const_iterator expIt = this->operands.begin();
	vector<intrusive_ptr<Expression> >::const_iterator expEnd = this->operands.end();
	assert(expIt!=expEnd);
	intrusive_ptr<RightValue> currValue = (*expIt++)->evalAs<RightValue>(runtimeEnv);
	while (expIt!=expEnd) {
		assert(opIt!=operators.end());
		const Token op = *opIt++;
		const TokenType tt = op.getType();
		assert(tt==TT_PLUS || tt==TT_MINUS);
		currValue = runtimeEnv->binaryOperatorDispatch(
					currValue,
					(*expIt++)->evalAs<RightValue>(runtimeEnv),
					tt == TT_PLUS ? RuntimeEnvironment::opPlusMap : RuntimeEnvironment::opMinusMap,
					op.getBegin(),
					op.getEnd());
	}
	return currValue;
}

intrusive_ptr<Value> MultiplicationExpression::implEval(RuntimeEnvironment *runtimeEnv) const {
	vector<Token>::const_iterator opIt = this->operators.begin();
	vector<intrusive_ptr<Expression> >::const_iterator expIt = this->operands.begin();
	vector<intrusive_ptr<Expression> >::const_iterator expEnd = this->operands.end();
	assert(expIt!=expEnd);
	intrusive_ptr<RightValue> currValue = (*expIt++)->evalAs<RightValue>(runtimeEnv);
	while (expIt!=expEnd) {
		assert(opIt!=operators.end());
		const Token op = *opIt++;
		RuntimeEnvironment::DispatchMapType *map;
		switch(op.getType()) {
		case TT_STAR:
			map = &RuntimeEnvironment::opStarMap;
			break;
		case TT_SLASH:
			map = &RuntimeEnvironment::opSlashMap;
			break;
		case TT_MOD:
			map = &RuntimeEnvironment::opModMap;
			break;
		case TT_COLON:
			map = &RuntimeEnvironment::opColonMap;
			break;
		default:
			assert(false);
			throw RuntimeException("Something rather unpleasant has happened (op.getType() returned an unknown value) and I'm way too confused to continue", op.getBegin(), op.getEnd());
		}
		currValue = runtimeEnv->binaryOperatorDispatch(currValue, (*expIt++)->evalAs<RightValue>(runtimeEnv), *map, op.getBegin(), op.getEnd());
	}
	return currValue;
}

Frame *StaticEnv::VarData::tryToFindFrame(RuntimeEnvironment *runtimeEnv) const {
	assert(this->depth>StaticEnv::IN_THE_WILD);
	Frame *f;
	if (this->depth==StaticEnv::TOP_LEVEL) {
#ifdef VERBOSE_RUNTIME_DEBUG
		cout << "StaticEnv::VarData::tryToFindFrame f = toplevel\n";
#endif // #ifdef VERBOSE_RUNTIME_DEBUG
		f = runtimeEnv->getTopLevelFrame();
		assert(f);
	} else {
		f = runtimeEnv->getCurrentFrame();
		assert(f);
#ifdef VERBOSE_RUNTIME_DEBUG
		cout << "StaticEnv::VarData::tryToFindFrame f=" << f << ", depth=" << this->depth << endl;
#endif // #ifdef VERBOSE_RUNTIME_DEBUG
		for(int a=0; a<this->depth; ++a) {
#ifdef VERBOSE_RUNTIME_DEBUG
			cout << "StaticEnv::VarData::tryToFindFrame a=" << a << ", f=";
#endif // #ifdef VERBOSE_RUNTIME_DEBUG
			f = f->accessLink.toCheckedPointer();
#ifdef VERBOSE_RUNTIME_DEBUG
			cout << "StaticEnv::VarData::tryToFindFrame final f=" << f << endl;
#endif // #ifdef VERBOSE_RUNTIME_DEBUG
			if (!f)
				break;
		}
	}
	return f;
}

#ifdef C5IDE
const Frame *StaticEnv::VarData::debuggerTryToFindFrame(const Frame *topLevelFrame, const Frame *currentFrame) const {
	assert(this->depth>StaticEnv::IN_THE_WILD);
	const Frame *f;
	if (this->depth==StaticEnv::TOP_LEVEL) {
		f = topLevelFrame;
		assert(f);
	} else {
		f = currentFrame;
		assert(f);
		for(int a=0; a<this->depth; ++a) {
			f = f->accessLink.toCheckedPointer();
			if (!f)
				break;
		}
	}
	return f;
}
#endif // #ifdef C5IDE

intrusive_ptr<Value> Identifier::implEval(RuntimeEnvironment *runtimeEnv) const {
#ifdef VERBOSE_RUNTIME_DEBUG
	cout << "Identifier::eval Finding frame for " << this->identifier << endl;
#endif // #ifdef VERBOSE_RUNTIME_DEBUG
	Frame * const f = this->varData.tryToFindFrame(runtimeEnv);
	if (!f)
		throw DeadEnviromentException(this);
	if (this->varData.isCapturedValue) {
		assert(f->userdefinedFun);
		assert(static_cast<vector<intrusive_ptr<RightValue> >::size_type>(this->varData.index) < f->userdefinedFun->capturedValues.size());
		const intrusive_ptr<Value> v(f->userdefinedFun->capturedValues[this->varData.index]);
#ifdef VERBOSE_RUNTIME_DEBUG
		cout << "Identifier::eval: it's the captured value v=" << v << endl;
#endif // #ifdef VERBOSE_RUNTIME_DEBUG
		return v;
	}
	const int index = this->varData.index;
#ifdef VERBOSE_RUNTIME_DEBUG
	cout << "Identifier::eval index=" << index << endl;
#endif // #ifdef VERBOSE_RUNTIME_DEBUG
	if (index>=0) {
		const intrusive_ptr<Value> v = f->varSlots[index].value;
#ifdef VERBOSE_RUNTIME_DEBUG
		cout << "Identifier::eval v=" << v << endl;
#endif // #ifdef VERBOSE_RUNTIME_DEBUG
		if (dynamic_pointer_cast<LeftValue>(v)) {
#ifdef VERBOSE_RUNTIME_DEBUG
			cout << "Identifier::eval returning a ReferenceVariable\n";
#endif // #ifdef VERBOSE_RUNTIME_DEBUG
			return new ReferenceVariable(this, f, index);
		}
	}
#ifdef VERBOSE_RUNTIME_DEBUG
		cout << "Identifier::eval returning a VariableName\n";
#endif // #ifdef VERBOSE_RUNTIME_DEBUG
	return new VariableName(runtimeEnv, f, index, this);
}

intrusive_ptr<Value> IsDefinedExp::implEval(RuntimeEnvironment *runtimeEnv) const {
	if (this->expId->varData.isCapturedValue) {
		runtimeEnv->interpreter->reportWarning("This expression will be always true, since imported values are always defined", this);
		return BoolValue::trueValue;
	}
	VariableSlot *vs;
	try {
		vs = this->expId->getVariableSlot(runtimeEnv);
	} catch (VariableNotFoundException &) {
		return BoolValue::falseValue;
	}
	assert(vs->value); // getVariableSlot should throw an exception if vs->value==0 (this might happen when a package is reloaded or an indeterminate removed)
	if (dynamic_pointer_cast<VoidValue>(vs->value)) // void means optional-arg without value
		return BoolValue::falseValue;
	if (const intrusive_ptr<LeftValue> lv = dynamic_pointer_cast<LeftValue>(vs->value))
		try {
			lv->asRightValue();
		} catch (const InterruptException &) {
			throw;
		} catch (const RuntimeException &) {
			return BoolValue::falseValue;
		}
	return BoolValue::trueValue;
}

intrusive_ptr<Value> FieldAccessExpression::implEval(RuntimeEnvironment *runtimeEnv) const {
	intrusive_ptr<Value> v = this->targetExp->eval(runtimeEnv);
	if (const intrusive_ptr<LeftValue> lv = dynamic_pointer_cast<LeftValue>(v)) {
		if (const intrusive_ptr<VariableName> vn = dynamic_pointer_cast<VariableName>(lv))
			if (vn->expId->identifier[0]=='$') {
				intrusive_ptr<PackageValue> pv = intrusive_ptr_cast<PackageValue>(vn->asRightValue());
				return pv->toVariableName(runtimeEnv, this->name, this->tokName);
			}
		return new FieldAccess(lv, this->name, intrusive_ptr<const FieldAccessExpression>(this));
	}
	intrusive_ptr<RecordValue> rv = dynamic_pointer_cast<RecordValue>(v);
	if (!rv)
		throw WrongTypeException(
				RecordValue::type->name+" or "+PackageValue::type->name,
				intrusive_ptr_cast<RightValue>(v)->getType()->name,
				this->targetExp->getBegin(),
				this->targetExp->getEnd());
	intrusive_ptr<RightValue> fieldValue = rv->getField(this->name);
	if (!fieldValue)
		throw FieldNotFoundException(this->name, this->tokName.getBegin(), this->tokName.getEnd(), rv);
	return fieldValue;
}

intrusive_ptr<Value> IndexedAccessExpression::implEval(RuntimeEnvironment *runtimeEnv) const {
	intrusive_ptr<Value> v = this->targetExp->eval(runtimeEnv);
	const CharPointer targetExpBegin = this->targetExp->getBegin();
	CharPointer targetExpEnd = this->targetExp->getEnd();
	typedef intrusive_ptr<Expression> expPtr;
	BOOST_FOREACH(expPtr indexExp, this->indexes) {
		const intrusive_ptr<RightValue> indexValue = indexExp->evalAs<RightValue>(runtimeEnv);
		const CharPointer indexExpBegin = indexExp->getBegin();
		const CharPointer indexExpEnd = indexExp->getEnd();
		if (intrusive_ptr<BigIntValue> N = dynamic_pointer_cast<BigIntValue>(indexValue))
			v = v->indexedByBigInt(N, targetExpBegin, targetExpEnd, indexExpBegin, indexExpEnd);
		else if (intrusive_ptr<StringValue> str = dynamic_pointer_cast<StringValue>(indexValue))
			v = v->indexedByString(str, targetExpBegin, targetExpEnd, indexExpBegin, indexExpEnd);
		else
			throw WrongTypeException(BigIntValue::type->name + " or " + StringValue::type->name, indexValue->getType()->name, indexExpBegin, indexExpEnd);
		targetExpEnd = indexExpEnd;
	}
	return v;
}

intrusive_ptr<Value> RecordExpression::implEval(RuntimeEnvironment *runtimeEnv) const {
	intrusive_ptr<RecordValue> r(new RecordValue());
	BOOST_FOREACH(const RecordField &rf, this->fields) {
		intrusive_ptr<const Identifier> idExp = intrusive_ptr_cast<const Identifier>(rf.name);
		if (r->getField(idExp->identifier))
			throw RuntimeException("Duplicate field", idExp);
		r->setField(idExp->identifier, rf.initExp->evalAs<RightValue>(runtimeEnv));
	}
	return r;
}

intrusive_ptr<Value> ListExpression::implEval(RuntimeEnvironment *runtimeEnv) const {
	intrusive_ptr<ListValue> l(new ListValue());
	BOOST_FOREACH(intrusive_ptr<const Expression> e, this->exps) {
		l->addValue(e->evalAs<RightValue>(runtimeEnv));
	}
	return l;
}

intrusive_ptr<Value> CartesianProductExpression::implEval(RuntimeEnvironment *runtimeEnv) const {
	intrusive_ptr<ListValue> previous(new ListValue());
	previous->addValue(new ListValue(this->operands.size()));
	intrusive_ptr<ListValue> current;
	BOOST_FOREACH(intrusive_ptr<const Expression> e, this->operands) {
		intrusive_ptr<ListValue> l = e->evalAs<ListValue>(runtimeEnv);
		const ListValue::ContainerType::size_type lSize = l->size();
		const ListValue::ContainerType::size_type previousSize = previous->size();
		current = new ListValue();
		for(ListValue::ContainerType::size_type a=0; a<previousSize; ++a) {
			for(ListValue::ContainerType::size_type b=0; b<lSize; ++b) {
				intrusive_ptr<ListValue> cloned = intrusive_ptr_cast<ListValue>(previous->getValue(a)->clone());
				cloned->addValue(l->getValue(b));
				current->addValue(cloned);
			}
		}
		previous = current;
	}
	return current;
}

intrusive_ptr<Value> ExpSuchThatIdInExpAndExp::implEval(RuntimeEnvironment *runtimeEnv) const {
	intrusive_ptr<ListValue> list = this->exp2->evalAs<ListValue>(runtimeEnv);
	intrusive_ptr<ListValue> result(new ListValue());
	Frame *f = runtimeEnv->pushIterationFrame(this);
	BOOST_SCOPE_EXIT( (runtimeEnv) ) {
		runtimeEnv->popFrame();
	} BOOST_SCOPE_EXIT_END
	VariableSlot &varSlot = f->varSlots.front();
	const int size = list->size();
	for(int a=0; a<size; ++a) {
		intrusive_ptr<RightValue> sourceElem = list->getValue(a);
		assert(sourceElem);
		varSlot.value = sourceElem;
		if (!this->optionalExp3 || this->optionalExp3->evalAs<BoolValue>(runtimeEnv)->theBool)
			result->addValue(this->exp1->evalAs<RightValue>(runtimeEnv));
	}
	return result;
}

intrusive_ptr<Value> IdInExpSuchThatExp::implEval(RuntimeEnvironment *runtimeEnv) const {
	intrusive_ptr<ListValue> list = this->exp1->evalAs<ListValue>(runtimeEnv);
	Frame *f = runtimeEnv->pushIterationFrame(this);
	BOOST_SCOPE_EXIT( (runtimeEnv) ) {
		runtimeEnv->popFrame();
	} BOOST_SCOPE_EXIT_END
	VariableSlot &varSlot = f->varSlots.front();
	intrusive_ptr<ListValue> result(new ListValue());
	const int size = list->size();
	for(int a=0; a<size; ++a) {
		intrusive_ptr<RightValue> elem = list->getValue(a);
		assert(elem);
		varSlot.value = elem;
		if (this->exp2->evalAs<BoolValue>(runtimeEnv)->theBool)
			result->addValue(elem);
	}
	return result;
}

intrusive_ptr<Value> DotDotExpression::implEval(RuntimeEnvironment *runtimeEnv) const {
	const intrusive_ptr<ListValue> l(new ListValue());
	const intrusive_ptr<RightValue> lowerV = this->leftExp->evalAs<RightValue>(runtimeEnv);
	if (const intrusive_ptr<BigIntValue> NV = dynamic_pointer_cast<BigIntValue>(lowerV)) {
		BigInt lower(NV->theInteger);
		const BigInt upper(this->rightExp->evalAs<BigIntValue>(runtimeEnv)->theInteger);
		for(; lower<=upper; ++lower)
			l->addValue(new BigIntValue(lower));
	} else if (const intrusive_ptr<RingElemValue> ringElemLowerV = dynamic_pointer_cast<RingElemValue>(lowerV)) {
		const intrusive_ptr<RingElemValue> ringElemUpperV = this->rightExp->evalAs<RingElemValue>(runtimeEnv);
		const ring theRing(owner(ringElemLowerV->theRingElem));
		if (theRing!=owner(ringElemUpperV->theRingElem))
			throw RuntimeException("Cannot eval the dot-dot notation with ring-elems owned by different rings", this);
		if (!IsPolyRing(theRing))
			throw RuntimeException("The ring is not a PolyRing", this);
		const PolyRing polyRing(AsPolyRing(theRing));
		long lower, upper;
		if (!IsIndet(lower, ringElemLowerV->theRingElem))
			throw RuntimeException("The lower bound is not an indeterminate", this->leftExp);
		if (!IsIndet(upper, ringElemUpperV->theRingElem))
			throw RuntimeException("The upper bound is not an indeterminate", this->rightExp);
		for(; lower<=upper; ++lower)
			l->addValue(new RingElemValue(indet(polyRing, lower)));
	} else
		throw WrongTypeException(BigIntValue::type->name+" or "+RingElemValue::type->name, lowerV->getType()->name, this);
	return l;
}

intrusive_ptr<Value> InvocationExpression::implEval(RuntimeEnvironment *runtimeEnv) const {
	intrusive_ptr<RightValue> target = this->targetExp->evalAs<RightValue>(runtimeEnv);
	intrusive_ptr<FunctionValue> fun = dynamic_pointer_cast<FunctionValue>(target);
	if (!fun)
		throw WrongTypeException(FunctionValue::type->name,
                             target->getType()->name + " (maybe you forgot \"*\"?)",
                             this->targetExp);
	return fun->eval(this, runtimeEnv);
}

namespace {
	intrusive_ptr<Value> dontKnow(const intrusive_ptr<const Statement> s) {
		throw RuntimeException("I don't know (yet) how to execute this statement", s->getBegin(), s->getEnd());
	}
}

void UseStatement::implExecute(RuntimeEnvironment *runtimeEnv) {
	const intrusive_ptr<RingValue> ring = this->ringDefinition->eval(runtimeEnv);
	string ringName;
	if (this->identifier) {
		ringName = this->identifier->identifier;
		const intrusive_ptr<LeftValue> leftValue = this->identifier->evalAs<LeftValue>(runtimeEnv);
		if (leftValue->assignmentNeedsOwnership())
			leftValue->obtainOwnership();
		leftValue->assign(ring, this->ringDefinition->getBegin(), this->ringDefinition->getEnd(), runtimeEnv);
	}
	if (runtimeEnv->currentRing)
		runtimeEnv->currentRing->removeInjectedIndeterminates(runtimeEnv);
	runtimeEnv->currentRing = ring;
	runtimeEnv->getTopLevelFrame()->varSlots[runtimeEnv->currentRingSlot].value = ring;
	ring->injectIndeterminates(runtimeEnv, this->ringDefinition, ringName);
}

void UsingStatement::implExecute(RuntimeEnvironment *) {
	dontKnow(this);
}

void DescribeStatement::implExecute(RuntimeEnvironment *runtimeEnv) {
	this->exp->evalAs<RightValue>(runtimeEnv)->describe(runtimeEnv->getOutputStream());
}

void HelpStatement::implExecute(RuntimeEnvironment *runtimeEnv) {
	try {
		ostringstream os;
		OnlineHelp::PrintMan(this->topic, os);
		runtimeEnv->getOutputStream()->print(os.str())->flush();
	} catch (const std::exception& err) {
		throw RuntimeException(err.what(), this);
	}
}

intrusive_ptr<RingValue> RingDefinition::eval(InterpreterNS::RuntimeEnvironment *runtimeEnv) const {
	try {
		intrusive_ptr<RingValue> ring = this->identifier->evalAs<RingValue>(runtimeEnv);
		if (this->optionalExp) {
			const intrusive_ptr<RightValue> quotient = this->optionalExp->evalAs<RightValue>(runtimeEnv);
			if (const intrusive_ptr<BigIntValue> N = dynamic_pointer_cast<BigIntValue>(quotient))
                          ring = new RingValue(NewQuotientRing(ring->theRing, ideal(RingElem(ring->theRing, N->theInteger))));
			else if (const intrusive_ptr<RingElemValue> poly = dynamic_pointer_cast<RingElemValue>(quotient))
				ring = new RingValue(NewQuotientRing(ring->theRing, ideal(poly->theRingElem)));
			else if (const intrusive_ptr<IdealValue> ideal = dynamic_pointer_cast<IdealValue>(quotient))
				ring = new RingValue(NewQuotientRing(ring->theRing, ideal->theIdeal));
			else
				throw WrongTypeException(BigIntValue::type->name + ", " + RingElemValue::type->name + " or " + IdealValue::type->name, quotient->getType()->name, this->optionalExp);
		}
		if (!this->indeterminates.empty()) {
			vector<symbol> symbols;
			BOOST_FOREACH(intrusive_ptr<IndeterminateDeclaration> ind, this->indeterminates) {
				if (ind->ranges.empty()) {
					symbols.push_back(symbol(ind->identifier->identifier));
					continue;
				}
				vector<pair<long, long> > ranges;
				typedef pair<intrusive_ptr<Expression>, intrusive_ptr<Expression> > ExpPair;
				BOOST_FOREACH(const ExpPair &p, ind->ranges) {
					intrusive_ptr<BigIntValue> lower = p.first->evalAs<BigIntValue>(runtimeEnv);
					long l1;
					if (!IsConvertible(l1, lower->theInteger))
						throw RuntimeException("The index does not fit a machine-integer", p.first);
					intrusive_ptr<BigIntValue> upper = p.second ? p.second->evalAs<BigIntValue>(runtimeEnv) : lower;
					long l2;
					if (!IsConvertible(l2, upper->theInteger))
						throw RuntimeException("The index does not fit a machine-integer", p.second);
					if (l2<l1)
						throw RuntimeException("The upper bound is actually smaller than the lower bound", p.first->getBegin(), p.second->getEnd());
					ranges.push_back(make_pair(l1, l2));
				}
				vector<long> indexes;
				for(vector<pair<long, long> >::size_type a=0; a<ranges.size(); ++a)
					indexes.push_back(ranges[a].first);
				for(;;) {
					symbols.push_back(symbol(ind->identifier->identifier, indexes));
					int incIndex = indexes.size()-1;
					while (incIndex>=0)
						if ( ++indexes[incIndex]>ranges[incIndex].second ) {
							indexes[incIndex] = ranges[incIndex].first;
							--incIndex;
						} else
							break;
					if (incIndex==-1)
						break;
				}
			}
			/*ostringstream os;
			BOOST_FOREACH(const symbol &s, symbols)
				os << ' ' << s;
			runtimeEnv->getOutputStream()->print("Symbols:")->println(os.str());*/
			switch (this->maybeOrderTT) {
			case TT_LEX:
				ring = new RingValue(NewPolyRing(ring->theRing, symbols, lex));
				break;
			case TT_DEGLEX:
				ring = new RingValue(NewPolyRing(ring->theRing, symbols, StdDegLex));
				break;
			case TT_DEGREVLEX:
				ring = new RingValue(NewPolyRing(ring->theRing, symbols, StdDegRevLex));
				break;
			default:
				ring = new RingValue(NewPolyRing(ring->theRing, symbols));
			}
		}
		return ring;
	} catch (const ErrorInfo& err) {
		runtimeEnv->announceCLE(err);
		throw RuntimeException(err.what(), this);
	}
}

void RingAssignStatement::implExecute(RuntimeEnvironment *runtimeEnv) {
	const intrusive_ptr<LeftValue> leftValue = this->leftExp->evalAs<LeftValue>(runtimeEnv);
	runtimeEnv->interpreter->checkForInterrupts(this->leftExp);
	const intrusive_ptr<RingValue> ring = this->ringDef->eval(runtimeEnv);
	if (leftValue->assignmentNeedsOwnership())
		leftValue->obtainOwnership();
	leftValue->assign(ring, this->ringDef->getBegin(), this->ringDef->getEnd(), runtimeEnv);
}

void TimeStatement::implExecute(RuntimeEnvironment *) {
	dontKnow(this);
}

void EmptyStatement::implExecute(RuntimeEnvironment *) {
	// nothing to do here, what a lovely statement ;-)
}

void SkipStatement::implExecute(RuntimeEnvironment *) {
	// nothing to do here, what a lovely statement ;-)
}

VariableSlot *Identifier::getVariableSlot(RuntimeEnvironment *runtimeEnv) {
	Frame * const f = this->varData.tryToFindFrame(runtimeEnv);
	if (!f)
		throw DeadEnviromentException(this);
	int index = this->varData.index;
	if (index<0) {
		assert(f==runtimeEnv->getTopLevelFrame());
		index = runtimeEnv->slotFor(this->identifier);
		if (index<0)
			throw VariableNotFoundException(this, runtimeEnv);
		this->varData.index = index;
	}
	VariableSlot &vs = f->varSlots[index];
	if (!vs.value) { // this can only happen when a package has been reloaded (and some of the previous version members are not defined anymore) or an indeterminate has been removed (by Use-ing another ring)
		assert(f==runtimeEnv->getTopLevelFrame());
		throw VariableNotFoundException(this, runtimeEnv);
	}
	return &vs;
}

void ProtectStatement::implExecute(RuntimeEnvironment *runtimeEnv) {
	if (this->expId->varData.isCapturedValue)
		throw RuntimeException("Cannot protect a(n imported) value", this->expId);
	VariableSlot *vs = this->expId->getVariableSlot(runtimeEnv);
	if (vs->isSystemProtected())
		throw RuntimeException("This variable is already system-protected", this->expId);
	if (vs->isProtected())
        {
          string reason;
          if (!vs->getProtectionReason().empty())
            reason = " ("+vs->getProtectionReason()+")";
		throw RuntimeException("Variable already protected"+reason, this->expId);
        }
	string reason;
	if (this->optExp)
		reason = this->optExp->evalAs<StringValue>(runtimeEnv)->theString;
	vs->protect(reason);
}

void UnprotectStatement::implExecute(RuntimeEnvironment *runtimeEnv) {
	if (this->expId->varData.isCapturedValue)
		throw RuntimeException("Cannot unprotect a(n imported) value", this->expId);
	VariableSlot *vs = this->expId->getVariableSlot(runtimeEnv);
	if (vs->isSystemProtected())
		throw RuntimeException("Cannot unprotect a system-protected variable", this->expId);
	if (vs->isPackage())
		throw RuntimeException("Cannot unprotect a package-exported variable", this->expId);
	if (!vs->isProtected())
		runtimeEnv->interpreter->reportWarning("This variable was not protected", this->expId);
	vs->unprotect();
}

void IfStatement::implExecute(RuntimeEnvironment *runtimeEnv) {
	BOOST_FOREACH(const IfBranch &branch, this->branches)
		if (!branch.optExp || branch.optExp->evalAs<BoolValue>(runtimeEnv)->theBool) {
			branch.statements->execute(runtimeEnv);
			return;
		}
}

void CiaoOrQuitStatement::implExecute(RuntimeEnvironment *runtimeEnv) {
	runtimeEnv->getOutputStream()->println("Bye.");
	throw Ciao();
}

void SourceStatement::implExecute(RuntimeEnvironment *runtimeEnv) {
	const string filename = this->exp->evalAs<StringValue>(runtimeEnv)->theString;
	try {
		runtimeEnv->interpreter->readAndExecute(filename, false, this->ttype!=TT_LOAD);
	} catch (const RuntimeException &) {
		throw;
	} catch (const BaseException &be) {
		throw RuntimeException(be.reason, this->getBegin(), this->getEnd());
	}
}

void SourceRegionStatement::implExecute(RuntimeEnvironment *runtimeEnv) {
  try
  {
    const long FromLine = ConvertTo<long>(this->expFromLine->evalAs<BigIntValue>(runtimeEnv)->theInteger);
    const long FromChar = ConvertTo<long>(this->expFromChar->evalAs<BigIntValue>(runtimeEnv)->theInteger);
    const long ToLine = ConvertTo<long>(this->expToLine->evalAs<BigIntValue>(runtimeEnv)->theInteger);
    const long ToChar = ConvertTo<long>(this->expToChar->evalAs<BigIntValue>(runtimeEnv)->theInteger);

    const string filename = this->expFileName->evalAs<StringValue>(runtimeEnv)->theString;

    runtimeEnv->interpreter->readAndExecute(filename, false, true, FromLine,FromChar, ToLine,ToChar);
  }
  catch (const RuntimeException &)
  {
    throw;
  }
  catch (const BaseException &be)
  {
    throw RuntimeException(be.reason, this->getBegin(), this->getEnd());
  }
  catch (const CoCoA::ErrorInfo& err)
  {
    // Get here only if line no or char posn are outside range of long integer.
    throw RuntimeException("Line number or char position too large or negative", this->getBegin(), this->getEnd());
  }
}

void EvalStatement::implExecute(RuntimeEnvironment *runtimeEnv) {
	intrusive_ptr<RightValue> v = this->exp->evalAsRightValueVoidIsOk(runtimeEnv);
	assert(v);
	if (!dynamic_pointer_cast<VoidValue>(v)) {
		Frame *currentFrame = runtimeEnv->getCurrentFrame();
		if (currentFrame->invocationExp)
			throw RuntimeException("Ignored value inside fn-proc!  All values must be assigned"
					" or operated upon by a command/procedure (maybe you forgot to assign, print, or return it?)", this->getBegin(), this->getEnd());
		else
			runtimeEnv->getTopLevelFrame()->varSlots[runtimeEnv->itSlot].value = v;
		const intrusive_ptr<OutputStreamValue> output(runtimeEnv->getOutputStream());
		output->print(runtimeEnv, v, this->exp);
		output->newline();
#ifdef PRINT_REFCOUNTS
		output->print(v->refCountAsString());
#endif // #ifdef PRINT_REFCOUNTS
	}
}

void PrintStatement::implExecute(RuntimeEnvironment *runtimeEnv) {
	intrusive_ptr<OutputStreamValue> out = this->onExp ?
			this->onExp->evalAs<OutputStreamValue>(runtimeEnv) :
			runtimeEnv->getOutputStream();
	BOOST_FOREACH(intrusive_ptr<const Expression> e, this->exps) {
		intrusive_ptr<RightValue> v = e->evalAs<RightValue>(runtimeEnv);
		out->print(runtimeEnv, v, e);
	}
	if (this->ttype==TT_PRINTLN)
		out->newline();
	out->flush();
}

void ForStatement::implExecute(RuntimeEnvironment *runtimeEnv) {
	BigInt begin(this->beginExp->evalAs<BigIntValue>(runtimeEnv)->theInteger);
	BigInt end(this->endExp->evalAs<BigIntValue>(runtimeEnv)->theInteger);
	BigInt step(1);
	if (this->stepExp) {
		step = this->stepExp->evalAs<BigIntValue>(runtimeEnv)->theInteger;
		if (IsZero(step))
			throw RuntimeException("Step value cannot be zero", this->stepExp);
	}
	bool goingUp = step>0;
	Frame *f = runtimeEnv->pushIterationFrame(this);
	BOOST_SCOPE_EXIT( (runtimeEnv) ) {
		runtimeEnv->popFrame();
	} BOOST_SCOPE_EXIT_END
	VariableSlot &varSlot = f->varSlots.front();
	while( (goingUp && begin<=end) || (!goingUp && begin>=end)) {
		runtimeEnv->interpreter->checkForInterrupts(this);
		varSlot.value = new BigIntValue(begin);
		try {
			this->statements->execute(runtimeEnv);
		} catch (const Break &b) {
			if (b.label.length() && this->label!=b.label)
				throw;
			return;
		} catch (const Continue &c) {
			if (c.label.length() && this->label!=c.label)
				throw;
		}
		begin += step;
	}
}

void Statements::implExecute(RuntimeEnvironment *runtimeEnv) {
	BOOST_FOREACH(intrusive_ptr<Statement> s, this->statements)
		s->execute(runtimeEnv);
}

void ForeachStatement::implExecute(RuntimeEnvironment *runtimeEnv) {
	intrusive_ptr<ListValue> list = this->inExp->evalAs<ListValue>(runtimeEnv);
	Frame *f = runtimeEnv->pushIterationFrame(this);
	BOOST_SCOPE_EXIT( (runtimeEnv) ) {
		runtimeEnv->popFrame();
	} BOOST_SCOPE_EXIT_END
	VariableSlot &varSlot = f->varSlots.front();
	const int size = list->size();
	for(int a=0; a<size; ++a) {
		runtimeEnv->interpreter->checkForInterrupts(this);
		intrusive_ptr<RightValue> sourceElem = list->getValue(a);
		assert(sourceElem);
		varSlot.value = sourceElem;
		try {
			this->statements->execute(runtimeEnv);
		} catch (const Break &b) {
			if (b.label.length() && this->label!=b.label)
				throw;
			return;
		} catch (const Continue &c) {
			if (c.label.length() && this->label!=c.label)
				throw;
			continue;
		}
	}
}

void TryStatement::implExecute(RuntimeEnvironment *runtimeEnv) {
	try	{
		this->tryStatements->execute(runtimeEnv);
	} catch (InterruptException &) {
		throw;
	} catch (RuntimeException &exc) {
		Frame *f = runtimeEnv->pushIterationFrame(this);
		BOOST_SCOPE_EXIT( (runtimeEnv) ) {
			runtimeEnv->popFrame();
		} BOOST_SCOPE_EXIT_END
		VariableSlot &varSlot = f->varSlots.front();
		varSlot.value = new ErrorValue(exc.reason);
		this->uponErrorStatements->execute(runtimeEnv);
	}
}

void BlockStatement::implExecute(RuntimeEnvironment *runtimeEnv) {
	this->statements->execute(runtimeEnv);
}

void AliasStatement::implExecute(RuntimeEnvironment *) {
	//if (this->statements)
	//	this->statements->execute(runtimeEnv);
}

void WhileStatement::implExecute(RuntimeEnvironment *runtimeEnv) {
	while(this->exp->evalAs<BoolValue>(runtimeEnv)->theBool) {
		runtimeEnv->interpreter->checkForInterrupts(this);
		try {
			this->statements->execute(runtimeEnv);
		} catch (const Break &b) {
			if (b.label.length() && this->label!=b.label)
				throw;
			return;
		} catch (const Continue &c) {
			if (c.label.length() && this->label!=c.label)
				throw;
		}
	}
}

void RepeatUntilStatement::implExecute(RuntimeEnvironment *runtimeEnv) {
	const bool forever = !this->optExp;
	do {
		runtimeEnv->interpreter->checkForInterrupts(this);
		try {
			this->statements->execute(runtimeEnv);
		} catch (const Break &b) {
			if (b.label.length() && this->label!=b.label)
				throw;
			return;
		} catch (const Continue &c) {
			if (c.label.length() && this->label!=c.label)
				throw;
		}
	} while(forever || !this->optExp->evalAs<BoolValue>(runtimeEnv)->theBool);
}

namespace {
	intrusive_ptr<PackageValue> findOwner(RuntimeEnvironment *runtimeEnv, string name) {
		BOOST_FOREACH(const VariableSlot &vs, runtimeEnv->getTopLevelFrame()->varSlots) {
			if (intrusive_ptr<PackageValue> p = dynamic_pointer_cast<PackageValue>(vs.value)) {
				BOOST_FOREACH(const Token &t, p->pkgDecl->exportedNames) {
					if (t.lexeme()==name)
						return p;
				}
			}
		}
		return 0;
	}
}

void PackageStatement::implExecute(RuntimeEnvironment *runtimeEnv) {
	Frame * const tlFrame = runtimeEnv->getTopLevelFrame();
	int tlIndex = runtimeEnv->slotFor(this->name);
	if (tlIndex>=0) {
		assert(tlFrame->varSlots[tlIndex].value);
		intrusive_ptr<PackageValue> oldPkg = intrusive_ptr_cast<PackageValue>(tlFrame->varSlots[tlIndex].value);
		runtimeEnv->interpreter->reportWarning("Redefinition of package "+this->name, this->beginSourcePosition, this->tokName.getEnd());
		BOOST_FOREACH(int index, oldPkg->ownedTopLevelIndexes) {
			assert(index>=0 && static_cast<vector<VariableSlot>::size_type >(index)<tlFrame->varSlots.size());
			VariableSlot &vs = tlFrame->varSlots[index];
			assert(vs.value);
			vs.resetFlags();
			vs.value = 0;
		}
	}
	const string protectionReason = "Owned by package "+this->name;
	intrusive_ptr<PackageValue> newPackage(new PackageValue(this));
	const int pkgIndex = runtimeEnv->setTopLevelVar(this->name, newPackage, VariableSlot::VSF_SystemProtected);
	tlFrame->varSlots[pkgIndex].setProtectionReason(protectionReason);
	ResolvePackageNamesVisitor rpnv(*this);
	typedef pair<string, intrusive_ptr<const AssignmentStatement> > Pair;
	vector<Pair> memberVars;
	BOOST_FOREACH(intrusive_ptr<Statement> statement, this->statements->statements) {
		if (dynamic_pointer_cast<const AliasStatement>(statement))
			continue;
		if (const intrusive_ptr<const AssignmentStatement> assignment = dynamic_pointer_cast<const AssignmentStatement>(statement)) {
			intrusive_ptr<Identifier> idExp = intrusive_ptr_cast<Identifier>(assignment->leftExp);
			const string identifier = this->name+"."+idExp->identifier;
			memberVars.push_back(make_pair(identifier, assignment));
			const int index = runtimeEnv->setTopLevelVar(identifier, VoidValue::theInstance, VariableSlot::VSF_None);
			newPackage->ownedTopLevelIndexes.push_back(index);
		} else if (const intrusive_ptr<const DefineStatement> define = dynamic_pointer_cast<const DefineStatement>(statement)) {
			define->funDecl->accept(&rpnv);
			const int index = runtimeEnv->setTopLevelVar(this->name+"."+define->name, new UserDefinedFunction(runtimeEnv, define->funDecl), VariableSlot::VSF_SystemProtected);
			newPackage->ownedTopLevelIndexes.push_back(index);
			tlFrame->varSlots[index].setProtectionReason(protectionReason);
		} else
			assert(false);
	}
	BOOST_FOREACH(const Token &tokName, this->exportedNames) {
		const string name(tokName.lexeme());
		int index = runtimeEnv->slotFor(name);
		if (index>=0) {
			if (tlFrame->varSlots[index].isProtected()) {
				runtimeEnv->interpreter->reportError("Cannot export the name \""+name+"\" because there is a top-level protected variable with the same name", tokName.getBegin(), tokName.getEnd());
				continue;
			}
			if (tlFrame->varSlots[index].isPackage()) {
				intrusive_ptr<PackageValue> owner = findOwner(runtimeEnv, name);
				assert(owner);
				runtimeEnv->interpreter->reportError("Cannot export the name \""+name+"\" because the same name was already exported by package "+owner->pkgName, tokName.getBegin(), tokName.getEnd());
				continue;
			}
		}
		const int indexFullyQualified = runtimeEnv->slotFor(this->name+"."+name);
		assert(indexFullyQualified>=0);
		index = runtimeEnv->setTopLevelVar(name, VoidValue::theInstance, VariableSlot::VSF_Package);
		newPackage->ownedTopLevelIndexes.push_back(index);
		intrusive_ptr<Identifier> id(new Identifier(tokName));
		id->varData.depth = StaticEnv::TOP_LEVEL;
		id->varData.index = indexFullyQualified;
		tlFrame->varSlots[index].value = new VariableName(runtimeEnv, tlFrame, indexFullyQualified, id);
		tlFrame->varSlots[indexFullyQualified].setHasBeenExported();
	}
	BOOST_FOREACH(const Pair &assignment, memberVars) {
		const intrusive_ptr<Expression> rightExp = assignment.second->rightExp;
		rightExp->accept(&rpnv);
		runtimeEnv->setTopLevelVar(assignment.first, rightExp->evalAs<RightValue>(runtimeEnv), VariableSlot::VSF_None);
	}
}

void DefineStatement::implExecute(RuntimeEnvironment *runtimeEnv) {
	const intrusive_ptr<UserDefinedFunction> fun(new UserDefinedFunction(runtimeEnv, this->funDecl));
	int slotIndex = runtimeEnv->slotFor(this->name);
	if (slotIndex<0)
		runtimeEnv->setTopLevelVar(this->name, fun);
	else {
		VariableSlot *vs = &(runtimeEnv->getTopLevelFrame()->varSlots[slotIndex]);
		checkProtection(vs, this->expId);
		vs->value = fun;
	}
}

intrusive_ptr<Value> FullyQualifiedIdentifier::implEval(RuntimeEnvironment *runtimeEnv) const {
	const int slot = runtimeEnv->slotFor(this->pkgName);
	if (slot<0)
		throw RuntimeException("Cannot find package \""+this->pkgName+"\"", this->tokPkgname);
	assert(runtimeEnv->getTopLevelFrame()->varSlots[slot].value);
	const intrusive_ptr<PackageValue> pv(intrusive_ptr_cast<PackageValue>(runtimeEnv->getTopLevelFrame()->varSlots[slot].value));
	return pv->toVariableName(runtimeEnv, this->id, this->tokId);
}

intrusive_ptr<Value> LambdaExpression::implEval(RuntimeEnvironment *runtimeEnv) const {
	return new UserDefinedFunction(runtimeEnv, this->funDecl);
}

void ReturnStatement::implExecute(RuntimeEnvironment *runtimeEnv) {
	throw Return(this->exp ? this->exp->evalAs<RightValue>(runtimeEnv) : VoidValue::theInstance);
}

void BreakStatement::implExecute(RuntimeEnvironment *) {
	throw Break(this->label);
}

void ContinueStatement::implExecute(RuntimeEnvironment *) {
	throw Continue(this->label);
}

void AssignmentStatement::implExecute(RuntimeEnvironment *runtimeEnv) {
	const intrusive_ptr<LeftValue> leftValue = this->leftExp->evalAs<LeftValue>(runtimeEnv);
	runtimeEnv->interpreter->checkForInterrupts(this->leftExp);
	intrusive_ptr<RightValue> rightValue = this->rightExp->evalAs<RightValue>(runtimeEnv);
	if (leftValue->assignmentNeedsOwnership())
		leftValue->obtainOwnership();
	leftValue->assign(rightValue, this->rightExp->getBegin(), this->rightExp->getEnd(), runtimeEnv);
}

} // namespace AST

namespace LexerNS {

using namespace InterpreterNS;

void ErrorReporter::reportError(const RuntimeException &exception) {
	this->reportError(exception.reason, exception.from, exception.to);
	const vector<SnapshotFrame>::size_type snapshotSize = exception.snapshot.size();
	if (snapshotSize) {
		this->printContext();
		int nesting=-1;
		bool dots=false;
		CharPointer from = exception.from;
		CharPointer to = exception.to;
		for(vector<SnapshotFrame>::size_type a=0; a<snapshotSize; ++a) {
			const SnapshotFrame &frame = exception.snapshot[a];
			++nesting;
			if ( (snapshotSize-a>10) && a>10 ) {
				if (dots) {
					--nesting;
					continue;
				}
				this->outputStream->print(string(nesting, ' '))->print("...\n");
				dots = true;
				continue;
			}
			if (nesting) {
				this->outputStream->print(string(nesting, ' '));
				this->printCalledBy();
			}
			const intrusive_ptr<const FunctionDeclaration> fnDecl(intrusive_ptr_cast<const FunctionDeclaration>(frame.block));
			if (fnDecl->fnName.length()) {
				this->outputStream->print("function ");
				this->printBold(fnDecl->fnName);
			} else
				this->printBold("anonymous-function");
			if (!this->reportLineNumberWhenMeaningful(from, to, false, false)) {
				if (frame.block)
					this->outputStream->print(" (previously defined at the prompt)");
				else
					this->outputStream->print(" at top-level");
			}
			this->outputStream->newline();
			from = frame.invocationExp->getBegin();
			to = frame.invocationExp->getEnd();
		}
		this->outputStream->print(string(++nesting, ' '))->print("called");
		if (!this->reportLineNumberWhenMeaningful(from, to, false, false))
			this->outputStream->print(" at top-level");
		this->outputStream->newline();
	}
}

} // namespace LexerNS

} // namespace CoCoA
